/* Copyright 2013 David Wolverton
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.googlecode.genericdao.search;

import java.lang.reflect.InvocationTargetException;
import java.math.BigDecimal;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang3.time.DateFormatUtils;
import org.joda.time.DateTime;

/**
 * Utilities for TRG Generic DAO Search
 * 
 * @author dwolverton
 */
public class InternalUtil {
	/**
	 * <p>
	 * Return an instance of the given class type that has the given value. For
	 * example, if type is <code>Long</code> and <code>Integer</code> type with
	 * the value 13 is passed in, a new instance of <code>Long</code> will be
	 * returned with the value 13.
	 * 
	 * <p>
	 * If the value is already of the correct type, it is simply returned.
	 * 
	 * @throws ClassCastException
	 *             if the value cannot be converted to the given type.
	 */
	public static Object convertIfNeeded(Object value, Class<?> type) throws ClassCastException {
		
		// Since we're returning an object, we will never be able to return a primitive value.
		// We will return the boxed type instead.
		if (type.isPrimitive()) {
			if (boolean.class.equals(type)) {
				type = Boolean.class;
			} else if (char.class.equals(type)) {
				type = Character.class;
			} else if (byte.class.equals(type)) {
				type = Byte.class;
			} else if (short.class.equals(type)) {
				type = Short.class;
			} else if (int.class.equals(type)) {
				type = Integer.class;
			} else if (long.class.equals(type)) {
				type = Long.class;
			} else if (float.class.equals(type)) {
				type = Float.class;
			} else if (double.class.equals(type)) {
				type = Double.class;
			}
		}
		
		if (value == null)
			return null;
		if (type.isInstance(value))
			return value;
		
		

		if (String.class.equals(type)) {
			return value.toString();
		} if (Boolean.class.equals(type) && value instanceof String && ("true".equals(value) || "false".equals(value))) {
			return Boolean.valueOf((String) value);
		} else if (Number.class.isAssignableFrom(type)) {
			// the desired type is a number
			if (value instanceof Number) {
				// the value is also a number of some kind. do a conversion
				// to the correct number type.
				Number num = (Number) value;

				if (type.equals(Double.class)) {
					return new Double(num.doubleValue());
				} else if (type.equals(Float.class)) {
					return new Float(num.floatValue());
				} else if (type.equals(Long.class)) {
					return Long.valueOf(num.longValue());
				} else if (type.equals(Integer.class)) {
					return Integer.valueOf(num.intValue());
				} else if (type.equals(Short.class)) {
					return Short.valueOf(num.shortValue());
				} else if (type.equals(BigDecimal.class)) {
					return new BigDecimal(num.toString());					
				} else {
					try {
						return type.getConstructor(String.class).newInstance(value.toString());
					} catch (IllegalArgumentException e) {
					} catch (SecurityException e) {
					} catch (InstantiationException e) {
					} catch (IllegalAccessException e) {
					} catch (InvocationTargetException e) {
					} catch (NoSuchMethodException e) {
					}
				}
			} else if (value instanceof String) {
				//the value is a String. attempt to parse the string
				//catch percentage strings
				boolean percentageConversion = false;
				if(((String) value).contains("%")) {
					percentageConversion = true;
					value = ((String) value).replace("%", "");
				}
				try {
					Number result = null;
					if (type.equals(Double.class)) {
						result = Double.parseDouble((String) value);
						if(result != null && percentageConversion) return ((Double)result) / 100;
					} else if (type.equals(Float.class)) {
						result = Float.parseFloat((String) value);
						if(result != null && percentageConversion) return ((Float)result) / 100;
					} else if (type.equals(Long.class)) {
						return Long.parseLong((String) value);
					} else if (type.equals(Integer.class)) {
						return Integer.parseInt((String) value);
					} else if (type.equals(Short.class)) {
						return Short.parseShort((String) value);
					} else if (type.equals(Byte.class)) {
						return Byte.parseByte((String) value);
					} else if (type.equals(BigDecimal.class)) {
						if(((String) value).isEmpty()) return BigDecimal.ZERO;
						result = new BigDecimal((String) value);
						if(result != null && percentageConversion) return ((BigDecimal)result).divide(BigDecimal.valueOf(100L));
					}	
					
					return result;
				} catch (NumberFormatException ex) {
					//fall through to the error thrown below
				}
			}
		} else if (Date.class.equals(type)) {
			try {
				if(value instanceof org.joda.time.DateTime) {
					return ((org.joda.time.DateTime)value).toDate();
				}
				if(((String) value).contains("T")) {
					DateFormat format = new SimpleDateFormat(DateFormatUtils.ISO_DATETIME_FORMAT.getPattern());
					return format.parse((String) value);
				} else {
					DateFormat format = new SimpleDateFormat(DateFormatUtils.ISO_DATE_FORMAT.getPattern());
					return format.parse((String) value);
				}
			} catch (ParseException e) {
				// begin mod: more flexible date parsing as localized date strings seems to get thorugh to here
				DateTime dateTime = DateTime.parse((String) value);
				if (dateTime != null) return dateTime.toDate();
				// end mod
				throw new ClassCastException("Unable to convert value " + value.toString() + " to type Date");
			}		
		} else if (Class.class.equals(type)) {
			try {
				return Class.forName(value.toString());
			} catch (ClassNotFoundException e) {
				throw new ClassCastException("Unable to convert value " + value.toString() + " to type Class");
			}
		} else if (type.isEnum()) {
			List<?> enumFields = Arrays.asList(type.getEnumConstants());
			// in case of a simple filter the value may be enclosed by % wildcards
			// hence we need to get rid of those
			String realValue = value.toString().toLowerCase().replace("%", "");
			for (Object e : enumFields) {
				if (e.toString().toLowerCase().equals(realValue)) {
					return e;
				}
			}
		}

		throw new ClassCastException("Unable to convert value of type " + value.getClass().getName() + " to type "
				+ type.getName());
	}

	public static String paramDisplayString(Object val) {
		return paramDisplayString(val,"\"");
	}
	
	public static String paramDisplayString(Object val, String delim) {
		if (val == null) {
			return "null";
		} else if (val instanceof String) {
			return delim + val + delim;
		} else if (val instanceof Collection) {
			StringBuilder sb = new StringBuilder();
			sb.append(val.getClass().getSimpleName());
			sb.append(" {");
			boolean first = true;
			for (Object o : (Collection<?>) val) {
				if (first) {
					first = false;
				} else {
					sb.append(", ");
				}
				sb.append(paramDisplayString(o));
			}
			sb.append("}");
			return sb.toString();
		} else if (val instanceof Object[]) {
			StringBuilder sb = new StringBuilder();
			sb.append(val.getClass().getComponentType().getSimpleName());
			sb.append("[] {");
			boolean first = true;
			for (Object o : (Object[]) val) {
				if (first) {
					first = false;
				} else {
					sb.append(", ");
				}
				sb.append(paramDisplayString(o));
			}
			sb.append("}");
			return sb.toString();
		} else {
			return val.toString();
		}
	}
}
