package net.pricefx.common.api.pa;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.ParseException;
import java.util.Date;
import java.util.Locale;

import net.pricefx.common.util.CalendarUtil;
import net.pricefx.common.util.SpecialValue;

import org.apache.commons.lang3.math.NumberUtils;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.ibm.icu.text.NumberFormat;

public enum DataType {

	STRING, NUMBER, DATE, BOOLEAN, DATETIME, LOB, LONG /*only for auto-inc id column at the moment; has no corresponding DMField.Type*/, UNKNOWN;

	private static final Logger log = LoggerFactory.getLogger(DataType.class);
	
	private static final ThreadLocal<DecimalFormat> DF = new ThreadLocal<DecimalFormat>() {

	    @Override
	    protected DecimalFormat initialValue() {

	        DecimalFormat df = new DecimalFormat();
	        DecimalFormatSymbols dfs = df.getDecimalFormatSymbols();
	        dfs.setDecimalSeparator('.');
	        dfs.setNaN("NaN");
	        dfs.setInfinity("Infinity");
	        df.setGroupingUsed(false);
	        df.setRoundingMode(RoundingMode.HALF_UP);
	        df.setDecimalFormatSymbols(dfs);
	        return df;
	    }
	};
	

	public static DataType get(final Object dataType){
		if (dataType != null && dataType instanceof String){
			try{
				return DataType.valueOf((String)dataType);
			}
			catch(IllegalArgumentException e){ /* return null is fine */ }
		}
		return null;
	}
	
	public Object format(Object value) {
		if (value != null){
			switch (this) {
				case NUMBER:
					if (value instanceof Double && Double.isNaN((Double)value)){
						// just return the NaN
						return value;
					}
					BigDecimal bd = null;
					if (value instanceof BigDecimal){
						bd = (BigDecimal)value;
					}
					else try{
						bd = new BigDecimal(value.toString());
					}
					catch(NumberFormatException nfe){
						// just return the garbage passed in (better for detecting the error condition)
						return value;
					}
					// keep 15 significant figures (in line with Excel's max precision, see PFCD-3242)
					int scale = 15 - bd.precision() + bd.scale();
					bd = bd.setScale(scale, RoundingMode.HALF_UP);
					return bd;
				case DATETIME:
					// return DateTime instead of just Date to make it clear this value carries datetime precision
					// and should be treated (for example in serialization) accordingly
					if (value instanceof Date) {
						return new DateTime(value);
					}
					break;
				case DATE:
				case STRING:
				case BOOLEAN:	
				case LOB:	
				case LONG:
				case UNKNOWN:	// note: no specific handling for these types (for now)
			}
		}
		return value;
	}
	
	
	// throws an IllegalArgumentException (incl. NumberFormatException) if not parsable to this data type. 
	public Object parse(final Object value) {
		return parse(value, null);
	}
	
	public Object parse(final Object value, final Locale locale) {
		Object parsedVal = null;
		if (SpecialValue.isNull(value)){		// isNull() also checks for SpecialValue.NULL
			parsedVal = value; // null remains null
		}
		else if (this != STRING && "".equals(value)){
			return null;
		}
		else switch (this){
			case STRING:
			case LOB:
			case UNKNOWN:
				parsedVal = value.toString().trim();
				break;
			case NUMBER:
				// if NaN made it to a serialization step (i.e. as a Double), the best we can do is reinstate this Double.
				// more typically, a Double is converted to BigDecimal before serialiation, blowing up in case of a NaN
				// (with the application code handling the resulting NumberFormatException as and where this happened).
				if (value instanceof Number) {
					parsedVal = value;
				}
				else if ("NaN".equals(value)) {
					parsedVal = Double.NaN;
				}
				else {
					parsedVal = NumberUtils.createBigDecimal(toNumber(value.toString(), locale).toString());
				}
				break;
			case BOOLEAN:
				if (value instanceof Boolean) {
					parsedVal = value;
				}
				else {
					parsedVal = toBoolean(value.toString());
				}
				break;
			case DATE:
				if (value instanceof Date) {
					parsedVal = value;
				}
				else if (value instanceof DateTime) {
					parsedVal = ((DateTime)value).toDate();
				}
				else {
					parsedVal = toDate(value.toString());
				}
				break;
			case DATETIME:
				if (value instanceof DateTime) {
					parsedVal = value;
				}
				else if (value instanceof Date) {
					parsedVal = new DateTime(value);
				}
				else {
					parsedVal = toDateTime(value.toString());
				}
				break;
			case LONG:
				if (value instanceof Long) {
					parsedVal = value;
				}
				else {
					parsedVal = NumberUtils.createLong(toNumber(value.toString(), locale).toString());
				}
				break;
			default:
				throw new IllegalArgumentException("Cannot parse'" + value + "' to " + this);
		}
		
		log.trace("parsed '{}' as {} to {}", value, this, parsedVal);
		return parsedVal;
	}
	
	private static Number toNumber(final String value, final Locale locale){
		try{
			return NumberFormat.getNumberInstance().parse(value);
		}
		catch(ParseException pe1){
			try{
				return NumberFormat.getNumberInstance(Locale.US).parse(value);
			}
			catch(ParseException pe2){
				if (locale != null && locale != Locale.US) try{
					return NumberFormat.getNumberInstance(locale).parse(value);
				}
				catch(ParseException pe3){ /* throwing IllegalArgumentException just below */ }
			}
		}
		throw new IllegalArgumentException("Not a Number: '" + value + "'");
	}

	public static String numberToString(Number number, int numberOfDecimals) {
	    DF.get().setMaximumFractionDigits(numberOfDecimals);
	    return DF.get().format(number);
	}
	 
	private static boolean toBoolean(final String value){
		if ("true".equalsIgnoreCase(value)) {
			return true;
		}
		else if ("false".equalsIgnoreCase(value)) {
			return false;
		}
		else if ("1".equalsIgnoreCase(value)) {
			return true;
		}
		else if ("0".equalsIgnoreCase(value)) {
			return false;
		}
		throw new IllegalArgumentException("Not a Boolean: '" + value + "'");
	}

	private static Date toDate(final String value){
		Date date = null; 
		if (value != null){
			date = new CalendarUtil().parseDate(value.toString());
		}
		if (date == null){ 
			throw new IllegalArgumentException("Not a Date: '" + value + "'");
		}
		return date;
	}

	private static DateTime toDateTime(final String value){
		DateTime datetime = null; 
		if (value != null){
			datetime = new CalendarUtil().parseDateTime(value.toString());
		}
		if (datetime == null){ 
			throw new IllegalArgumentException("Not a DateTime: '" + value + "'");
		}
		return datetime;
	}

	public String serialize(final Object value) {
		if (SpecialValue.isNull(value)){
			return null;
		}
		else if (value instanceof String){
			return (String)value;
		}
		else switch (this){
			case STRING:
			case LOB:
				return value.toString();		// we know it's not null
			case DATE:
				if (value instanceof Date){
					return CalendarUtil.getDefaultDateFormat().format((Date)value);
				}
				break;
			case DATETIME:
				if (value instanceof Date){
					return CalendarUtil.getDefaultDateTimeFormat().format((Date)value);
				}
				else if (value instanceof DateTime){
					return CalendarUtil.getDefaultDateTimeFormat().format(((DateTime)value).toDate());
				} 
				break;
			case NUMBER:
				if (value instanceof Number){
					return value.toString();
				}
				break;
			case LONG:
				if (value instanceof Long){
					return value.toString();
				}
				break;
			case BOOLEAN:
				if (value instanceof Boolean){
					return Boolean.toString((boolean)value);	// we know it's not null
				}
				break;
			case UNKNOWN:
				//returns null below
		}
		return null;
	}

	public Object deserialize(final String sValue) {
		return parse(sValue);
	}

	public boolean isAssignable(final Object value) {
		if (SpecialValue.isNull(value)){		// isNull() also checks for SpecialValue.NULL
			return true;
		}
		
		switch (this){
			case STRING:
			case LOB:
			case UNKNOWN:				
				return true;	// can always convert to string (as not null)
			case NUMBER:
				return DataType.isNumber(value);
			case BOOLEAN:
				return DataType.isBoolean(value);
			case DATE:
				return DataType.isDate(value);
			case DATETIME:
				return DataType.isDateTime(value);
			case LONG:
				return DataType.isLong(value);
			default:
				// do nothing
		}
		return false;
	}
	
	public boolean isNumeric(){
		return NUMBER.equals(this);
	}
	
	public boolean isTime(){
		return DATE == this || DATETIME == this;
	}

	public static DataType toDataType(final Class<?> clazz) {
		if (clazz == null){
			return null;
		}
		else if (String.class.isAssignableFrom(clazz)){
			return STRING;
		}
		else if (Number.class.isAssignableFrom(clazz)){
			return NUMBER;
		}
		else if (Boolean.class.isAssignableFrom(clazz)){
			return BOOLEAN;
		}
		else if (Date.class.isAssignableFrom(clazz)){
			return DATE;
		}
		else if (DateTime.class.isAssignableFrom(clazz)){
			return DATETIME;
		}
		else if (Long.class.isAssignableFrom(clazz)){
			return LONG;
		}
		return null;
	}

	public static DataType toDataType( final Object value, final DataType preferred ) {
		if (value == null || STRING == preferred)	// everything can be a string
			return preferred;
		
		Boolean isBool = null;
		Boolean isDate = null;
		Boolean isDateTime = null;
		Boolean isNumber = null;
		
		if (preferred != null){
			switch (preferred){
				case BOOLEAN:
					isBool = isBoolean(value);
					if (isBool.booleanValue()) return preferred;
					break;
				case DATE:
					isDate = isDate(value);
					if (isDate.booleanValue()) return preferred;
					break;
				case DATETIME:
					isDateTime = isDateTime(value);
					if (isDateTime.booleanValue()) return preferred;
					break;
				case NUMBER:
					isNumber = isNumber(value);
					if (isNumber.booleanValue()) return preferred;
					break;
				default:
					throw new IllegalArgumentException("Unsupported preferred DataType: " + preferred);
			}
		}
		
		if (isNumber == null) isNumber = isNumber(value);
		if (isNumber.booleanValue()) return NUMBER;

		if (isBool == null) isBool = isBoolean(value);
		if (isBool.booleanValue()) return BOOLEAN;

		if (isDate == null) isDate = isDate(value);
		if (isDate.booleanValue()) return DATE;
		
		if (isDateTime == null) isDateTime = isDateTime(value);
		if (isDateTime.booleanValue()) return DATETIME;

		return STRING;
	}

	public static boolean isBoolean(final Object value){
		try{
			BOOLEAN.parse(value);
			return true;
		}
		catch(IllegalArgumentException e){
			return false;
		}
	}

	public static boolean isDate(final Object value){
		try{
			DATE.parse(value);
			return true;
		}
		catch(IllegalArgumentException e){
			return false;
		}
	}

	public static boolean isDateTime(final Object value){
		try{
			DATETIME.parse(value);
			return true;
		}
		catch(IllegalArgumentException e){
			return false;
		}
	}		

	public static boolean isNumber(final Object value){
		try{
			NUMBER.parse(value);
			return true;
		}
		catch(IllegalArgumentException e){
			return false;
		}
	}

	public static boolean isLong(final Object value){
		try{
			LONG.parse(value);
			return true;
		}
		catch(IllegalArgumentException e){
			return false;
		}
	}

}