package net.pricefx.common.apibuilder.calculationflowaction;

import java.util.ArrayList;
import java.util.List;

import net.pricefx.common.apibuilder.Invocations;
import net.pricefx.common.dto.ChainedJobInfo;

/**
 * This action starts calculation of selected LPG for all skus or restricted set
 * of skus. It's a programmatic version of the same action available in UI. As
 * such it allows user to schedule the action in calculation flow formula.
 */
public class LivePriceGridAction<ATTACHEDTO extends Help> extends Action {

	private String livePriceGridLabel;
	private Boolean calculate;
	private List<String> restrictToSkus = new ArrayList<String>();
	private int priority;
	private int distributedCalculationThreshold;
	private ATTACHEDTO backLink;
	
	public LivePriceGridAction(ATTACHEDTO backLink, Invocations<MethodInvocationRecord> invocations, String livePriceGridLabel) {
		super(invocations);
		this.backLink = backLink;
		this.livePriceGridLabel = livePriceGridLabel;
		this.calculate = false;
	}

	/**
	 * Whether the action is valid and should be triggered or not.
	 * 
	 * @param calculate
	 *            <ul>
	 *            <li>true - calculation of the LPG is started as background
	 *            process</li>
	 *            <li>false - calculation of the LPG is NOT started at all</li>
	 *            </ul>
	 */
	public LivePriceGridAction<ATTACHEDTO> setCalculate(Boolean calculate) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setCalculate", calculate));
		this.calculate = Boolean.TRUE.equals(calculate) ? true : false;
		return this;
	}
	
	/**
	 * Sets the prority of this job. 
	 * 
	 * @param priority Default is 0 (lowest user-definable priority). Max value is 9.
	 * @return
	 */
	public LivePriceGridAction<ATTACHEDTO> setPriority(int priority) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setPriority", priority));
		this.priority = priority;
		return this;		
	}
	
	/**
	 * Sets a job-individual threshold to start a distributed calculation.
	 * Please note: Any values below 1000 may very well be counter-productive and make the job slower!
	 * 
	 * @param threshold The number of items that need to be part of the job in order to execute it in distributed mode.
	 * @return
	 */
	public LivePriceGridAction<ATTACHEDTO> setDistributedCalculationThreshold(int threshold) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setDistributedCalculationThreshold", threshold));
		this.distributedCalculationThreshold = threshold;
		return this;		
	}	
	
	/**
	 * Restricts LPG calculation to list of presented skus
	 * 
	 * @param skus
	 *            list of skus to process.s
	 */
	public LivePriceGridAction<ATTACHEDTO> restrictToSkus(List<String> skus) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".restrictToSkus", skus));
		if (skus != null) restrictToSkus = skus;
		else skus = new ArrayList<String>();
		return this;
	}

	public ATTACHEDTO back() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".back"));
		return backLink;
	}

	/**
	 * Defines subsequent follow-on jobs
	 *
	 * @param jobs List of ChainedJobInfo objects
	 * @return
	 */
	public LivePriceGridAction<ATTACHEDTO> setNextJobs(List<ChainedJobInfo> jobs) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setNextJobs", jobs));
		this.jobs = jobs;
		return this;
	}

	@Override
	public String getObjectName() {
		return backLink.getObjectName() + ".LivePriceGridAction";
	}

	public String getLivePriceGridLabel() { return livePriceGridLabel; }
	public Boolean getCalculate() {	return calculate; }
	public List<String> getRestrictToSkus() { return restrictToSkus; }
	public int getPriority() { return priority; }
	public int getDistributedCalculationThreshold() {	return distributedCalculationThreshold; }
	public List<ChainedJobInfo> getNextJobs() { return jobs; }
	
	@Override
	public String toString() {
		return "LivePriceGridAction [livePriceGridLabel=" + livePriceGridLabel
				+ ", calculate=" + calculate + ", restrictToSkus="
				+ restrictToSkus + ", backLink=" + backLink + ", hashCode()="
				+ hashCode() + "]";
	}
	
}