package net.pricefx.common.apibuilder.calculationflowaction;

import java.math.BigDecimal;

import net.pricefx.common.api.CFTraitType;
import net.pricefx.common.api.CalculationFlowTimeUnit;
import net.pricefx.common.apibuilder.Invocations;

/**
 * This action is helping you to get an information about schedule as seen in the UI + if needed alter it. <br/>
 * <br/>
 * This sample is showing how to change original schedule interval of 10 minutes (as defined in UI), to 2 minute interval and back.
 *
 * Running this element in CF logic would run CF in intervals START, 2, 8, 2, 8, 2, .....  → original 10 minute interval is intact, we are putting in there another 2 minute interval.
 * 
 * <pre>
 * 
 *   api.retainGlobal = true
 *
 *  // Schedules are: original interval, 2 minutes interval, original interval, 2 minutes interval, original interval
 *
 *  // let's say the original interval is 10 minutes and start at 9:34 then the
 *
 *  // sequence of schedules will be: 9:44, 9:46, 9:54, 9:56, 10:04, 10:06, 10:14                     
 *
 *  // you see that the 10 minute interval is kept, with one run inside that interval
 *
 *  def sharedState = api.global
 *
 *
 *  if (sharedState.condititionMet) {
 *
 *    // switching from original schedule per each run  to 2 minut schedule
 *
 *    sharedState.condititionMet = false
 *
 *    def settings = actionBuilder.getScheduleSettings()
 *
 *    // make note of original schedule
 *
 *    sharedState.startingDate = settings.getStartingDate()
 *
 *    sharedState.numberOfTimeUnits = settings.getNumberOfTimeUnits()
 *
 *    settings.setNewNumberOfTimeUnits(2)
 *
 *  } else {
 *
 *    // switching from 2 minute schedule to original schedule
 *
 *    sharedState.condititionMet = true
 *
 *
 *    def settings = actionBuilder.getScheduleSettings()
 *
 *    def startingDate = sharedState.remove("startingDate")
 *
 *    def numberOfTimeUnits = sharedState.remove("numberOfTimeUnits")
 *
 *     
 *
 *    settings.setNewStartingDate(startingDate)
 *
 *    settings.setNewNumberOfTimeUnits(numberOfTimeUnits)
 *
 *
 *    // note the logic will go there also for the first ever run, global state is not set yet,
 *
 *    // properties are null and using setting.setNew... with null values means "keep the original values"
 *
 *    // -> therefore first run is always the original schedule interval
 *
 *  }
 * 
 * </pre>
 * 
 */
public class ScheduleSettings<ATTACHEDTO extends Help> implements Help {
	

	public ScheduleSettings(ATTACHEDTO backLink, Invocations<MethodInvocationRecord> invocations, CalculationFlowTimeUnit timeUnit, BigDecimal numberOfTimeUnits, String startingDate, Boolean periodic, CFTraitType traitType) {
		this.backLink = backLink;
		this.invocations = invocations;
		this.timeUnit = timeUnit;
		this.numberOfTimeUnits = numberOfTimeUnits;
		this.startingDate = startingDate;
		this.periodic = periodic;
		this.traitType = traitType;
	}

	final private ATTACHEDTO backLink;
	final Invocations<MethodInvocationRecord> invocations;

	final private CalculationFlowTimeUnit timeUnit;
	final private BigDecimal numberOfTimeUnits;
	final private String startingDate;
	final private Boolean periodic;
	final private CFTraitType traitType;
	
	private CalculationFlowTimeUnit newTimeUnit;
	private BigDecimal newNumberOfTimeUnits;
	private String newStartingDate;
	private Boolean newPeriodic;
	
	@Override
	public String getObjectName() {
		return backLink.getObjectName() + ".ScheduleSettings";
	}

	/**
	 * Gets you time unit as seen in the UI. Possible options are MINUTE, HOUR, DAY.
	 */
	public CalculationFlowTimeUnit getTimeUnit() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".getTimeUnit"));
		return timeUnit;
	}

	/**
	 * Number of time units to next triggering of schedule.
	 */
	public BigDecimal getNumberOfTimeUnits() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".getNumberOfTimeUnits"));
		return numberOfTimeUnits;
	}

	/**
	 * Starting date and time of the associated formula
	 */
	public String getStartingDate() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".getStartingDate"));
		return startingDate;
	}

	/**
	 * If this CF item is period or not. Default option is that it is periodic.
	 */
	public Boolean isPeriodic() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".isPeriodic"));
		return periodic;
	}

	/**
	 * Method that alters the schedule for next run. Sets new time unit.
	 */
	public void setNewTimeUnit(CalculationFlowTimeUnit newTimeUnit) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setNewTimeUnit", newTimeUnit));
		this.newTimeUnit = newTimeUnit;
	}

	/**
	 * Method that alters the schedule for next run. Sets new number of time units.
	 */
	public void setNewNumberOfTimeUnits(BigDecimal newNumberOfTimeUnits) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setNewNumberOfTimeUnits", newNumberOfTimeUnits));
		this.newNumberOfTimeUnits = newNumberOfTimeUnits;
	}

	/**
	 * Method that alters the schedule for next run. Sets new starting date.
	 */
	public void setNewStartingDate(String newStartingDate) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setNewStartingDate", newStartingDate));
		this.newStartingDate = newStartingDate;
	}

	/**
	 * Method that alters the schedule for next run. Sets periodicity.
	 */
	public void setNewPeriodic(Boolean newPeriodic) {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".setNewPeriodic", newPeriodic));
		this.newPeriodic = newPeriodic;
	}

	/**
	 * Gets newly set time unit.
	 */
	public CalculationFlowTimeUnit getNewTimeUnit() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".getNewTimeUnit"));
		return newTimeUnit;
	}

	/**
	 * Gets newly set number of time units.
	 */
	public BigDecimal getNewNumberOfTimeUnits() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".getNewNumberOfTimeUnits"));
		return newNumberOfTimeUnits;
	}

	/**
	 * Gets newly set starting date.
	 */
	public String getNewStartingDate() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".getNewStartingDate"));
		return newStartingDate;
	}

	/**
	 * Gets newly set periodicity.
	 */
	public Boolean isNewPeriodic() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".isNewPeriodic"));
		return newPeriodic;
	}
	
	public CFTraitType getTraitType() {
		return traitType;
	}


	/**
	 * 
	 * @return Retuns true if the CF was run on user request (not by scheduler
	 *         itself) either manually in UI or via REST API action
	 */
	public boolean isStartImmediatelyTraitActive() {
		return CFTraitType.START_IMMEDIATELY.equals(traitType);
	}

	public boolean isScheduleChangeRequested() {
		boolean isScheduleChangeRequested = (
				newNumberOfTimeUnits != null ||
				newPeriodic != null ||
				newStartingDate != null ||
				newTimeUnit != null
				);
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".isScheduleChangeRequested", isScheduleChangeRequested));
		return isScheduleChangeRequested;
	}
	
	public void clearNewSchedule() {
		this.newNumberOfTimeUnits = null;
		this.newPeriodic = null;
		this.newStartingDate = null;
		this.newTimeUnit = null;
	}
	
	public ATTACHEDTO back() {
		invocations.add(MethodInvocationRecord.of(getObjectName() + ".back"));
		return backLink;
	}
	
	public String getDateTimeFormat() {
		return "yyyy-MM-dd'T'HH:mm:ss";
	}

	@Override
	public String toString() {
		return "ScheduleSettings [backLink=" + backLink + ", timeUnit="
				+ timeUnit + ", numberOfTimeUnits=" + numberOfTimeUnits
				+ ", startingDate=" + startingDate + ", periodic=" + periodic
				+ ", newTimeUnit=" + newTimeUnit + ", newNumberOfTimeUnits="
				+ newNumberOfTimeUnits + ", newStartingDate=" + newStartingDate
				+ ", newPeriodic=" + newPeriodic + ", hashCode()=" + hashCode()
				+ "]";
	}

}
