package net.pricefx.formulaengine.scripting;

import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import net.pricefx.common.api.InputType;
import net.pricefx.common.dto.creationworkflow.CreationWorkflowDefinition;
import net.pricefx.common.dto.creationworkflow.CreationWorkflowStepDefinition;
import net.pricefx.domain.CalculatedFieldSet;
import net.pricefx.domain.Customer;
import net.pricefx.domain.CustomerAssignment;
import net.pricefx.domain.CustomerGroup;
import net.pricefx.domain.LookupTable;
import net.pricefx.domain.ManualPricelist;
import net.pricefx.domain.PriceGrid;
import net.pricefx.domain.Pricelist;
import net.pricefx.domain.ProductGroup;
import net.pricefx.domain.RebateRecordSet;
import net.pricefx.domain.Simulation;
import net.pricefx.formulaengine.AbstractProducer.ResultIterator;
import net.pricefx.formulaengine.DatamartContext;
import net.pricefx.formulaengine.DatamartContext.DataSlice;
import net.pricefx.formulaengine.DatamartRowSet;
import net.pricefx.formulaengine.TableContext;
import net.pricefx.formulaengine.XExpression;
import net.pricefx.formulaengine.scripting.inputbuilder.InputBuilderFactory;
import net.pricefx.server.dto.SummaryCalculableObjectItem;
import net.pricefx.server.dto.calculation.AttributedResult;
import net.pricefx.server.dto.calculation.ConfiguratorEntry;
import net.pricefx.server.dto.calculation.ConfiguratorEntryArray;
import net.pricefx.server.dto.calculation.ContextParameter;
import net.pricefx.server.dto.calculation.DashboardController;
import net.pricefx.server.dto.calculation.ResultFlexChart;
import net.pricefx.server.dto.calculation.ResultGauge;
import net.pricefx.server.dto.calculation.ResultHighchart;
import net.pricefx.server.dto.calculation.ResultHighmap;
import net.pricefx.server.dto.calculation.ResultMatrix;
import net.pricefx.server.dto.calculation.WrappedResult;
import net.pricefx.service.ItemSummaryQuery;
import net.pricefx.workflowengine.WorkflowInfo;

import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;

import com.googlecode.genericdao.search.Filter;

import groovy.lang.Closure;

/**
 * This interface represents what can be reached in the Groovy element syntax as
 * {@code api}. That {@code api} variable or binding is merely an instance of
 * this interface to expose certain "safe" methods and functionality.<br>
 * <br>
 * <b>Caching usage</b><br>
 * The API exposes three ways to store transient and temporary data during formula
 * executions. Each variant has its own specifics and therefore is not suitable for all use cases. The options are:<br>
 * <ul>
 * <li>api.local</li>
 * <li>api.global</li>
 * <li>Shared Cache</li>
 * </ul>
 * <br>
 * The first two can be understood as mere in-memory hash maps. For example,
 * api.local.put("key","value") will work fine.<br>
 * The difference between them is the potential scope and retention of the
 * contained data.<br>
 * <br>
 * <u>api.local:</u><br>
 * This is a hash map that is available during the lifetime of exactly one
 * logic execution. It can (and should) be used to store intermediate results
 * that need to be accessed by other subsequent logic elements.<br>
 * <br>
 * <u>api.global:</u><br>
 * By default, api.global is very similar to api.local. Without any further
 * settings, the only difference is that api.global is also available in Groovy
 * library functions (while api.local is not).<br>
 * However, by setting api.retainGlobal = true you can instruct the formula engine to keep the hash map in between logic runs. It works between line items and, in case of Quotes, Contracts and Rebate Agreements, between the header and the line items (but only in the pre-phase, not in the post-phase). The values are also carried forward between two list calculation passes (i.e. form initial run to dirty item run; ONLY in non-distributed mode.).<b><br>
 * Important notes:
 * <br>retainGlobal can be set to true either in the logic or globally in Configuration > 
 * All Modules > General Settings > api.retainGlobal defaults to TRUE.
 * retainGlobal is effective only once. In order to keep it for
 * the 3rd item as well (and so on) you need to call it in every formula
 * execution.<br>
 * api.global data also stick to their execution thread. That means that they are NOT shared between
 * the different parallel executions of a distributed calculation. It effectively builds up multiple
 * caches. Another side effect of this is that api.global is not retained and passed to a subsequent 
 * calculation pass if the original calculation was in a distributed mode.
 * <br></b>
 * <br>
 * <u>Shared Cache:</u><br>
 * The main limitation of api.global is the restriction of its availability to
 * the current execution thread. In a distributed calculation however, there are
 * multiple threads and even multiple servers involved in the calculation. If all
 * these should share a single cache, you must use the shared cache. The shared
 * cache is even available to different calculations (such as other lists or list &amp;
 * CFS or dashboard logic etc). Be careful with key naming so that you do not
 * mix and match data.<br>
 * These advantages come at a cost: Due to the distributed "superglobal" nature,
 * the read/write operations of that cache are slower compared to a real
 * local in-memory map. Use the shared cache only for data that is expensive
 * to calculate or for avoiding frequent database calls. Data in the shared cache
 * expires automatically, the TTL is extended everytime a key is written or read.
 * Values stored are of the String type only. Serialization and deserialization of other
 * objects like dates and numbers are to be done in logic itself.
 * <br><br>
 * <b>Object modifications</b><br>
 * The API contains methods to manipulate data (add, addOrUpdate, delete, etc).
 * These methods can only be used in certain formula contexts (otherwise they are ignored).
 * These contexts are: <br>
 * <ul>
 * <li>From within a CFS execution (non-distributed only)</li>
 * <li>From within a calculation flow</li>
 * <li>From within a direct logic execution via JSON API </li>
 * </ul>
 * <br>
 * On top of this, some restrictions apply which data objects can be manipulated - i.e. not every possible object can be modified.
 * Currently available whitelisted typecodes are:
 * <ul>
 * <li>All customer extension types</li>
 * <li>All product extension types</li>
 * <li>All matrix price parameter values and normal price parameter values (LTV) as well as price parameter header (LT)</li>
 * <li>PGI and XPGI</li>
 * <li>PLI and XPLI</li>
 * <li>MPLI</li>
 * <li>P</li>
 * <li>C</li>
 * <li>PCOMP</li>
 * <li>PDESC</li>
 * <li>PGI</li>
 * <li>TODO</li>
 * <li>Data export, import and archive</li>
 * </ul>
 */
public interface PublicGroovyAPI {

	/**
	 * Returns true if the formula is executed in the syntax check mode.
	 * The syntax check mode is a special type of run used to determine the logic parameters in the configuration dialogs
	 * where the logics are selected (e.g. in configuration of LPGs, PLs or dashboards).
	 * The syntax check mode runs against no calculable item, so all the elements that depend on item data
	 * (e.g. {@link #product()} or {@link #currentItem()} as well as {@link #getElement(String)}) produce null or mocked values.
	 * <p>
	 * Note: The best practise is to gather all parameters in the first elements of the logic and then abort the calculation
	 * after all parameters are gathered, otherwise all code which depends on the item data will need a special treatment
	 * (it would have to be surrounded using this function).
	 * <p>
	 * Example:
	 * <p>
	 * <b>Element Amount:</b>
	 * <blockquote>
	 * <pre>
	 * return api.decimalUserEntry("Amount")
	 * </pre>
	 * </blockquote>
	 * <b>Element Comment:</b>
	 * <blockquote>
	 * <pre>
	 * return api.stringUserEntry("Comment")
	 * </pre>
	 * </blockquote>
	 * <b>Element SyntaxCheckAbort:</b>
	 * <blockquote>
	 * <pre>
	 * if (api.isSyntaxCheck()) {
	 *   api.abortCalculation()
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @return true if the logic is executed in the syntax check mode
	 */
	boolean isSyntaxCheck();

	/**
	 * Returns true if the logic is called via the UI debug/test execution mode.
	 * 
	 * @return true if executed in the debug mode
	 */
	boolean isDebugMode();

	/**
	 * Returns true if the calculation is executed in the distributed mode.
	 * Useful to suppress warnings (or not execute methods) for operations that do not work in the distributed mode.
	 * These operations are basically all write operations that affect other items.
	 * Examples: all object modification ops (add, addOrUpdate, update, delete), marking other items dirty
	 * 
	 * @return true if executed in the distributed mode
	 */
	boolean isDistributedMode();
	
	/**
	 * A publicly available cache (as a hashmap) for storing and sharing values across logic elements.
	 * If {@code api.retainGlobal} is set to true, this map will also be available with the entire list calculation,
	 * i.e across item boundaries. It can be also accessed directly as {@code api.global}.
	 *
	 * <strong>Important</strong>: This map must be treated as a cache.
	 * It means that the logic should never rely on the cached values, as in various situations this cache is purged
	 * or structured differently. For example, in a distributed calculation, every calculation thread has its own cache;
	 * or in a multi-pass list calculation, the subsequent passes may not have the cache of the previous pass available.
	 * The system always tries to make it available, but the logic should never rely on it.
	 *
	 * @return the hashmap
	 */
	Map<String, Object> getGlobal();

	/**
	 * Puts an entire map of variables into the global cache.
	 *
	 * @param globalVars The map of variables to put into the global cache
	 */
	void putGlobal(final Map<String, Object> globalVars);

	/**
	 * Retrieves the CalculationContext of the logic evaluation (logic element group).<br>
	 * <ol type="a">
	 * <li>For a PA Calculation or Flush logic this can be (from PriceFxInterface):
	 * <ul>
	 *      <li>FORMULAELEMENTGROUP_DM_INIT = "init" - executed before rows</li>
	 *      <li>FORMULAELEMENTGROUP_DM_ROW = "row" - executed for every line of the Target/Source</li>
	 *      <li>FORMULAELEMENTGROUP_DM_SUMMARY = "summary" - executed after rows</li>
     * </ul>
     * </li>
     * <li>In RebateAgreement calculation there is:
	 * <ul>
     *      <li>FORMULAELEMENTGROUP_RM_AGREEMENT = "agreement"</li>
     *      <li>FORMULAELEMENTGROUP_RM_AGREEMENT_READONLY = "agreementReadOnly"</li>
     *      <li>FORMULAELEMENTGROUP_RM_REBATERECORD = "rebateRecord"</li>
     * </ul>
     * </li>
     * <li>In the Optimizer module, when evaluating Model formulas, there is:
     * <ul>
     *      <li>FORMULAELEMENTGROUP_MODEL_DEFINITION = "definition"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_CALCULATION = "calculation"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_ANALYSIS = "analysis"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_FINALIZATION = "finalization"</li>
     *      <li>FORMULAELEMENTGROUP_MODEL_RECORD = "record"</li>
     * </ul>
     * </li>
     * </ol>
     * @return CalculationContext as String
	 */
    String getCalculationContext();

	/**
	 * Gets the info about the current batch of items being calculated by a calcuation thread.
	 * The calculation batch size is defined by a server variable calculationTasks.commitBatchSize and by default it is 200 records.
	 * For the best use, use the functions from the <a href="https://bitbucket.org/pricefx/pricefx-logic/src/master/CalculationLogic/SharedLib/elements/BatchUtils.groovy">Shared Library</a>.
	 *
	 * @return List of arrays [sku, key2] for PLI, XPLI, PGI, XPGI contexts or [itemId, typedId] for CFS logics run on P, PX
	 * @see <a href="https://bitbucket.org/pricefx/pricefx-logic/src/master/CalculationLogic/SharedLib/elements/BatchUtils.groovy">SharedLib/elements/BatchUtils.groovy</a>
	 */
	public List<String[]> getBatchInfo();

	/**
	 * Returns the ID part of an object. The method tries to be smart and extracts also IDs from maps
	 * or typedId strings.
	 * @param obj The given object (Map)
	 * @return The ID of the object
	 */
	Long getId(Object obj);

	/**
	 * Gets the result of the previous logic element.
	 * <p>
	 * <b>Note:</b> Instead of this function, you can use out.<ElementName> or output.<ElementName>.
	 *
	 * @param name The name of the logic element
	 * @return The value or {@code null} if the element is not found
	 */
	Object getElement(String name);

	/**
	 * Same as {@code api.currentItem("key2")}. Retrieves the secondary key (field "key2") of a matrix item record. Applicable in the following contexts:
     * <ul>
     *      <li>matrix price grid (XPGI.key2)</li>
     *      <li>matrix price list (XPLI.key2)</li>
     *      <li>matrix simulation (XSIMI.key2)</li>
     * </ul>
	 *
	 * @return The value of the secondary key or {@code null} if calculating a single item record (= not a matrix item record from the list above)
	 */
	Object getSecondaryKey();

	/**
	 * Gets a named binding variable from a library function.
	 * Example:
	 * <p>
	 * In a workflow formula, the workflow can be referenced in a Groovy element directly by the binding/variable "workflow":
	 * {@code workflow.addApprovalStep("First Approval")}
	 * <p>
	 * However, this does not work from a library function as this script does not have the binding available.
	 * You need to use:
	 * <blockquote>
	 * <pre>
	 * api.getBinding("workflow").addApprovalStep("First Approval")
	 * </pre>
	 * </blockquote>
	 *
	 * @param name The name of the variable
	 * @return The binding
	 */
	Object getBinding(String name);

	/**
	 * Retrieves the mandatory parameter targetDate from the current execution context. The target date is {@code new Date()} by default or overriden in:
     * <ul>
     *     <li>Quotes: Quote view</li>
	 *     <li>Price grids: Price grids overview window, Target date column</li>
     *     <li>Price lists: Price list creation dialog, tab Get parameters</li>
	 *     <li>PA calculation dataloads: Dataload definition, Calculation tab</li>
	 *     <li>Calculation flow: Calculation flow detail, Target date column</li>
     * </ul>
	 *
	 * @return The targetDate
	 */
	Date targetDate();

	/**
	 * Gets a {@code java.util.Calendar} object initialized with the current targetDate (or new Date() if no target date is given).
	 * Used for adding days/months/year and/or setting hours, minutes, seconds.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def months = 6
	 *
	 * def calendar = api.calendar()
	 *
	 * // adds months to the date
	 * calendar.add(Calendar.MONTH, months)
	 * calendar.set(Calendar.HOUR_OF_DAY, startHour)
	 * calendar.set(Calendar.MINUTE, 0)
	 * calendar.set(Calendar.SECOND, 0)
	 *
	 * // return the date 6 months later after targetDate at 00:00:00
	 * calendar.getTime()
	 * </pre>
	 * </blockquote>
	 * @return The initialized Calendar object
	 */
	Calendar calendar();

	/**
	 * Gets a calendar object initialized with {@code date}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def months = 6
	 * def targetDate = api.targetDate()
	 *
	 * def calendar = api.calendar(targetDate)
	 *
	 * // adds months to the date
	 * calendar.add(Calendar.MONTH, months)
	 * calendar.set(Calendar.HOUR_OF_DAY, startHour)
	 * calendar.set(Calendar.MINUTE, 0)
	 * calendar.set(Calendar.SECOND, 0)
	 *
	 * // return the date 6 months later after the given date at 00:00:00
	 * calendar.getTime()
	 * </pre>
	 * </blockquote>
	 *
	 * @param date The date to initialize the calendar with
	 * @return The initialized Calendar object
	 */
	Calendar calendar(Date date);

	/**
	 * Provides the access to the standard {@code java.util.Random} generator instance created for the duration of the calculation task
	 * (= not get recreated for each logic element evaluation).
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def randomPct = (90 + api.random().nextInt(20)) / 100
	 * // randomPct contains the random Integer number between 90 (inclusive) and 110 (exclusive)
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def random = api.random().nextLong()
	 * // random contains the random Long number e.g. -2249116201139970752
	 * </pre>
	 * </blockquote>
	 * @return The {@code java.util.Random} generator instance created for the duration of the calculation task
	 * (i.e. does not get re-created for each formula (element) evaluation).
	 * @see <a href="https://docs.oracle.com/javase/8/docs/api/java/util/Random.html">https://docs.oracle.com/javase/8/docs/api/java/util/Random.html</a>
	 */
	Random random();

	/**
	 * @return A {@link net.pricefx.domain.DMCalendar DatamartCalendar} object.
	 * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#calendar} instead. 
	 */
	Object datamartCalendar();

	/**
	 * Creates a new DatamartSlice, which allows setting filter criteria along the Time, CustomerGroup, ProductGroup
	 * or other dimensions in a Datamart.
	 *
	 * @return The instantiated, empty, {@link DataSlice} object
	 * @deprecated Use {@link DatamartContext#newDatamartSlice()} instead.
	 */
	@Deprecated 
	Object newDatamartSlice();

	/**
	 * Creates a new DatamartSlice, which allows setting filter criteria along the Time, CustomerGroup, ProductGroup
	 * or other dimensions in a Datamart, initialized with the name of the time dimension field and an optional set
	 * filter criteria.
	 *
	 * @param dateFieldName Name of the time dimension field
	 * @param timePeriodsAndProductAndCustomerGroups TimePeriod, CustomerGroup, ProductGroup filters
	 * @return The initialized {@link DataSlice} object
	 * @deprecated Use {@link DatamartContext#newDatamartSlice(String, Object...)} instead.
	 */
	Object newDatamartSlice(String dateFieldName, Object... timePeriodsAndProductAndCustomerGroups);

	/**
	 * Translates a CustomerGroup/ProductGroup/DMDataSlice to its equivalent filter representation,
	 * mapping the domain level field names to the corresponding Datamart field names in the process.
	 * <p>
	 * Note: A DMDataSlice object can be instantiated by api.newDatamartSlice, or in a Rebate calculation context, referenced by the 'calulationBase' binding.
	 * <blockquote>
	 * <pre>
	 * def filter = api.customerGroupEntry("Customer Filter")
	 * return filter != null ? api.datamartFilter(filter) : Filter.custom("1=1")
	 * </pre>
	 * </blockquote>
	 * @param groupOrSlice The CustomerGroup/ProductGroup/DMDataSlice to translate into Datamart level filter criteria
	 * @return Filter object representing either the CustomerGroup/ProductGroup dimension slice or the passed slice in the Datamart
	 */
	Filter datamartFilter(Object groupOrSlice);

	/**
	 * This method is deprecated and exists only for backward compatibility for older in-production formulas.
	 * Use the Groovy query API instead (see {@link net.pricefx.formulaengine.DatamartContext DatamartContext}).
	 *
	 * Runs a query against a PA Datamart. The query is only executed once in the scope of a calculation job (PL generation etc.).
	 * The query is identified by the first argument. This is also the 'handle' by which datamartLookup refers to the result set of the query.
	 * 
	 * @param queryAndDatamartAndProjectionsAndFilters  The first argument is the query ID/handle, subsequent arguments are either DM field names (projections)
	 * or filter objects.
	 * @return The numbers of rows in the result set. The actual result set data is stored in memory for a subsequent lookup by datamartLookup(...).
	 * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#newQuery(Table)}
	 */
	Object datamartQuery(Object... queryAndDatamartAndProjectionsAndFilters);

	/**
	 * This method is deprecated and only exists for backward compatibility for older in-production formulas.
	 * Use the Groovy query API instead (see {@link net.pricefx.formulaengine.DatamartContext DatamartContext}).
	 *
	 * Looks up rows in the result of a datamartQuery(...) call.  
	 * @param queryAndProjectionsAndFilters  The first arguments is the ID/handle from a datamartQuery(...) call.  Subsequent arguments refer to
	 * projections specified in the datamartQuery call or filter objects leading to additional filtering of rows in the result set.
	 * @return A single primitive value if the result of the lookup results in one row for just one projection.
	 * A list of values if the result is multiple rows for one projection.
	 * A map of (field name, value) pairs for a single row result with multiple projections.
	 * A Matrix2D object for multiple rows and projections.
	 * @deprecated Use {@link net.pricefx.formulaengine.DatamartContext#executeQuery(DatamartContext.Query)},
	 * {@link net.pricefx.formulaengine.DatamartContext#executeSqlQuery(String, DatamartContext.Query...)} or
	 * {@link net.pricefx.formulaengine.DatamartContext#streamQuery(DatamartContext.Query)}
	 */
	Object datamartLookup(Object... queryAndProjectionsAndFilters);


	/**
	 * Provides an API for querying and loading PA data.
	 * The most common cases are:
	 * <ul>
	 *     <li>getting the PA Datamart using {@link DatamartContext#getDatamart(String)} to be further queried</li>
	 *     <li>getting the PA Data Source using {@link DatamartContext#getDataSource(String)} to be further queried</li>
	 *     <li>getting the PA calendar using {@link DatamartContext#calendar()}</li>
	 * </ul>
	 *
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def ctx = api.getDatamartContext()
	 *
	 * def dm = ctx.getDatamart("Transactions_DM")
	 *
	 * def query = ctx.newQuery(dm, true)
	 *     .select("ProductGroup")
	 *     .select("SUM(SalesValue)")
	 *     .where(
	 *         Filter.equal("Year", "2019"),
	 *     )
	 *     .groupBy("ProductGroup")
	 *     .orderBy("ProductGroup")
	 *
	 * def result = ctx.executeQuery(query)
	 *
	 * return result?.getData()
	 * </pre>
	 * </blockquote>
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def date = api.parseDate("yyyy-MM-dd", "2018-01-24")
	 * def week = api.getDatamartContext().calendar().getWeek(date)
	 * return week
	 * </pre>
	 * </blockquote>
	 *
	 * @return {@link net.pricefx.formulaengine.DatamartContext DatamartContext} 
	 */
	DatamartContext getDatamartContext();

	/**
	 * Provides an SQL query (SELECT only) interface to PA query results.
	 * @return {@link net.pricefx.formulaengine.TableContext TableContext}
	 * @throws Exception if a DB connection cannot be established
	 */
	TableContext getTableContext() throws Exception;
	
	/**
	 * Available in a PA DataLoad context only, a {@link DatamartRowSet} provides access to the rows being loaded or calculated.
	 * @return The {@link DatamartRowSet} holding the data rows to load in the target FC.
	 */
	DatamartRowSet getDatamartRowSet();

	/**
	 * Available in a PA DataLoad context only, a {@link DatamartRowSet} provides access to the rows currently being loaded or calculated.
	 * The following Data Load types make use of rowsets:
	 * <ul>
	 * 		<li>'Flush': The source rowset gives access to the DataFeed data rows, while the target rowset holds the rows to be loaded in the target DataSource (DS).</li>
	 * 		<li>'Calculate': The source rowset gives access to the source rows, which can either come from a DF or from the target FC being calculated.
	 * 						The target rowset holds the rows to be loaded in the target FC. Initially, this rowset will be empty unless the DL.withTargetSnapshot
	 * 						option is set, in which case the rowset will be pre-populated with the target FC rows in the scope of the DL (determined by the DL.filter).</li>
	 * </ul>
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def newRow = [
	 *      "sku" : "1234456",
	 *      "cost" : 123
	 * ]
	 *
	 * def target = api.getDatamartRowSet("target")
	 *
	 * // target is null when running the logic test execute, so you can use trace for debugging purposes
	 * if (target) {
	 *     target.addRow(newRow)
	 * } else {
	 *     api.trace("newRow", null, newRow)
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def source = api.getDatamartRowSet("source")
	 * def target = api.getDatamartRowSet("target")
	 *
	 * while (source?.next()) {
	 *     def row = source?.getCurrentRow()
	 *
	 *     // perform the required calculation of row
	 *     target["TotalSales"] = ...
	 *
	 *     // target is null when running the logic test execute, so you can use trace for debugging purposes
	 *     if (target) {
	 *         target.addRow(row)
	 *     }
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param name Either 'source' to select the source FC or 'target' to select the target FC rowset.
	 * @return The {@link DatamartRowSet} holding either the source or target rowset.
	 * @see #getDatamartContext()
	 * @see DatamartContext#newQuery(Table)
	 */
	DatamartRowSet getDatamartRowSet(String name);

	/**
	 * Temporarily switches the current SKU context to another SKU.
	 * This can be used to call functions that rely on a SKU context and do not allow to specify an alternate SKU with a different SKU.
	 * <ul>
	 *     <li>{@link #product()}</li>
	 *     <li>{@link #productCompetition()}</li>
	 *     <li>{@link #productExtension(String)}</li>
	 *     <li>{@link #productXRef()}</li>
	 *     <li>{@link #bom()}</li>
	 *     <li>{@link #bomList()}</li>
	 *     <li>{@link #pricelist(String)}</li>
	 *     <li>{@link #getPricelistItem(String)}</li>
	 * </ul>
	 * The switch is reset by {@link #resetSKUContextSwitch} or automatically at the beginning of every element.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def masterSku = api.productXRef()
	 * def costs
	 *
	 * if (masterSku) {
	 *     api.switchSKUContext(masterSku[0].targetSku)
	 *
	 *     // take the cost from the masterSku item
	 *     costs = api.productExtension("Cost")
	 *
	 *     api.resetSKUContextSwitch()
	 * } else {
	 *     // take the cost from the item being calculated
	 *     costs = api.productExtension("Cost")
	 * }
	 *
	 * if (costs) {
	 *     ....
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param newSKU The new SKU to switch to.
	 * @see #resetSKUContextSwitch()
	 * @see #product(String, String)
	 */
	void switchSKUContext(String newSKU);

	/**
	 * Resets a SKU context switch to the originally specified SKU.
	 * @see #switchSKUContext(String)
	 */
	void resetSKUContextSwitch();

	/**
	 * Retrieves the product object for the SKU currently being calculated.
	 * The product object is cached and the next call reads the value from cached record.
	 * <p>
     * If this method is called with the parameter value “sku”, this specific call is optimized
     * in the backend, so it does NOT make a lookup to the Product Master when you only need the “sku” value.
     * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def item = api.product()
	 * def message = checkItem(item)
	 *
	 * def checkItem(item) {
	 *     if (!item.attribute2) {          // Product group
	 *         return "Product group not set"
	 *     } else if (!item.attribute5) {   // Manufacturer
	 *         return "Manufacturer not set"
	 *     }
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @return The full current product Map
	 * @see #currentItem()
	 */
	Object product();
	
	/**
	 * Retrieves the given attribute value from the product table for the SKU being calculated.
	 * The product object is cached and the next call reads the value from the cached record.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def label = api.product("label")
	 * def productGroup = api.product("attribute1") // Product Group
	 * def manufacturer = api.product("attribute4") // Manufacturer
	 * </pre>
	 * </blockquote>
	 * @param attributeName The attribute value to be returned
	 * @return The value of the attribute
	 * @see #currentItem(String)
	 */
	Object product(String attributeName);

	/**
	 * Retrieves the given attribute value from the product table for a given SKU.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def masterSku = ...
	 * def label
	 *
	 * if (masterSku) {
	 *     // take the label from the masterSku item
	 *     label = api.product("label", masterSku)
	 * } else {
	 *    // take label from the item being calculated
	 *    label = api.product("label")
	 * }
	 * </pre>
	 * </blockquote>
	 * @param attributeName The attribute whose value should be returned
	 * @param sku           The product
	 * @return The value of the attribute (or {@code null} if product not found)
	 */
	Object product(String attributeName, String sku);

	/**
	 * Retrieves the given attribute value from the product table for a given SKU.
	 * If the given SKU is null, filterFormulaName and filterFormulaParam are used to apply
	 * filter logic for the Product picker.
	 * @param attributeName 		The attribute whose value should be returned
	 * @param sku					The product
	 * @param filterFormulaName  	Unique name of a formula that will be triggered when the Product picker opens
	 * @param filterFormulaParam 	Additional data for filterFormula
	 * @return The value of the attribute (or {@code null} if product not found)
	 */
	Object product(String attributeName, String sku, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Triggers a custom-named product picker input parameter.
	 *
	 * @param parameterName The name of the parameter
	 * @param attributeName The attribute whose value should be returned ({@code null} for the entire object)
	 * @return The attribute value
	 */
	Object otherProduct(String parameterName, String attributeName);

	/**
	 * Triggers a custom-named product picker input parameter.
	 *
	 * @param parameterName The name of the parameter
	 * @param attributeName The attribute whose value should be returned ({@code null} for the entire object)
	 * @param filterFormulaName  	Unique name of a formula that will be triggered when the Product picker opens
	 * @param filterFormulaParam 	Additional data for filterFormula
	 * @return The attribute value
	 */
	Object otherProduct(String parameterName, String attributeName, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Triggers a custom-named product picker input parameter.
	 *
	 * @param parameterName The name of the parameter
	 * @return The product object
	 */
	Object otherProduct(String parameterName);

	/**
	 * Retrieves a list of all product extentions records (PX) from a PX table named {@code extensionName} for the SKU being calculated.
	 * Consider use of {@link #getBatchInfo()} for the performance reasons.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def promotions = api.productExtension("Promotion")
	 *
	 * if (promotions) {
	 *     def promotion = promotions.getAt(0)
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param extensionName The name of the extension
	 * @return List of PX records (represented as a Map)
	 * @see #getBatchInfo()
	 */
	Object productExtension(String extensionName);

	/**
	 * Retrieves a list of product extentions records (PX) from a PX table named {@code extensionName} for the SKU being calculated filtered by {@code filters}.
	 * Consider use of {@link #getBatchInfo()} for the performance reasons.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def filter = [
	 *         Filter.greaterOrEqual("attribute1", targetDate),   // Valid from
	 *         Filter.lessOrEqual("attribute2", targetDate)       // Valid to
	 * ]
	 *
	 * def promotions = api.productExtension("Promotion", *filter)
	 *
	 * if (promotions) {
	 *     def promotion = promotions.getAt(0)
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param extensionName The name of the extension
	 * @param filters       One or more {@link Filter} objects that narrow down the search
	 * @return List of matching PX records (represented as a Map)
	 * @see #getBatchInfo()
	 */
	Object productExtension(String extensionName, Filter... filters);

	/**
	 * Retrieves a list of all product competition records (PCOMP) for the SKU being calculated.
	 * Consider using {@link #productExtension(String)} instead since that gives you a better flexibility.
	 *
	 * @return List of PCOMP records (represented as a Map)
	 * @see #getBatchInfo()
	 */
	Object productCompetition();

	/**
	 * Retrieves a list of product competition records (PCOMP) for the SKU being calculated filtered by {@code filters}.
	 * Consider using {@link #productExtension(String)} instead since that gives you a better flexibility.
	 *
	 * @param filters One or more {@link Filter} objects that narrow down the search
	 * @return List of matching PX records (represented as a Map)
	 */
	Object productCompetition(Filter... filters);

	/**
	 * Retrieves a list of all product reference records (PXREF) for the SKU being calculated.
	 * Consider use of {@link #getBatchInfo()} for the performance reasons.
	 *
	 * @return List of PXREF records (represented as a Map)
	 * @see #getBatchInfo()
	 */
	Object productXRef();

	/**
	 * Prompts the user to input parameters for a rebate agreement (RBA).
	 *
	 * @param parameterName - The name of an entry in the input parameters listing (mandatory)
	 * @param attributeName - The name of the attribute you need (mandatory)
     * @return The attribute of the rebate agreement or {@code null} if RA does not exists
     */
	Object rebateAgreementUserEntry(String parameterName, String attributeName);

	/**
	 * Retrieves a list of all product reference records (PXREF) for the SKU being calculated.
	 *
	 * @param filters One or more {@link Filter} objects that narrow down the PXREF search
	 * @return List of matching PXREF records (represented as a Map)
	 * @see #getBatchInfo()
	 */
	Object productXRef(Filter... filters);

	/**
	 * Retrieves the customer object for the SKU currently being calculated.
	 * The customer object is cached and the next call reads the value from cached record.
	 *
	 * @param attributeName The attribute value to be returned
	 * @return The value of the attribute
	 */
	Object customer(String attributeName);

	/**
	 * Equivalent of the standard syntax "Customer" function. Retrieves the given attribute value of the customer context parameter.
	 * If the value is {@code null} and defaultFromParent is {@code true}, it retrieves (and returns) the value from
	 * the customer's parent regardless if that value is {@code null} as well or not. If there is no parent customer set,
	 * {@code null} will be returned.
	 *
	 * @param attributeName     The attribute value to be returned
	 * @param defaultFromParent Determines whether to look up the value from the parent customer (if any) in case it is {@code null} in the customer
	 * @return The value of the attribute
	 */
	Object customer(String attributeName, boolean defaultFromParent);

	/**
	 * Retrieves the given attribute value from the customer table for the SKU being calculated.
	 * The customer object is cached and the next call reads the value from the cached record.
	 *
	 * @param attributeName The attribute value to be returned
	 * @param customerId    The customerId value for the customer search
	 * @return The value of the attribute (or {@code null} if no customer found)
	 */
	Object customer(String attributeName, String customerId);

	/**
	 * Retrieves the given attribute value of the given customer. If the value is
	 * {@code null} and defaultFromParent is {@code true}, it retrieves (and returns) the value from the customer's
	 * parent regardless if that value is {@code null} as well or not. If there is no parent customer set, {@code null} will be returned.
	 *
	 * @param attributeName     The attribute value to be returned
	 * @param customerId        The customerId value for the customer search
	 * @param defaultFromParent Determines whether to look up the value from the parent customer (if any) in case it is {@code null} in the customer
	 * @return The value of the attribute (or {@code null} if no customer found)
	 */
	Object customer(String attributeName, String customerId, boolean defaultFromParent);

	/**
	 * Retrieves a named attribute value of the named customer (if the value is
	 * {@code null} and defaultFromParent is {@code true}, retrieves (and returns) the value from the customer's
	 * parent regardless if that value is {@code null} as well or not. If there is no parent customer set, {@code null} will be returned)
	 *
	 * @param attributeName      The attribute value to be returned
	 * @param customerId         The customerId value for the customer search
	 * @param defaultFromParent  Determines whether to look up the value from the parent customer (if any) in case it is {@code null} in the customer
	 * @param filterFormulaName  Unique name of a formula that will be triggered when the Customer picker opens
	 * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...)
	 * @return The value of the attribute (or {@code null} if no customer found)
	 */
	Object customer(String attributeName, String customerId, boolean defaultFromParent, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName The name of the parameter
	 * @return The customer object
	 */
	Object otherCustomer(String parameterName);

	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName The name of the parameter
	 * @param attributeName The attribute which value to return ({@code null} for the entire object)
	 * @return The attribute value
	 */
	Object otherCustomer(String parameterName, String attributeName);

	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName     The name of the parameter
	 * @param attributeName     The attribute whose value should be returned ({@code null} for the entire object)
	 * @param defaultFromParent Determines whether to look up the attribute value from the parent customer (if any) in case it is {@code null} in the customer
	 * @return The attribute value
	 */
	Object otherCustomer(String parameterName, String attributeName, boolean defaultFromParent);


	/**
	 * Triggers a custom-named customer picker input parameter.
	 *
	 * @param parameterName     The name of the parameter
	 * @param attributeName     The attribute whose value should be returned ({@code null} for the entire object)
	 * @param defaultFromParent Determines whether to look up the attribute value from the parent customer (if any) in case it is {@code null} in the customer
	 * @param filterFormulaName  Unique name of a formula that will be triggered when the Customer picker opens
	 * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...)
	 * @return The attribute value
	 */
	Object otherCustomer(String parameterName, String attributeName, boolean defaultFromParent, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Retrieves a list of all customer extentions records (CX) from a CX table named {@code extensionName}
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def filter = Filter.equal("attribute1", "ShipTo")   // Type
	 *
	 * def addresses = api.customerExtension("Addresses", filter)
	 *
	 * if (addresses) {
	 *     def address = addresses.getAt(0)
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param extensionName The name of the extension
	 * @param filters       None, one or more {@link Filter} object that narrow down the CX search
	 * @return List of matching CX records (represented as a Map)
	 */
	Object customerExtension(String extensionName, Filter... filters);

	/**
	 * Returns a line item of the given calculable line item collection (CLIC).
	 * Currently, there are three types of CLICs: Quote, Contract and Rebate Agreement.
	 *
	 * @param clic   The CLIC to which the line item belongs, it may either be the CLIC object as such (its map representation to be exact) or typedId or a uniqueName.
	 * @param lineId The lineId of the line item
	 * @param <T> ContractLineItem (CTLI), QuoteLineItem (QLI) or RebateAgreementLineItem (RBALI)
	 * @return The calculable line item
	 */
	Object getCalculableLineItem(Object clic, String lineId);

	/**
	 * Returns results of the given names (or labels) from a line item object (its map representation).
	 * <p>
	 * Note: The line item object will typically be fetched via {@link #getCalculableLineItem(Object, String)}.
	 *
	 * @param cli The calculable line item
	 * @param resultNamesOrLabels The result names or labels
	 * @return Either a single result or a map(nameOrLabel,result) of results if there were more names (labels) specified
	 */
	Object getCalculableLineItemResult(Object cli, String... resultNamesOrLabels);

	/**
	 * Returns the full object (header and line items) of e.g. a quote, rebate agreement or contract.
	 *
	 * @param typedId The typeId of the object
	 * @return A map specifying the line item collection or {@code null} if an invalid typedId was passed
	 */
	Object getCalculableLineItemCollection(String typedId);

	/**
	 * Creates a revision of CalculableLineItemCollection (Quote, Rebate Agreement, Contract).
	 *
	 * @param typedId The typeId of the object
	 * @return A map specifying the revised line item collection or {@code null} if an invalid typedId was passed
	 */
	Object createNewRevision(String typedId);
	
	/**
	 * Adds a warning to the formula execution warning log. The warnings are displayed:
	 * <ul>
	 *     <li>in a column in the element results list (Quotes, Rebate agreements, Contract agreements)</li>
	 *     <li>in a column in the grid list items (Pricelists, Pricegrids, Datamarts, Policy records)</li>
	 * </ul>
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def cost =
	 *
	 * if (!cost) {
	 *      api.addWarning("Missing cost")
	 * }
	 * </pre>
	 * </blockquote>
	 * 
	 * @param message The warning text message
	 */
	void addWarning(String message);

	/**
	 * Retrieves the previous calculation result by the logic element label (not the element result name).
	 * @param label The logic element label
	 * @return The calculation result of that element
	 */
	Object contextByLabel(String label);

	/**
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the column 'name' matches the value
	 * entered by the user and returns the value from the column 'value'.
	 * @param parameterName The name of the price parameter table
	 * @return The user selected value, in case of matrix lookups the entire object/row
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName);

	/**
	 * Searches for a record in the price parameter table named {@code parameterName}
	 * where the column 'name' (for LTV or MLTV) or 'key1' (for MLTV2-6)
	 * matches the attributeNameOrKeyOrRangeValue value
	 * and returns the value from the column 'value' (LTV) or 'attribute1' (MLTV).
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
     * <p>
     * Example:
	 * <blockquote>
	 * <pre>
	 * def productGroup = api.product("attribute2")
	 * def shippingCost = api.vLookup("ShippingCosts", productGroup)
	 * </pre>
	 * </blockquote>
     *
	 * @param parameterName                  The name of the price parameter table
	 * @param attributeNameOrKeyOrRangeValue A predefined key or range value or in case of a matrix lookup a user entry
	 * @return The value of the column 'value' for SIMPLE (LTV) or 'attribute1' for MATRIX (MLTV)
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, Object attributeNameOrKeyOrRangeValue);

	/**
	 * This syntax is only relevant for MATRIX 1-key price parameter tables (MLTV).
	 * The function searches for a record in the price parameter table named {@code parameterName}
	 * where the column 'key1' matches the {@code key1} parameter
	 * and returns the value of the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from cache.
	 *
	 * <p>
     * Example:
	 * <blockquote>
	 * <pre>
	 * def prevailing = api.vLookup("Manufacturer", "attribute1", "Meat company")
	 * </pre>
	 * </blockquote>
     *
	 * @param parameterName The name of the price parameter table
	 * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
	 * @param key           The key of the parameter table
	 * @return The value of the column {@code attributeName} or full object if attribute name is null
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key);

	/**
	 * This syntax is only relevant for MATRIX 2-key price parameter tables (MLTV2).
	 * The function searches for a record in a price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key2' match the {@code key1}-{@code key2} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def exchangeRate = api.vLookup("ExchangeRate", "attribute1", "EUR", "USD")
	 * </pre>
	 * </blockquote>
	 *
	 * @param parameterName The name of the price parameter table
	 * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
	 * @param key1          The key1 of the parameter table
	 * @param key2          The key2 of the parameter table
	 * @return The value of the column {@code attributeName} or full object if attribute name is null
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2);

	/**
	 * This syntax is only relevant for MATRIX 3-key price parameter tables (MLTV3).
	 * The function searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key3' match the {@code key1}-{@code key3} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def exchangeRate = api.vLookup("ExchangeRate", "attribute1", "EUR", "USD", "2018")
	 * </pre>
	 * </blockquote>
	 *
	 * @param parameterName The name of the price parameter table
	 * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
	 * @param key1          The key1 of the parameter table
	 * @param key2          The key2 of the parameter table
	 * @param key3          The key3 of the parameter table
	 * @return The value of the column {@code attributeName} or full object if attribute name is null
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3);

	/**
	 * This syntax is only relevant for MATRIX 4-key price parameter tables (MLTV4).
	 * The function searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key4' match the {@code key1}-{@code key4} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * @param parameterName The name of the price parameter table
	 * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
	 * @param key1          The key1 of the parameter table
	 * @param key2          The key2 of the parameter table
	 * @param key3          The key3 of the parameter table
	 * @param key4          The key4 of the parameter table
	 * @return The value of the column {@code attributeName} or full object if attribute name is null
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4);

	/**
	 * This syntax is only relevant for MATRIX 5-key price parameter tables (MLTV5).
	 * The function searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key5' match the {@code key1}-{@code key5} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * @param parameterName The name of the price parameter table
	 * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
	 * @param key1          The key1 of the parameter table
	 * @param key2          The key2 of the parameter table
	 * @param key3          The key3 of the parameter table
	 * @param key4          The key4 of the parameter table
	 * @param key5          The key5 of the parameter table
	 * @return The value of the column {@code attributeName} or full object if attribute name is null
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4, String key5);

	/**
	 * This syntax is only relevant for MATRIX 6-key price parameter tables (MLTV6).
	 * The function searches for a record in the price parameter table named {@code parameterName}
	 * where the columns 'key1'-'key6' match the {@code key1}-{@code key6} parameters
	 * and returns the value from the column {@code attributeName}.
	 * Once the record is found, it remains cached during the whole calculation and
	 * the next call for the same arguments reads the value from the cache.
	 *
	 * @param parameterName The name of the price parameter table
	 * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
	 * @param key1          The key1 of the parameter table
	 * @param key2          The key2 of the parameter table
	 * @param key3          The key3 of the parameter table
	 * @param key4          The key4 of the parameter table
	 * @param key5          The key5 of the parameter table
	 * @param key6          The key6 of the parameter table
	 * @return The value of the column {@code attributeName} or full object if attribute name is null
	 * @see #findLookupTableValues(String, Filter...)
	 */
	Object vLookup(String parameterName, String attributeName, String key1, String key2, String key3, String key4, String key5, String key6);

    /**
     * This syntax can be applied to all MATRIX lookup types.
     * Essentially the keys are passed in a map.
     * The key name can be the technical name ("key1", etc) or the meta-renamed name.
     *
     * @param parameterName The name of the price parameter table
     * @param attributeName The name of the attribute to retrieve, or a comma separated list of attributes
     * @param keys A map with the keys of the row to retrieve. Note: All keys need to be present for the table in question!
     * @return A map with value of the columns of {@code attributeNames} as keys or full object if attributeNames is null
     */
    Object vLookup(String parameterName, String attributeName, Map keys);

	/**
	 * This syntax can be applied to all MATRIX lookup types.
	 * Essentially the keys are passed in a map.
	 * The key name can be the technical name ("key1", etc) or the meta-renamed name.
	 *
	 * @param parameterName The name of the price parameter table
	 * @param attributeNames The names of the attribute to retrieve
	 * @param keys A map with the keys of the row to retrieve. Note: All keys need to be present for the table in question!
	 * @return A map with value of the columns of {@code attributeNames} as keys or full object if attributeNames is null
	 */
	Map<String,Object> vLookup(String parameterName, List<String> attributeNames, Map keys);

	/**
	 * Triggers a numeric input element.
     * <ul>
     * <li>When the user entered number is a whole number (integer),
     * the value returned is java.lang.Integer.</li>
     * <li>When the user entered number is a decimal number,
     * the value returned is java.math.BigDecimal.</li>
     * <li>If you always need the result as a BigDecimal, you can cast it,
     * e.g. api.decimalUserEntry("number") as BigDecimal.</li>
     * </ul>
	 *
	 * @param entryName The name of the user entry parameter
	 * @return The user entered value
	 * @see #input(String)
	 * @see #stringUserEntry(String)
	 * @see #textUserEntry(String) (String...)
	 * @see #decimalUserEntry(String)
	 * @see #dateUserEntry(String)
	 * @see #dateRangeUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #booleanUserEntry(String)
	 * @see #option(String, Object...)
	 * @see #options(String, Object...)
	 * @see #filterBuilderUserEntry(String, String)
	 * @see #datamartFilterBuilderUserEntry(String, String, Object...)
	 * @see #otherProduct(String)
	 * @see #otherCustomer(String)
	 * @see #productGroupEntry(String...)
	 * @see #customerGroupEntry(String...)
	 * @see #rebateAgreementUserEntry(String, String)
	 * @see #anyUser(String)
	 * @see #inputMatrix(String, String...)
	 */
	Object userEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a decimal value.
	 * Rendered as a numeric field.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return Decimal value entered by the user as a BigDecimal
	 * @see #integerUserEntry(String)
	 */
	Object decimalUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter an integer value.
	 * Rendered as a numeric field.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return Integer value entered by the user as a BigDecimal
	 * @see #decimalUserEntry(String)
	 */
	Object integerUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a short text value.
	 * Rendered as a single-line text field.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return Text value entered by the user as a String
	 * @see #textUserEntry(String)
	 */
	Object stringUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a short text value.
	 * Rendered as a single-line text field.
	 *
	 * @param entryName The name of the user entry parameter
	 * @param textMask  Optional text mask that is displayed in the UI
	 * @return Text value entered by the user as a String
	 * @see #textUserEntry(String)
	 */
	Object stringUserEntry(String entryName, String textMask);

	/**
	 * Creates an input parameter that lets the user enter a short text value.
	 * Rendered as single-line text field.
	 *
	 * @param entryName The name of the user entry parameter
	 * @param textMask  Optional text mask that is displayed in the UI
	 * @param label		Label of the entry box
	 * @return Text value entered by the user as String
	 * @see #textUserEntry(String, String)
	 */
	Object stringUserEntry(String entryName, String textMask, String label);
	
	/**
	 * Creates an input parameter that lets the user enter a longer text value.
	 * Rendered as a multi-line text box.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return Text value entered by the user as a String
	 * @see #stringUserEntry(String)
	 */
	Object textUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a longer text value.
	 * Rendered as a multi-line text box.
	 *
	 * @param entryName The name of the user entry parameter
	 * @param label		The label of the entry box
	 * @return Text value entered by the user as a String
	 * @see #stringUserEntry(String, String, String)
	 */
	Object textUserEntry(String entryName, String label);
	
	/**
	 * Creates an input parameter that lets the user enter a boolean (yes or no) value.
	 * Rendered as a checkbox.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return The boolean value entered by the user
	 */
	Object booleanUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a date value.
	 * Rendered as a date picker field.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return String value entered by the user in the format yyyy-MM-dd
	 * @see #parseDate(String, String)
	 * @see #timeUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #dateRangeUserEntry(String)
	 */
	Object dateUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user enter a date range (start and end date).
	 * Rendered as two date picker fields.
	 * UNITY UI ONLY!
	 *
	 * @param entryName The name of the user entry parameter
	 * @return String array or list with values entered by the user in the format yyyy-MM-dd
	 * @see #parseDate(String, String)
	 * @see #timeUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #dateUserEntry(String)
	 */
	Object dateRangeUserEntry(String entryName);
	/**
	 * Creates an input parameter that lets the user to enter a time value in format "HH:mm" (e.g. "23:59").
	 * Rendered as a time picker field.
	 *
	 * @see #parseDate(String, String)
	 * @param entryName The name of the user entry parameter
	 * @return String value entered by the user in the format HH:mm
	 * @see #dateUserEntry(String)
	 * @see #dateTimeUserEntry(String)
	 * @see #dateRangeUserEntry(String)
	 */
	Object timeUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user to enter a date time value in format "dd/MM/yyyy HH:mm" (e.g. "23/01/2019 10:30").
	 * <p>
	 * Rendered as a datetime picker field.
	 * <p>
	 * Important: The format implicitly uses the GMT/UTC time zone and so the returned value may
	 * differ from what was entered in on the client (as the client time may be in a different time zone).
	 *
	 * @see #parseDate(String, String)
	 * @param entryName The name of the user entry parameter
	 * @return String value entered by the user in the format "yyyy-MM-dd'T'HH:mm:ss"<br><br>
	 */
	Object dateTimeUserEntry(String entryName);

	/**
	 * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a product group picker widget.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return The entered product group as a Map
	 * <pre>
	 * [
	 *   "productFieldName"  : "attribute3",
	 *   "productFieldLabel" : "Business Unit",
	 *   "productFieldValue" : "MeatBall"
	 * ]
	 * </pre>
	 */
	Object productGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a product group picker widget.
	 * Only PA searchable Product Extensions will be available.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return The entered product group as a Map
	 * <pre>
	 * [
	 *   "productFieldName"  : "attribute3",
	 *   "productFieldLabel" : "Business Unit",
	 *   "productFieldValue" : "MeatBall"
	 * ]
	 * </pre>
	 */
	Object datamartProductGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of products and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a product group picker widget.
	 *
	 * @param entryName The name of the user entry parameter
	 * @param filterFormulaName A unique name of a formula that will be triggered when the customer picker opens
	 * @param filterFormulaParam Additional data for filterFormula
	 * @return The entered product group as a Map
	 * <pre>
	 * [
	 *   "productFieldName"  : "attribute3",
	 *   "productFieldLabel" : "Business Unit",
	 *   "productFieldValue" : "MeatBall"
	 * ]
	 * </pre>
	 * @return
	 */
	Object productGroupEntry(String entryName, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Creates an input parameter that lets the user select a group of customers and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a customer group picker widget.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return The entered customer group as a Map
	 * <pre>
	 * [
	 *   "customerFieldName" : "attribute3",
	 *   "customerFieldLabel" : "Customer Type",
	 *   "customerFieldValue" : "Restaurant"
	 * ]
	 * </pre>
	 * @see #productGroupEntry(String...)
	 */
	Object customerGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of customers and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a customer group picker widget.
	 * Only PA searchable Customer Extensions will be available.
	 *
	 * @param entryName The name of the user entry parameter
	 * @return The entered customer group as a Map
	 * <pre>
	 * [
	 *   "customerFieldName" : "attribute3",
	 *   "customerFieldLabel" : "Customer Type",
	 *   "customerFieldValue" : "Restaurant"
	 * ]
	 * </pre>
	 * @see #productGroupEntry(String...)
	 */
	Object datamartCustomerGroupEntry(String... entryName);

	/**
	 * Creates an input parameter that lets the user select a group of customers and returns the selection as a Map.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a customer group picker widget.
	 *
	 * @param entryName The name of the user entry parameter
	 * @param filterFormulaName  A unique name of a formula that will be triggered when the customer picker opens
	 * @param filterFormulaParam Additional data for filterFormula (e.g. agreement's typedId,...)
	 * @return The entered customer group as a Map
	 * <pre>
	 * [
	 *   "customerFieldName" : "attribute3",
	 *   "customerFieldLabel" : "Customer Type",
	 *   "customerFieldValue" : "Restaurant"
	 * ]
	 * </pre>
	 * @see #productGroupEntry(String...)
	 */
	Object customerGroupEntry(String entryName, String filterFormulaName, Object filterFormulaParam);

	/**
	 * Creates an input parameter that lets the user enter multiple discounts/surcharges/tiers by amount.
	 * This is commonly used in PromotionManager and RebateManager.
	 * It is rendered as a multi tier entry widget. Every line holds an amount and the given discount/surcharge in an ascending or descending order.
	 *
	 * @param entryName  The name of the user entry parameter
	 * @param arguments Two value hints: sort type (can be either "ASC" which is the default or "DESC") and order validation type
	 *               (can be "VALIDATE" or "NO_VALIDATION").
	 * @return The values entered by the user as a TieredValue object
	 */
	Object multiTierEntry(String entryName, String... arguments);

	/**
	 * Creates an input parameter that lets the user select a single value from the predefined options.
	 * It is rendered as a drop down list of possible options.
	 *
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * return api.option("Currency", ["EUR", "USD"])
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def salesOrgs = api.findLookupTableValues("SalesOrg").collect{
	 *   it.name
	 * }
	 *
	 * return api.option("SalesOrg", salesOrgs)
	 * </pre>
	 * </blockquote>
	 *
	 * @param entryName The name of the user entry parameter
	 * @param options   An arbitrary number of options that should be displayed in the drop down list
	 * @return The user entered value
	 * @see #options(String, Object...)
	 */
	Object option(String entryName, Object... options);

    /**
	 * Creates an input parameter that lets the user select a single value from the predefined options.
	 * It is rendered as a drop down list of possible options.
	 * You can add custom (display-only) labels.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def salesOrgMap = api.findLookupTableValues("SalesOrg").collectEntries{
	 *   [(it.name) : it.name]
	 * }
	 *
	 * return api.option("SalesOrg", salesOrgMap.keySet() as List, salesOrgMap)
	 * </pre>
	 * </blockquote>
     *
     * @param entryName The name of the user entry parameter
     * @param options   An arbitrary number of options that should be displayed in the drop down list
     * @param labels    A map [value : label] of labels that should be shown in the UI. If null or empty will behave same as api.option() without a label definition.
     * @return The user entered value
	 * @see #options(String, List, Map)
     */
    Object option(String entryName, List<Object> options, Map<String, Object> labels);	

    /**
     * Creates an input parameter that lets the user select multiple values from the predefined options.
	 * It is rendered as a drop down list of possible options. Each selected value appears at the top.
     *
     * @param entryName The name of the user entry parameter
     * @param options   An arbitrary number of options that should be displayed in the drop down list
     * @return List of user selected values
	 * @see #option(String, Object...)
     */
    Object options(String entryName, Object... options);

    /**
     * Creates an input parameter that lets the user select multiple values from the predefined options.
	 * It is rendered as a drop down list of possible options. Each selected value appears at the top.
     * You can add custom (display-only) labels.
     *
     * @param entryName The name of the user entry parameter
     * @param options   An arbitrary number of options that should be displayed in the drop down list
     * @param labels    A map [value : label] of labels that should be shown in the UI. If null or empty will behave same as {@link #option(String, Object...)} without a label definition.
     * @return List of user selected values
     */
    Object options(String entryName, List<Object> options, Map<String, Object> labels);

	/**
	 * Creates an input parameter that lets the user build a filter for a table of objects of the type defined by {@code typeCode} parameter.
	 * <p>
	 * It is rendered as a filter builder widget allowing to construct a filter (e.g. to define filter for Products, set the {@code typeCode} to {@code P}).
	 * <p>
	 * The filter object can be used in other functions, like {@link #find(String, Filter...)}, {@link #stream(String, String, Filter...)} {@link #count(String, Filter...)} etc.
	 *
	 * @param entryName The name of the user entry parameter
	 * @param typeCode  A type code string of the type of the object for which the filter is set. The supported type codes currently are "P" (Product data) and "C" (Customer data).
	 * @return A filter object defined by the user
	 */
	Object filterBuilderUserEntry(String entryName, String typeCode);

	/**
	 * Creates an input parameter that lets the user build a filter for a given Datamart.
	 * <p>
	 * The filter can be used in {@link net.pricefx.formulaengine.DatamartContext.Query#where(Filter...)}
	 * of the {@link ISandboxAPI#datamartQuery(Object...)}.
	 * <p>
	 * It will render a FilterBuilder widget entry allowing to construct a filter object filtering on the specified Data Source.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def filter1 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM")
	 *
	 * def dimFilterList = ["CustomerId", "ProductId"]
	 * def filter2 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM", dimFilterList)
	 *
	 * def dimFilterMap = ["CustomerId" : "CD-00155", "ProductId" : "MB-0005"]
	 * def filter3 = datamartFilterBuilderUserEntry("ProductGroup", "Transactions_DM", dimFilterMap, Filter.equal("Country", "DE"))
	 * </pre>
	 * </blockquote>
	 * @param entryName    The name of the user entry parameter
	 * @param source       The name of the source for which the filter is set.  The resolution of the source is as follows:
     * <ol>
     *                  <li> any Datamart - by typedId, sourceName, uniqueName or label (uniqueName and label for backwards compatibility) </li>
     *                  <li> any FieldCollection - by typedId or sourceName </li>
     * </ol>
	 * @param args There can be up to 2 arguments: one can be a Map or List and the other a Filter. The order does not matter.
	 *             If the argument is a Map, each pair [fieldName : value] will be used as a filter.
	 * @return A filter object to the given source
	 */
	Object datamartFilterBuilderUserEntry(String entryName, String source, Object... args);

	/**
	 * Looks up the result price (considering a manual override) of an SKU being calculated in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If multiple price lists are found, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName The name of the price list
	 * @return The {@code resultPrice} of the price list record for the SKU
	 */
	Object pricelist(String listName);

	/**
	 * Looks up the attribute defined by {@code attributeName} (considering a manual override) of an SKU being calculated in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If multiple price lists are found, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName      The name of the price list
	 * @param attributeName The name of the attribute to retrieve
	 * @return The value of the {@code attributeName} column of the price list record
	 * @see #pricelistItem(String, String)
	 */
	Object pricelist(String listName, String attributeName);

	/**
	 * Looks up the result price (considering a manual override) of a given SKU in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName      The name of the price list
	 * @param attributeName The name of the attribute to retrieve
	 * @param sku           The SKU to use
	 * @return The value of the {@code attributeName} column of the price list record
	 * @see #pricelistItem(String, String)
	 */
	Object pricelist(String listName, String attributeName, String sku);

	/**
	 * Looks up the attribute defined by {@code attributeName} (considering a manual override) of the given SKU in a matrix price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName      The name of the price list
	 * @param attributeName The name of the attribute to retrieve
	 * @param sku           The SKU to use
	 * @param key2          The key2 to use
	 * @return The value of the {@code attributeName} column of the matrix price list record
	 * @see #pricelistItem(String, String, String)
	 */
	Object pricelist(String listName, String attributeName, String sku, String key2);

	/**
	 * Looks up the price list item record (PLI or XPLI) of an SKU being calculated in a price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName The name of the price list
	 * @return The entire price list item record (PLI or XPLI)
	 * @see #pricelist(String)
	 */
	Object pricelistItem(String listName);

	/**
	 * Looks up the price list item record (PLI) of the given SKU in a price list of a given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName The name of the price list
	 * @param sku      The SKU to use
	 * @return The entire price list item record (PLI)
	 * @see #pricelist(String, String, String)
	 */
	Object pricelistItem(String listName, String sku);

	/**
	 * Looks up the matrix price list item record (XPLI) of the given SKU in a matrix price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName The name of the price list
	 * @param sku      The SKU to use
	 * @param key2     The key2 to use
	 * @return The entire matrix price list item record (XPLI)
	 * @see #pricelist(String, String, String, String)
	 */
	Object pricelistItem(String listName, String sku, String key2);

	/**
	 * Looks up the matrix price list item record (XPLI) of the given SKU in a matrix price list of the given name.
	 * The price list line has to be approved and the targetDate of the price list has to be prior to the calculation targetDate.
	 * If there are multiple price lists of the same name, the one with the targetDate closer to the calculation targetDate is chosen.
	 *
	 * @param listName The name of the price list
	 * @param sku      The SKU to use
	 * @param key2     The key2 to use
	 * @param targetDate The targetDate for teh search
	 * @return The entire matrix price list item record (XPLI)
	 * @see #pricelist(String, String, String, String)
	 */
	Object pricelistItem(String listName, String sku, String key2, Date targetDate);

	/**
	 * Similar to the {@link #pricelistItem(String)} function. However, it optimizes the subsequent processing by providing the results as a map directly.
	 *
	 * @param listName The name of the price list. Target date and SKU matching applies.
	 * @return The price list object
	 * @see #pricelistItem(String)
	 */
	Map<String, Object> getPricelistItem(String listName);

	/**
	 * Accesses and retrieves values from a rolled-up bill of materials for the SKU being calculated.
	 * All quantities are rolled up and summed.
	 *
	 * @return A list with the summed up quantities
	 */
	Object bom();

	/**
	 * Accesses and retrieves values from a rolled-up bill of materials for the SKU being calculated.
	 * The summed up quantities are multiplied with a factor that is retrieved by a lookup by the material name from the provided parameter.
	 * Example: material-cost table
	 *
	 * @param lookupTableName The parameter table name
	 * @return A list with the summed up and multiplied quantities
	 */
	Object bom(String lookupTableName);

	/**
	 * Accesses and retrieves values from a rolled-up bill of materials for the SKU being calculated.
	 * Only BoM records with the specified category are considered.
	 * The summed up quantities are multiplied with a factor that is retrieved by a lookup by the material name from the provided parameter.
	 * Example: material-cost table
	 *
	 * @param lookupTableName The parameter table name
	 * @param categoryName    The category name to filter on
	 * @return A list with the summed up and multiplied quantities
	 */
	Object bom(String lookupTableName, String categoryName);

	/**
	 * Accesses and retrieves the values from a rolled-up bill of materials for the SKU being calculated
	 * Only BoM records with the specified category and the material are considered.
	 * The summed up quantities are multiplied with a factor that is retrieved by a lookup by the material name from the provided parameter.
	 * An example would be a material-cost table.
	 *
	 * @param lookupTableName The parameter table name
	 * @param categoryName    The category name to filter on
	 * @param materialName    The material to filter on
	 * @return A list with the summed up and multiplied quantities
	 */
	Object bom(String lookupTableName, String categoryName, String materialName);

	/**
	 * Rolls up the current SKU's BoM and returns the rolled up BoM records (not just the quantities).
	 *
	 * @return A list of BoM records
	 */
	Object bomList();

	/**
	 * Creates an input parameter that lets the user select a user from the list of all users.
	 * It is rendered as a user picker dialog / select list.
	 *
	 * @param entryName The name of the entry parameter
	 * @return The login name of the selected user or {@code null}
	 */
	Object anyUser(String entryName);
	
	/**
	 * Returns a value from the current user's master data.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * // retrieves the email of the user
	 * def email = api.user("email")
	 * </pre>
	 * </blockquote>
	 *
	 * @param attributeName The attribute to return
	 * @return The user's attribute value
	 */
	Object user(String attributeName);

	/**
	 * Returns the entire master data object of the current user.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * // retrieves the email of the user
	 * def email = api.user()?.email
	 * </pre>
	 * </blockquote>
	 *
	 * @return The user object (U)
	 */
	Object user();

	/**
	 * Retrieves the value of the input parameter of the given name.
	 * It only reads the value, it does not render any input widget.
	 *
	 * @param inputName The parameter name
	 * @return The parameter value
	 * @see #userEntry(String)
	 */
	Object input(String inputName);

	/**
	 * Creates an input parameter that lets the user select a matrix of values.
	 * It renders as a grid-style input widget with the specified columns.
	 * The result will be a list (= the rows) of maps (= the row's columns). The map attribute names are the column names.
	 * Make sure the column names are valid JSON identifiers (spaces are ok but try to avoid special chars).
	 *
	 * @param inputMatrixName The name of the parameter
	 * @param columnNames     The column names/labels
	 * @return A list (=the rows) of maps (=the row's columns)
	 */
	Object inputMatrix(String inputMatrixName, String... columnNames);

    /**
     * Creates an input parameter that lets the user select a matrix of values.
	 * It renders as a grid-style input widget with the specified columns.
     * The result will be a list (=the rows) of maps (=the row's columns). The map attribute names are the column names.
     * Make sure the column names are valid JSON identifiers (spaces are ok but try to avoid special chars).
     *
     * @param inputMatrixName The name of the parameter
     * @param paramConfig     Map of configuration settings with the following options (key, value):
     *                        <ul>
     *                          <li>'canModifyRows':    false/true (the user can add and delete a row)</li>
     *                          <li>'readOnlyColumns':  List of columns (names) to be uneditable </li>
     *                          <li>'noCellRefresh':    false/true (do not update Configurator on each edit)</li>
     *                          <li>'defaultHeight':    number which specifies default table height</li>
     *                          <li>'fixTableHeight':   false/true (preserve default height of table)</li>
     *                          <li>'fitFieldWidths':   false/true (Should fields autofit their widths to titles or content?)</li>
     *                        </ul>
     * @param columns     The column names/labels
     * @return A list (=the rows) of maps (=the row's columns)
     */
	Object inputMatrix(String inputMatrixName, Map<String,Object> paramConfig, String...columns);

	/**
	 * Creates an input parameter that lets the user pick a file.
	 * It renders as a file input widget that allows to use data from a XLSX file in a logic.
	 * @param inputName The name of the parameter
	 * @return A handle that uniquely identifies the binary and its version that was assigned to the input
	 */
	Object parsableInputFile(String inputName);

	/**
	 * If a parsableInputFile has a value, this function opens the file and parses it to basic Groovy data structures.
	 * @param inputName The name of the parameter
	 * @return [data: [sheetName: [row1col1, row1col2, ...], [row2col1, row2col2, ...], ...]
	 */
	Object parsableInputFileData(String inputName);

	/**
	 * If a parsableInputFile has a value, this function opens the file and parses it to basic groovy datastructures.
	 * @param fileHandle The value of the parameter
	 * @return [data: [sheetName: [row1col1, row1col2, ...], [row2col1, row2col2, ...], ...]
	 */
	Object parsableInputFileDataFromHandle(String fileHandle);
	
	/**
	 * Creates a filter object that can be used in other API functions. The operator is always "equals", i.e. property = value.
     * <p>
     * Example:
     * <blockquote>
	 * <pre>
	 * def cas = api.find("CA",
	 * 		api.filter("assignmentId", mplId),
	 * 		api.filter("assignmentType", "MPL")
	 * )
	 * </pre>
	 * </blockquote>
	 *
	 * @param property The name of the property to filter on
	 * @param value    The specified value
	 * @return A filter object
	 * @deprecated Use {@link Filter#equal(String, Object)} instead
	 */
	Object filter(String property, Object value);

	/**
	 * Constructs a Filter object from the map representation of that filter (from the "wire format").
	 * The map representation is the way the filter is described in the JSON requests between the UI and the server.
	 *
	 * @param filterMap The filter in the form of nested maps
	 * @return The filter object
	 */
	Filter filterFromMap(Map<String, Object> filterMap);

	/**
	 * Creates a filter object that can be used in other API functions.
	 * <p>
	 * Note: This method is sort of dynamic. Dynamic in a sense that if the {@code opString} does not actually represent a valid operator (e.g. "="),
	 * the passed value will then be treated as the lower bound and the {@code value} as the upper bound. In other words, the resulting filter will look like this:
	 * {@code Filter.and(Filter.greaterOrEqual(property, opString), Filter.lessThan(property, value))}. If either parameter is {@code null}, its part
	 * is skipped.
     *
     * <p>Example:
     * <blockquote>
	 * <pre>{@code
	 * datamartQuery.where(api.filter("PricingDate", ">", targetDate.minus(365).format("yyyy-MM-dd")))
	 * }</pre>
	 * </blockquote>
     *
	 * @param property The name of the property to filter on
	 * @param opString One out of : "=", "&lt;", "&lt;=", "&gt;", "&gt;=", "&lt;&gt;", "~". If different, it gets treated as the lower bound value. Please see the remark.
	 * @param value    The specified value (or an upper bound)
	 * @return A filter object
	 * @see Filter#equal(String, Object)
	 * @see Filter#lessThan(String, Object)
	 * @see Filter#lessOrEqual(String, Object)
	 * @see Filter#greaterOrEqual(String, Object)
	 * @see Filter#greaterThan(String, Object)
	 * @see Filter#notEqual(String, Object)
	 * @see Filter#like(String, String)
	 * @see Filter#and(Filter...)
	 * @see Filter#or(Filter...)
	 * @see Filter#in(String, Object...)
	 * @see Filter#notIn(String, Object...)
	 * @deprecated Use the {@link Filter} functions instead
	 */
	Object filter(String property, Object opString, Object value);

	/**
	 * Creates a filter to be applied to a search on a related object type, limiting the result to objects of the related type that have a link to the given customer.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * 		def filter = api.customerToRelatedObjectsFilter("RBA", "104083")	// customerId=104083
	 * 		def rbas = api.find("RBA", 0, 0, null, filter)
	 * 		api.trace("rbas for customer #104083", null, rbas)
	 * </pre>
	 * </blockquote>
	 *
	 * @param relatedObjectTypeCode	TypeCode of the related object type. The possible values are:
	 * <ul>
	 * 		<li>'PR':    PriceRecord</li>
	 * 		<li>'CT':    Contract</li>
	 * 		<li>'CTLI':  ContractLineItem</li>
	 * 		<li>'RBA':   RebateAgreement</li>
	 * 		<li>'RBALI': RebateAgreementLineItem</li>
	 * </ul>
	 * @param customerId Customer.id to filter on
	 * @return A filter object
	 */
	Filter customerToRelatedObjectsFilter(String relatedObjectTypeCode, String customerId);

	/**
	 * Creates a filter to be applied to a customer search, limiting the result to customers that are linked to the given related object, the latter defined by its typedId.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def ras = api.find("RBA", 0, 1, null, null)
	 * ras.each {
	 * 		def filter = api.relatedObjectToCustomersFilter(it.typedId)
	 * 		def customers = api.find("C", 0, 0, null, filter)
	 * 		...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param relatedObjectTypedId	TypedId of the related object. Its type needs to be one of:
	 * <ul>
	 * 		<li>'PR':    PriceRecord</li>
	 * 		<li>'CT':    Contract</li>
	 * 		<li>'CTLI':  ContractLineItem</li>
	 * 		<li>'RBA':   RebateAgreement</li>
	 * 		<li>'RBALI': RebateAgreementLineItem</li>
	 * </ul>
	 * @return A filter object
	 */
	Filter relatedObjectToCustomersFilter(String relatedObjectTypedId);

	/**
	 * Creates a filter to be applied to a search on a related object type, limiting the result to objects of the related type that have a link to the given product.
	 <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def ras = api.find("RBA", 0, 1, null, null)
	 * ras.each {
	 * 		def raId = it.typedId.tokenize(".")[0] as Long
	 *   	def raFilter = api.relatedObjectToCustomersFilter("RBA", raId)
	 * 		def customers = api.find("C", 0, 0, null, filter)
	 * 		...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param relatedObjectTypeCode	TypeCode of the related object type. The possible values are:
	 * <ul>
	 * 		<li>'PR':    PriceRecord</li>
	 * 		<li>'CT':    Contract</li>
	 * 		<li>'CTLI':  ConttractLineItem</li>
	 * 		<li>'RBA':   RebateAgreement</li>
	 * 		<li>'RBALI': RebateAgreementLineItem</li>
	 * </ul>
	 * @param sku Product.id to filter on
	 * @return A filter object
	 */
	Filter productToRelatedObjectsFilter(String relatedObjectTypeCode, String sku);

	/**
	 * Creates a filter to be applied to a product search, limiting the result to products that are linked to the given related object, the latter defined by its typedId.
	 *
	 * @param relatedObjectTypedId	TypedId of the related object.  Its type needs to be one of:
	 * <ul>
	 * 		<li>'PR':    PriceRecord</li>
	 * 		<li>'CT':    Contract</li>
	 * 		<li>'CTLI':  ContractLineItem</li>
	 * 		<li>'RBA':   RebateAgreement</li>
	 * 		<li>'RBALI': RebateAgreementLineItem</li>
	 * </ul>
	 * @return A filter object
	 */
	Filter relatedObjectToProductsFilter(String relatedObjectTypedId);

	/**
	 * Checks if the current user is a member of the given user group.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def user = api.user("loginName")
	 * if (api.isUserInGroup("Manager", user)) {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param userGroupName The group name
	 * @param loginUserName The login name of the user
	 * @return true or false
	 */
	boolean isUserInGroup(String userGroupName, String loginUserName);

	/**
	 * Generates a trace message that can be used in the logic test drive. During the regular logic execution trace messages are ignored.
	 * The parameters specify the output of the trace.
	 *
	 * @param functionName A function name (free text of your choice)
	 * @param parameters   Parameters to trace
	 * @param result       The result of the trace
	 */
	void trace(String functionName, String parameters, Object result);

	/**
	 * Generates a trace message that can be used in the logic test drive. During the regular logic execution trace messages are ignored.
	 * The parameters specify the output of the trace.
	 *
	 * @param functionName A function name (free text of your choice)
	 * @param result       The result of the trace
	 */
	void trace(String functionName, Object result);
	
	/**
	 * Generates a trace message that can be used in the logic test drive. During the regular logic execution trace messages are ignored.
	 * @param result       The value to trace
	 */
	void trace(Object result);

	/**
	 * This function is meant to be used by the Pricefx internally only. You should use {@link #logInfo(String, Object)} or {@link #logWarn(String, Object)}.
	 * Logs a DEBUG message to the system error log which can be displayed in menu Tools / Logs / View Log.
	 * Log message can contain placeholders like {} which are then filled in dynamically with the argument's value.
	 * <b>Note:</b> To enable this function, you need to tick the setting "Allow Groovy logging to system log" in menu Tools / Configuration / General Settings.
	 *
	 * @param msg  The log message to print
	 * @param arg1 The arguments for the log message.
	 * @see #logInfo(String, Object)
	 * @see #logWarn(String, Object)
	 */
	void log(String msg, Object arg1);

	/**
	 * This function can be used by the support team only. You should use {@link #logInfo(Object)} or {@link #logWarn(Object)}.
	 * Logs a DEBUG message to the system error log which can be displayed in menu Tools / Logs / View Log.
	 * Log message can contain placeholders like {} which are then filled in dynamically with the argument's value.
	 * <b>Note:</b> To enable this, you need to tick the setting "Allow Groovy logging to system log" in menu Tools / Configuration / General Settings.
	 *
	 * @param msg  The message accompanying the exception
	 * @see #logInfo(Object)
	 * @see #logWarn(Object)
	 */
	void log(Object msg);

	/**
	 * Logs an INFO warning message to the system error log which can be displayed in menu Tools / Logs / View Log.
	 * Log message can contain placeholders like {} which are then filled in dynamically with the argument's value.
	 * <b>Note:</b> To enable this, you need to tick the setting "Allow Groovy logging to system log" in menu Tools / Configuration / General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 * api.markItemDirty()
	 * api.logInfo("SKU marked as dirty", sku)
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg  The log message to print
	 * @param arg1 The arguments for the log message
	 * @see #log(String, Object)
	 * @see #logWarn(String, Object)
	 */
	void logInfo(String msg, Object arg1);

	/**
	 * Logs a INFO message to the system error log which can be displayed in menu Tools / Logs / View Log.
	 * Log message can contain placeholders like {} which are then filled in dynamically with the argument's value.
	 * <b>Note:</b> To enable this, you need to tick the setting "Allow Groovy logging to system log" in menu Tools / Configuration / General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.logInfo("No items to be calculated")
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg The log message to print
	 * @see #log(Object)
	 * @see #logWarn(Object)
	 */
	void logInfo(Object msg);

	/**
	 * Logs a WARN message to the system error log which can be displayed in menu Tools / Logs / View Log.
	 * Log message can contain placeholders like {} which are then filled in dynamically with the argument's value.
	 * <b>Note:</b> To enable this, you need to tick the setting "Allow Groovy logging to system log" in menu Tools / Configuration / General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * if (checkItem()) {
	 *      api.logWarn("Incorrect or missing 'sku' parameter", sku)
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg  The log message to print
	 * @param arg1 The arguments for the log message
	 * @see #log(String, Object)
	 * @see #logInfo(String, Object)
	 */
	void logWarn(String msg, Object arg1);

	/**
	 * Logs a WARN message to the system error log which can be displayed in menu Tools / Logs / View Log.
	 * Log message can contain placeholders like {} which are then filled in dynamically with the argument's value.
	 * <b>Note:</b> To enable this, you need to tick  "Allow Groovy logging to system log" in menu Tools / Configuration / General Settings.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.logWarn("SKU cannot be null")
	 * </pre>
	 * </blockquote>
	 *
	 * @param msg  The message accompanying the exception
	 * @see #log(Object)
	 * @see #logInfo(Object)
	 */
	void logWarn(Object msg);

	/**
	 * Retrieves an already generated (in the previous logic step) input parameter in the form of a {@code ContextParameter} object.
	 * This can then be further modified as required.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def p = api.getParameter("Quantity")        // retrieve the context parameter with the same name as the input
	 * if (p != null && p.getValue() == null) {
	 *   p.setLabel("Required Quantity")           // set the displayed label
	 *   p.setRequired(true)                       // set the mandatory hint
	 *   p.setReadOnly(false)                      // set the read only flag
	 *   p.setValue(1)                             // set the default value with which the input is pre-populated
	 * }
	 * return qty ?: 1                             // if user doesn't specify a value, return 1
	 * }</pre>
	 * </blockquote>
	 *
	 * @param parameterName The named parameter to retrieve
	 * @return The parameter or {@code null}
	 */
	ContextParameter getParameter(String parameterName);

	/**
	 * In case of list processing (such as a price list or price grid), this method allows an easy access to values from the same list.
	 * It searches for the result record in the same list with the provided SKU.
	 * <p>
	 * Note: If used within a context that has a secondary key, only the first object is returned.
	 *
	 * @param sku The SKU to look for
	 * @return The result record (entire record)
	 */
	Map<String, Object> currentContext(String sku);

	/**
	 * In case of list processing (such as a price list or price grid), this method allows an easy access to values from the same list.
	 * It searches for the result record in the same list with the provided SKU.
	 *
	 * @param sku  The SKU to look for
	 * @param key2 The key2 to look for
	 * @return The result record (entire record)
	 */
	Map<String, Object> currentContext(String sku, String key2);

	/**
	 * In case of (chained) price lists, this method allows for an easy retrieval of records from the previous list.
	 * <p>
	 * Note: If used within a context that has a secondary key, only the first object is returned.
	 *
	 * @param sku The SKU to look for
	 * @return The result record (entire record)
	 */
	Map<String, Object> previousContext(String sku);

	/**
	 * In case of (chained) price lists, this method allows for an easy retrieval of records from the previous list.
	 *
	 * @param sku  The SKU to look for
	 * @param key2 The key2 to look for
	 * @return The result record (entire record)
	 */
	Map<String, Object> previousContext(String sku, String key2);

	/**
	 * The name of the current context, e.g. price list name, quote name etc.
	 *
	 * @return The name of a context
	 */
	String contextName();

	/**
	 * The type (type code string) of the current context.
	 * @return The type code string
	 */
	String contextType();

	/**
	 * The typedId of the current context.
	 * @return The typedId string
	 */
	String contextTypedId();

	/**
	 * Serves as an optimized way to get to certain object attributes. The sandbox automatically converts "real" server objects into
	 * a map representation when the object is accessed. This is for safety (sandboxing) reasons. This process converts
	 * all object attributes. If only one specific attribute is of interest, this method shortcuts that as the underlying object is not exposed
	 * (and thus does not force the sandbox into conversion).
	 *
	 * @param object The object whose value to retrieve
	 * @param path   The attribute name
	 * @return The value of the attribute
	 */
	Object getProperty(Object object, String path);

	/**
	 * Gets object attributes by elementName. It works only for objects that have meta attributes which contain the element name.
	 * These are e.g. PLI, SIMI, PGI, MPLI.
	 *
	 * @param object      The object whose value to retrieve
	 * @param elementName The elementName value
	 * @return The value of the attribute or {@code null} if invalid object or no matching meta found
	 */
	Object getPropertyByElementName(Object object, String elementName);

	/**
	 * In case of a list processing operation, this method gets the currently worked on record
     * (e.g., price list item or price grid item or quote item).
	 * <p>
     * Note: The price list item is initially created only after the first formula execution
     * (i.e. the PLI is created from the formula results).
     * At the execution time, the PLI does not exist yet and the method will return “null”.
     * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def currentItem = api.currentItem()
	 * def currentSku = currentItem != null ? currentItem.attribute20 : api.product("attribute20")
	 * </pre>
	 * </blockquote>
	 *
	 * @return The item object as a Map
	 */
	Object currentItem();

	/**
	 * In case of a list processing operation, this method gets the currently worked on record's property by elementName.
	 *
     *
	 * @param elementName The element name
	 * @return The item's property value
	 */
	Object currentItemByElementName(String elementName);

	/**
	 * Same as {@link #currentItem} but returns only the specified attribute.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def activePrice = api.currentItem("activePrice")
	 * </pre>
	 * </blockquote>
	 *
	 * @param attributeName The attribute to retrieve
	 * @return The value of the attribute
	 */
	Object currentItem(String attributeName);

	/**
	 * Returns a set of SKUs which comprise so-called 'SKU context' of
	 * the current formula execution.
	 * <p>
	 * The form of the SKU context differs, e.g., in case of the
	 * product detail formula executed via price grid details,
	 * the resulting set will contain all SKUs of the left-pane grid.
	 * <p>
	 * Note #1: If the SKU context is not supported by the current formula
	 * execution, {@code null} is returned.
	 * <p>
	 * Note #2: A maximum of 200 SKUs is returned per one call. If there are more,
	 * {@code startRow} must be used to get all SKUs in the set.
	 *
	 * @param startRow The starting row of the result set
	 * @return A set of SKUs or {@code null} if the SKU context is not
	 * available for the current formula execution
	 */
	Set<Object> contextSkuSet(int startRow);

	/**
	 * Returns the maximum number of records that {@link #find(String, int, int, String, Filter...)} and its variations can return in one call.
	 * The value is set in {@code formulaEngine.script.findMaxResults} system parameter.
	 * The default value is 2000 records.
	 *
	 * @return the maximum number of records
	 * @see #find(String, int, int, String, Filter...)
	 */
	int getMaxFindResultsLimit();
	
	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. A maximum of 200 items is returned.
	 * You should use {@link #find(String, int, int, String, List, Filter...)} instead because of performance reasons.
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @return List of objects represented as Maps
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
     * @deprecated use {@link #find(String, int, int, String, List, Filter...)} in order to always specify the fields to be returned because of performance.
     */
	List<Object> find(String typeCode, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. A maximum of 200 items is returned.
	 * The start row parameter can be used to page through more than 200 results.
	 * If you don't need all the columns to be fetched, you should use {@link #find(String, int, int, String, List, Filter...)} instead because of performance reasons.
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @param startRow The starting row of the result set
	 * @return List of objects represented as Maps
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
     * @deprecated use {@link #find(String, int, int, String, List, Filter...)} because of performance you should always specify the fields to be returned.
	 */
	List<Object> find(String typeCode, int startRow, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. A maximum of 200 items is returned.
	 * The start row parameter can be used to page through more than 200 results.
	 * If you don't need all the columns to be fetched, you should use {@link #find(String, int, int, String, List, Filter...)} instead because of performance reasons.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * def filters = [
	 *     Filter.equal("name", "Competition"),
	 *     Filter.equal("sku", sku),
	 * ]
	 *
	 * // sort by competitor's price ASC
	 * def competitors = api.find("PX", 0, "attribute1", *filters)
	 *
	 * competitors.each {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param startRow The starting row of the result set
	 * @param sortBy   String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @return List of objects represented as Maps
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
     * @deprecated use {@link #find(String, int, int, String, List, Filter...)} because of performance you should always specify the fields to be returned.
	 */
	List<Object> find(String typeCode, int startRow, String sortBy, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter. The maxRows parameter can only set to lower values.
	 * <p>
     * Example:
     * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * def filters = [
	 *     Filter.equal("name", "Competition"),
	 *     Filter.equal("sku", sku),
	 * ]
	 *
	 * // sort by competitor's price ASC
	 * def competitors = api.find("PX", 0, api.getMaxFindResultsLimit(), "-attribute2", *filters)
	 *
	 * competitors.each {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param startRow The starting row of the result set
	 * @param maxRows  The maximum number of rows to retrieve. Value can be up to {@link #getMaxFindResultsLimit()}. If set to 0 it will default to 200
	 * @param sortBy   String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @return List of objects represented as Maps
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, Filter...)
	 * @see #count(String, Filter...)
	 */

	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only set to lower values than that.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def sku = api.product("sku")
	 *
	 * def filters = [
	 *     Filter.equal("name", "Competition"),
	 *     Filter.equal("sku", sku),
	 * ]
	 *
	 * // sort by competitor's price ASC
	 * def competitors = api.find("PX", 0, api.getMaxFindResultsLimit(), "-attribute2", ["attribute2", "attribute5", "attribute6"], *filters)
	 *
	 * competitors.each {
	 *     ...
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param startRow The starting row of the result set
	 * @param maxRows  The maximum number of rows to retrieve. Value can be up to {@link #getMaxFindResultsLimit()}. If set to 0 it will default to 200
	 * @param sortBy   String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param fields   A list of fields to returned in the result. If you really need all fields, you can set this parametr to null
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @return List of objects represented as Maps
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, List, Filter...)
	 * @see #count(String, Filter...)
	 */
	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, List<String> fields, Filter... filters);

	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only set lower values than that.
	 *
	 * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param startRow           The starting row of the result set
	 * @param maxRows            The maximum number of rows to retrieve. Value can be up to {@link #getMaxFindResultsLimit()}. If set to 0 it will default to 200
	 * @param sortBy             String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param fields             A list of fields to returned in the result. If you really need all fields, you can set this parametr to null
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            An optional list of filters (that will evaluated using AND operator)
	 * @return List of objects represented as Maps
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, List, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, List<String> fields, boolean distinctValuesOnly, Filter... filters);
	
	/**
	 * The Swiss army knife method of the API. With this method you can search for all types of objects with a freestyle query.
	 * The filters are ANDed together. The maximum number of rows retrieved is determined by the formulaEngine.script.findMaxResults system parameter.
	 * The maxRows parameter can only set lower values than that.
	 *
	 * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param startRow           The starting row of the result set
	 * @param maxRows            The maximum number of rows to retrieve. Value can be up to {@link #getMaxFindResultsLimit()}. If set to 0 it will default to 200
	 * @param sortBy             String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param fields             A map of pairs [fieldname : aggregation] holding fields to be returned in the result  where aggregation can be one of SUM, AVG, MIN or MAX
	 * 							 The key of the map is the field name. The value can be a field operator like SUM, AVG, MIN or MAX.
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            An optional list of filters (that will evaluated using AND operator)
	 * @return List of objects represented as Maps
	 * @see #getMaxFindResultsLimit()
	 * @see #stream(String, String, Map, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	List<Object> find(String typeCode, int startRow, int maxRows, String sortBy, Map<String,String> fields, boolean distinctValuesOnly, Filter ... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
	 * If you don't need all the columns to be fetched, you should use {@link #stream(String, String, List, Filter...)} instead because of performance reasons.
	 * <p>
	 * Important note: you have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def filters = [
	 *      Filter.equal("attribute2", "Active")
	 * ]
	 *
	 * def products = api.stream("P", "-attribute1", *filters)
	 *
	 * products.each { product ->
	 *      ...
	 * }
	 * }</pre>
	 * </blockquote>
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param sortBy   String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @return The result implements the Iterator interface, allowing for iteration over all records in the scope of the query.
	 * @see #find(String, int, String, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, Filter... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 * <p>
	 * Important note: you have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def filters = [
	 *      Filter.equal("attribute2", "Active")
	 * ]
	 *
	 * def products = api.stream("P", "-attribute1", ["attribute2"], *filters)
	 *
	 * products.each { product ->
	 *      ...
	 * }
	 * }</pre>
	 * </blockquote>
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param sortBy   String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param fields   A list of attributes you want to return in the result, i.e. skip all attributes not contained in the list that the search would return otherwise.
	 * @param filters  An optional list of filters (that will evaluated using AND operator)
	 * @return The result implements the Iterator interface, allowing for iteration over all records in the scope of the query.
	 * @see #find(String, int, int, String, List, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, List<String> fields, Filter... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 * <p>
	 * Important note: you have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 *
	 * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param sortBy             String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param fields             A list of attributes you want to return in the result, i.e. skip all attributes not contained in the list that the search would return otherwise.
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            An optional list of filters (that will evaluated using AND operator)
	 * @return The result implements the Iterator interface, allowing for iteration over all records in scope of the query.
	 * @see #find(String, int, int, String, List, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, List<String> fields, boolean distinctValuesOnly, Filter... filters);

	/**
	 * Searches for all types of objects with a freestyle query and iterates over the result set.
	 * The filters are ANDed together.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 * <p>
	 * Important note: you have to close the iterator after iterating the data using {@link ResultIterator#close()}.
	 *
	 * @param typeCode           A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param sortBy             String of fields/attributes, separated by comma, based on which the resultset should be sorted. Any field can be prefixed with '-' for descending sort order
	 * @param fields             Map of pairs [fieldname : aggregation] holding fields to be returned in the result  where aggregation can be one of SUM, AVG, MIN or MAX
	 * 							 The key of the map is the field name. The value can be a field operator like SUM, AVG, MIN or MAX
	 * @param distinctValuesOnly Fetches distinct/all values. The use case here is fetching distinct single column values.
	 * @param filters            An optional list of filters (that will evaluated using AND operator)
	 * @return The result implements the Iterator interface, allowing for iteration over all records in scope of the query.
	 * @see #find(String, int, int, String, Map, boolean, Filter...)
	 * @see #count(String, Filter...)
	 */
	ResultIterator stream(String typeCode, String sortBy, Map<String,String> fields, boolean distinctValuesOnly, Filter ... filters);

	/**
	 * When given a list of entities (result of api.find,...), it converts the list of meta
	 * attributed entities to a list of maps where each map is a serialized entity which keys are using field names
	 * from attribute meta object.
	 *
	 * i.e. converts attributeX fields to names defined in metadata
	 * @param objects list of entities (results of api.find,...)
	 * @return list of serialized entities where keys have the names from meta data
	 */
	List<Map<String, Object>> namedEntities(List<Object> objects);

	/**
	 * Emits an object to be consumed by an external thread. Requires a blocking emitQueue to be set in the FormulaEngineContext and
	 * a consuming thread to take the emitted objects from the queue. See also FeederFormulaExecutor.
     * <p>
     * This method is not supported in the Syntax Check mode.
	 *
	 * @param obj The object to be emitted/inserted in the emit queue
	 * @return True if the object is successfully inserted in the queue, false otherwise.
	 * @throws InterruptedException if interrupted while waiting
	 */
	boolean emit(Object obj) throws InterruptedException;

	/**
	 * Emits the rebate records matching the given search criteria - to be consumed by a PA allocation task (a PA calculation task with a feeder formula calling
	 * this method, and a regular formula accepting each emitted RebateRecord as its currentItem context, and distributing one or more RebateRecord accruals
	 * over the set of PA rows which have contributed to those accruals in some way (by revenue, bvolume...).
	 *
	 * @param dateFieldName	The name of the field to be substituted in the RebateRecord calculationBase (a DMDataSlice object that includes a generic time dimension filter)
	 * @param sortBy		The RebateRecord field on which to sort
	 * @param filters		The filters to apply to the RebateRecord search
	 * @throws InterruptedException if interrupted while waiting
	 */
	void emitRebateRecords(String dateFieldName, String sortBy, Filter... filters) throws InterruptedException;

	/**
	 * Finds the PersistedObject with the given typedId and emits it if it exists. The assumption is that there is a receiving consumer, as in {@link #emitRebateRecords}.
	 * @param typedId	The typedId of the PersistedObject to emit
	 * @return false if failed
	 * @throws InterruptedException if interrupted while waiting
	 */
	boolean emitPersistedObject(String typedId) throws InterruptedException;
	
	/**
	 * Finds the PersistedObject for the given type and ID and emits it if it exists. The assumption is that there is a receiving consumer, as in {@link #emitRebateRecords}.
	 * @param typeCode	The typeCode of the PersistedObject to emit
	 * @param id		The ID of the PersistedObject to emit
	 * @return false if failed
	 * @throws InterruptedException if interrupted while waiting
	 */
	boolean emitPersistedObject(String typeCode, Long id) throws InterruptedException;

	/**
	 * Returns the number of records that a corresponding {@link #find(String, int, int, String, List, Filter...)} would return.
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param filters  A list of {@link Filter} objects that are ANDed together
	 * @return The number of matching records
	 */
	int count(String typeCode, Filter... filters);

	/**
	 * Finds a price parameter table object (LT) of the given name. It does apply target date and simulation set matching.
	 * There is no caching of the values, so the caching should be considered.
	 *
	 * @param tableName The name of the price parameter table
	 * @return The price parameter table object (as Map) or {@code null} if the table does not exists or it's validAfter &gt; targetDate
	 */
	LookupTable findLookupTable(String tableName);

	/**
	 * Returns all values of a named price parameter table (optionally filtered).
	 * There is no caching of the values, so you should should consider caching it yourself.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.retainGlobal = true
	 *
	 * // for the first time
	 * if (api.global.salesOrgs == null) {
	 *     // store all values as a Map [name : value] to global cache
	 *     api.global.salesOrgs = api.findLookupTableValues("SalesOrg").collectEntries {
	 *         [(it.name): it.value]
	 *     }
	 * }
	 *
	 * return api.global.salesOrgs
	 * </pre>
	 * </blockquote>
	 *
	 * @param tableName The name of the price parameter table
	 * @param filters   Optional extra {@link Filter}
	 * @return The price parameter table values or {@code null}
	 */
	List<Object> findLookupTableValues(String tableName, Filter... filters);

	/**
	 * Returns all values of a named price parameter table (optionally filtered and sorted).
	 * There is no caching of the values, so you should should consider caching it yourself.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.retainGlobal = true
	 *
	 * // for the first time
	 * if (api.global.salesOrgs == null) {
	 *     // store all values as a Map [name : value] to global cache
	 *     api.global.salesOrgs = api.findLookupTableValues("SalesOrg").collectEntries {
	 *         [(it.name): it.value]
	 *     }
	 * }
	 *
	 * return api.global.salesOrgs
	 * </pre>
	 * </blockquote>
	 *
	 * @param tableName The name of the price parameter table
	 * @param sortBy  A sort key. {@code null} if no sorting is required.
	 * @param filters   Optional extra {@link Filter}
	 * @return The price parameter table values or {@code null}
	 */
	List<Object> findLookupTableValues(String tableName, String sortBy, Filter... filters);

	/**
	 * Finds active workflows for the given approvable. You can get the same information as displayed in the workflow UI.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def quoteUniqueName = "P-123"
	 * def quote = api.find('Q', 0, 1, Filter.equal("uniqueName", quoteUniqueName)).getAt(0)
	 * return quote ? api.findWorkflowInfo('Q', quote.id) : null
	 * </pre>
	 * </blockquote>
	 *
	 * @param approvableTypeCode The type code of approvable. Approvables are e.g. quotes, rebate agreements, data change requests etc.
	 * @param approvableId       ID of the approvable
	 * @return Table like structure resembling workflow UI or {@code null} if not found
	 * @see #findPreviousWorkflowInfo(String, String...)
	 * @see #findPreviousApprovableState(String, String...)
	 */
	WorkflowInfo findWorkflowInfo(String approvableTypeCode, Object approvableId);

	/**
	 * Finds the previous approvable entity state if some DENIED or WITHDRAWN workflow for the approvable exists.
	 * @param typedId Approvable entity typed ID
	 * @param workflowStatuses Only DENIED or WITHDRAWN are supported
	 * @return State of the approvable entity when its previous workflow was DENIED or WITHDRAWN
	 * @see #findWorkflowInfo(String, Object)
	 */
	Map<String, Object> findPreviousApprovableState(String typedId, String... workflowStatuses);

	/**
	 * Finds the approvable entity previous workflow history if some DENIED or WITHDRAWN workflow for the approvable exists.
	 * @param typedId Approvable entity typed ID
	 * @param workflowStatuses Only DENIED or WITHDRAWN are supported
	 * @return Workflow history, see net.pricefx.workflowengine.WorkflowInfo
	 * @see #findWorkflowInfo(String, Object)
	 */
	WorkflowInfo findPreviousWorkflowInfo(String typedId, String... workflowStatuses);

	/**
	 * Finds all assignment records for the given customerId.
	 * Also matches by CustomerGroup definitions.
	 *
	 * @param customerId The customer whose assignments are searched for
	 * @return A list of assignment records
	 */
	List<CustomerAssignment> findCustomerAssignments(String customerId);

	/**
	 * Returns a list of price lists that match the targetDate (i.e. list's targetDate &lt;= targetDate and expiryDate &gt; targetDate).
	 *
	 * @param targetDate The targetDate to use
	 * @param listName   Price list name. If {@code null}, the list name is not included in search.
	 * @return The list of matching price lists
	 */
	List<Pricelist> findPricelists(Date targetDate, String listName);

	/**
	 * Returns a list of simulations that match the targetDate (i.e. list's targetDate &lt;= targetDate).
	 *
	 * @param targetDate     The targetDate to use
	 * @param simulationName Simulation name. If {@code null}, the simulation name is not included in search.
	 * @return The list of matching simulations
	 */
	List<Simulation> findSimulations(Date targetDate, String simulationName);

	/**
	 * Returns a list of price grids that match the targetDate (i.e. price grid's targetDate &lt;= targetDate).
	 *
	 * @param targetDate    The targetDate to use
	 * @param priceGridName Price grid name. If {@code null}, the price grid name is not included in search.
	 * @return The list of matching price grids
	 */
	List<PriceGrid> findPriceGrids(Date targetDate, String priceGridName);

	/**
	 * Returns a list of calculated field sets that match the targetDate.
	 *
	 * @param targetDate Target date for the CFS
	 * @param cfsName    CFS name. If {@code null}, the CFS name is not included in search.
	 * @return The list of matching calculated field sets
	 */
	List<CalculatedFieldSet> findCalculatedFieldSets(Date targetDate, String cfsName);

	/**
	 * Returns a list of manual price lists that match the targetDate (i.e. list's validAfter &lt;= targetDate and expiryDate &gt; targetDate).
	 *
	 * @param targetDate The targetDate to use
	 * @param listName   Price list name. If {@code null}, the list name is not included in search.
	 * @return The list of matching price lists
	 */
	List<ManualPricelist> findManualPricelists(Date targetDate, String listName);

	/**
	 * Returns a list of summary items. One item is returned for every matching (=same SKU) PLI or XPLI which is in an approved list.
	 *
	 * @param sku The SKU to filter on
	 * @return A list of summary items
	 */
	List<SummaryCalculableObjectItem> findApprovedPricelistItems(String sku);

	/**
	 * Retrieves a list of customers that match the given CustomerGroup (returns first 200 entries).
	 *
	 * @param group The group to match
	 * @return A list of customers
	 */
	List<Customer> findCustomersInGroup(CustomerGroup group);

	/**
	 * Retrieves a list of customers that match the given CustomerGroup with custom paging and sorting.
	 *
	 * @param group The customer group
	 * @param startRow The starting row number (maximum 200 rows returned at once)
	 * @param sortBy   A sort key. {@code null} if no sorting is required.
	 * @return A list of customers
	 */
	List<Customer> findCustomersInGroup(CustomerGroup group, int startRow, String sortBy);

	/**
	 * Returns a user entered value that overrides the calculated result of a logic element.
     * <p>
     * Note that the manualOverride works as a flag - when it is set for an element, the final
     * value of that element is taken from the manually entered value and the calculation of that
     * element does not run at all. If this flag is not set, the final value is the calculated result.
	 *
	 * @param elementName The formula/result element that is overridden
	 * @return The overridden value or {@code null} if not overridden
	 */
	Object getManualOverride(String elementName);

	/**
	 * Removes a user entered value that overrides the calculated result of a logic element.
     * <p>
     * If you have a condition to remove the manual override, it is recommended to put this condition
     * in a separate logic element which is always executed (not in the same one which has the manual
     * override allowed; the reason is that once the user enters a value manually, the element calculation
     * does not run at all, so the condition would not be evaluated).
	 *
     * Note also that a logic element with api.removeManualOverride() must
     * be before an element which is cleared.
     *
	 * @param elementName The formula/result element that is overridden
	 */
	void removeManualOverride(String elementName);

	/**
	 * Aborts the current item calculation. The subsequent logic elements are not executed and
	 * the calculation engine continues with calculation of the next item.
	 * This function only sets the abortCalculation flag, so the element where abortCalculation is called
	 * still returns a value (and is the last element returning a value).
	 * This function can be called together with {@link #isSyntaxCheck()} after the parameters are gathered.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * if (api.isSyntaxCheck()) {
	 *   api.abortCalculation()
	 * }
	 * </pre>
	 * </blockquote>
	 */
	void abortCalculation();

	/**
	 * Aborts the current item calculation. The subsequent logic elements are not executed and
	 * the calculation engine continues with calculation of the next item.
	 * This function only sets the abortCalculation flag, so the element where abortCalculation is called
	 * still returns a value (and is the last element returning a value).
	 * This function can be called together with {@link #isSyntaxCheck()} after the parameters are gathered.
	 *
	 * <p>
	 * @param keepOld Should the old outputs be retained? Only applicable for quote and contract calculation!
	 */
	void abortCalculation(boolean keepOld);

	/**
	 * Aborts the current item parameter determination. This helps in logic with many elements to tell the engine
	 * to not to spend more time on next elements when logic author is sure that there is no any statement
	 * that would generate any new inputs.
	 * This function can be called together with {@link #isSyntaxCheck()} after the parameters are gathered.
	 *
	 */
	void abortSyntaxCheck();

	/**
	 * Throws a calculation exception with the specified message and aborts the calculation.
	 *
	 * @param errorMessage The error message that is displayed
	 * @throws XExpression The custom calculation exception
	 */
	void throwException(String errorMessage) throws XExpression;

	/**
	 * Sets the dirty flag on the specified item, i.e. the second calculation run will be triggered and this item will be included in the 2nd pass.
	 * If used in a MatrixPricelist/MatrixSimulation (i.e. on objects that have a secondary key), all items with that SKU are marked as dirty.
	 * See also {@link #getIterationNumber()}, {@link #markItemClean()}.
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance. The default is 1 dirty pass.
	 *
	 * @param sku The item to have the dirty flag set
	 * @see #markItemClean(String)
	 */
	void markItemDirty(String sku);

	/**
	 * Sets the dirty flag on the specified item, i.e. the second calculation run will be triggered and this item will be included in the 2nd pass.
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance. The default is 1 dirty pass.
	 * See also {@link #getIterationNumber()}, {@link #markItemClean()}.
	 *
	 * @param sku The item to have the dirty flag set
	 * @param key2 key2 field
	 * @see #markItemClean(String, String)
	 */
	void markItemDirty(String sku, String key2);

	/**
	 * Sets the dirty flag on the specified item, i.e. the second calculation run will be triggered and this item will be included in the 2nd pass.
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance. The default is 1 dirty pass.
	 * See also {@link #getIterationNumber()}, {@link #markItemClean()}.
	 *
	 * @see #markItemClean()
	 */
	void markItemDirty();

	/**
	 * Removes the dirty flag from the current item set previously by {@link #markItemDirty()}.
	 *
	 * @see #markItemDirty()
	 */
	void markItemClean();

	/**
	 * Removes the dirty flag from the specified matrix item set previously by {@link #markItemDirty(String, String)}.
	 *
	 * @param sku The item from which to remove the dirty flag
	 * @param key2 key2 field
	 * @see #markItemDirty(String, String)
	 */	
	void markItemClean(String sku, String key2);

	/**
	 * Removes the dirty flag from the specified item set previously by {@link #markItemDirty(String)}.
	 *
	 * @param sku The item from which to remove the dirty flag
	 * @see #markItemDirty(String)
	 */	
	void markItemClean(String sku);
	
	/**
	 * Determines if the calculation is a dirty pass. Does not necessarily mean it is a 2nd pass calculation.
	 *
	 * @return true if is in a dirty pass
	 * @deprecated Use {@link #getIterationNumber()} &gt; 0 instead.
	 */
	boolean isSecondPass();

	/**
	 * Returns the number of iterations this list has been calculated during this cycle.
	 * 0 means: initial calculation (non-dirty)
	 * 1 means: 1st dirty pass, {@link #markItemDirty()} was called in the previous pass
	 * 2 means: 2nd dirty pass, {@link #markItemDirty()} was called in the previous pass
	 * and so on
	 * The maximum number of dirty passes is defined by the server configuration constant {@code maxDirtyCalculations} defined for the whole instance. The default is 1 dirty pass.
	 * @return The iteration number
	 */
	int getIterationNumber();

	/**
	 * Formats a string like {@link java.lang.String#format}.
	 *
	 * @param locale The locale (as a String)
	 * @param format The format string
	 * @param args   The arguments to pass on
	 * @return The formatted string
	 */
	String formatString(String locale, String format, Object... args);

	/**
	 * Formats a number as a String. The same semantics as {@link java.text.NumberFormat} and {@link java.text.DecimalFormat#applyPattern} apply.
	 *
	 * @param pattern The pattern string
	 * @param n       The number to format
	 * @return The formatted number as a String
	 */
	String formatNumber(String pattern, Object n);

	/**
	 * Formats a number as a String. The same semantics as {@link java.text.NumberFormat} and {@link java.text.DecimalFormat#applyPattern} apply.
	 *
	 * @param pattern The pattern string
	 * @param n       The number to format
	 * @param locale  The locale to use for formatting
	 * @return The formatted number as a String
	 */
	String formatNumber(String locale, String pattern, Object n);

	/**
	 * A utility method to decode a JSON string into a Map representation.
	 *
	 * @param json The input JSON string
	 * @return The parsed JSON as a Map
	 */
	Object jsonDecode(String json);

	/**
	 * A utility method to decode a JSON string into a List representation.
	 *
	 * @param json The input JSON string
	 * @return The parsed JSON as a List
	 */
	Object jsonDecodeList(String json);

	/**
	 * A utility method to encode an object to a JSON string. Omits all null values in the output.
	 *
	 * @param input The input object, usually a map
	 * @return Encoded object
	 */
	Object jsonEncode(Object input);

	/**
	 * 
	 * A utility method to encode an object to a JSON string.
	 *
	 * @param input The input object, usually a map
	 * @param omitNull Determines is null values should be omitted
	 * @return Encoded object
	 */
	Object jsonEncode(Object input, boolean omitNull);

	/**
	 * Marks the current logic element as a critical alert and provides the given message.
	 * In a price grid, price list or in a quote line item, it marks the whole record with a dark red background color.
	 * In the calculation results, it displays a red circle with an exclamation mark next to the element.
	 * To display the error message for price grids and price lists, click the arrow in the Detail column.
	 * Displayed only if called from an element that has its 'Display mode' enabled for the given context ('PriceGrids', 'Pricelists' or 'Quote Configurator').
	 * In a quote context, it blocks submission of the quote.
	 *
	 * @param msg The alert message
	 * @see #redAlert(String)
	 * @see #yellowAlert(String)
	 */
	void criticalAlert(String msg);

	/**
	 * Marks the current logic element with a red alert and provides the given message.
	 * In a price grid, price list or in a quote line item, it marks the corresponding field background with a red background color.
	 * In the calculation results, it displays a plain red circle next to the element.
	 * To display the error message for price grids and price lists, click the arrow in the Detail column.
	 * Displayed only if called from an element that has its 'Display mode' enabled for the given context ('PriceGrids', 'Pricelists' or 'Quote Configurator').
	 *
     * @param msg The alert message
	 * @see #criticalAlert(String)
	 * @see #yellowAlert(String)
	 */
	void redAlert(String msg);

	/**
	 * Marks the current logic element with a yellow alert and provides the given message.
	 * In a price grid, price list or in a quote line item, it marks the corresponding field background with a yellow background color.
	 * In the calculation results, it displays a plain yellow circle next to the element.
	 * To display the error message for price grids and price lists, click the arrow in the Detail column.
	 * Displayed only if called from an element that has its 'Display mode' enabled for the given context ('PriceGrids', 'Pricelists' or 'Quote Configurator').
	 *
	 * @param msg The alert message
	 * @see #criticalAlert(String)
	 * @see #redAlert(String)
	 */
	void yellowAlert(String msg);

	/**
	 * Marks the current formula element with the given alert message.
	 * If one was already set by means of a yellowAlert, redAlert or criticalAlert call, then that message is replaced. 
	 * When not raising an alert through any of these calls, but still setting a message with this method, 
	 * then the message will be displayed wherever calculation results are shown with alert messages, but withoud any yellow or red background coloring. 
	 *
	 * @param msg The alert message
	 */
	void setAlertMessage(String msg);

	/**
	 * @deprecated
	 * Calls {@link #buildFlexChart(String, Object) buildFlexChart(null, definition)}
	 * See the full documentation in {@link #buildFlexChart(Object)}.
	 *
	 * @param definition the definition of the {@code FlexChart}
	 * @return ResultFlexChart object
	 * @see #buildHighchart(Map)
	 * @see #buildFlexChart(String, Object)
	 */
	ResultFlexChart buildFlexChart(Object definition);

	/**
	 * @deprecated
	 * Creates a {@link FlexChart} from the passed {@code definition} by
	 * merging it with {@code options} defined within the
	 * {@code baseTemplateToUse} template.
	 * <p>
	 * Although it may change in future releases, for now, a
	 * {@code FlexChart} is nothing else than a {@code Highcharts}
	 * chart. Therefore, the passed {@code definition} (also applies to the
	 * {@code baseTemplateToUse} template) must denote
	 * {@code Highcharts options} - please consult {@code Highcharts}
	 * documentation for more details, and <b>note that FlexCharts must comply
	 * to the outdated Highcharts release 4</b>.
	 * <p>
	 * Since {@code Highcharts options} are normally specified by a
	 * {@code JavaScript object literal}, the passed
	 * {@code definition} must be either a {@code JSON} string
	 * representing such {@code object literal}, or any {@code object}
	 * which can be mapped to such {@code JSON} representation. Therefore,
	 * typically the {@code definiton} will be passed as a
	 * {@code map of maps}.
	 * <p>Example:
	 * <blockquote>
	 * <pre>
	 * def definition = [
	 *   chart: [
	 *     type: "bar"
	 *   ],
	 *   title: [
	 *     text: "A common FlexChart"
	 *   ],
	 *   series: [
	 *    [
	 *      data: [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21]
	 *    ]
	 *   ]
	 * ]
	 *
	 * api.buildFlexChart(definition)
	 * </pre>
	 * </blockquote>
	 * Note that not all {@code Highcharts options} are supported.
	 * Moreover, usage of some options is prohibited deliberately. These are
	 * currently {@code global}, {@code events} and every
	 * {@code JavaScript function}-based options.
	 * <p>
	 * Note: {@code FlexChart} templates are defined by an administrator in
	 * the '{@code Configuration}' section.
	 * <p>
	 *
	 * @param baseTemplateToUse The name of the template to use as a basis for merging with
	 *                          the passed {@code definition}; if {@code null}, the
	 *                          predefined Default template is used
	 * @param definition        The definition of the {@code FlexChart}; currently, this
	 *                          will typically be a string or map of maps containing
	 *                          {@code Highcharts options}
	 * @return A helper object linking the resulting {@code FlexChart},
	 * {@code baseTemplateToUse} and type of the result - which
	 * will always be 'FLEXCHART', here
	 * @see <a href="http://api.highcharts.com/highcharts">http://api.highcharts.com/highcharts</a>
	 * @see <a href="http://www.highcharts.com/docs">http://www.highcharts.com/docs</a>
	 * @see <a href="http://json.org/">http://json.org</a>
	 * @see #buildFlexChart(Object)
	 * @see #buildHighchart(Map)
	 */
	ResultFlexChart buildFlexChart(String baseTemplateToUse,
	                               Object definition);

	/**
	 * Creates a {@link ResultHighchart} from the passed {@code definition}.
	 * <br>
	 * As opposed to {@code FlexChart}, the definition is passed straight to the
	 * {@code Highcharts} library, currently release 7.1. There is no validation,
	 * no templates merging, and JavaScript functions are not supported.
	 * @param definition        The definition of the {@code Highchart} as a map of maps and arrays
	 * @return ResultHighchart object
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a>
	 * @see <a href="https://api.highcharts.com/highcharts">https://api.highcharts.com/highcharts</a>
	 * @see #buildFlexChart(Object)
	 */
	ResultHighchart buildHighchart(final Map<String, ?> definition);

	/**
	 * Creates a {@link ResultHighmap} from the passed {@code definition}.
	 * <br>
	 * As opposed to {@code FlexChart}, the definition is passed straight to the
	 * {@code Highmaps} library, currently release 7.1. There is no validation,
	 * no templates merging, and JavaScript functions are not supported.
	 * {@code HighChart} provide better performance than {@code FlexChart} since
	 * @param definition        The definition of the {@code Highmap} as a map of maps and arrays
	 * @return buildHighmap object
	 * @see <a href="https://www.highcharts.com/docs">https://www.highcharts.com/docs</a>
	 * @see <a href="https://api.highcharts.com/highmaps">https://api.highcharts.com/highmaps</a>
	 */
	ResultHighmap buildHighmap(final Map<String, ?> definition);

	/**
	 * Creates a histogram that adapts to an unknown data distribution.
	 * It keeps a more or less constant resolution throughout the data range by increasing 
	 * the resolution where the data is more dense. For example, if the data has
	 * such a distribution that most of the values lie in the 0-5 range and only a few are 
	 * in the 5-10 range, the histogram would adapt and assign more counting buckets to
	 * the 0-5 range and less to the 5-10 range.
	 * The histogram provides a method to obtain the accumulative density function
	 * for a given data point (getValueForPercentile), and a method to obtain the data point that splits the 
	 * data set at a given percentile (getValueForPercentile). 
	 * @return A new AdaptiveHistogram object
	 */
	Object newAdaptiveHistogram();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against price list(s).
	 *
	 * @return The query definition object
	 */
	ItemSummaryQuery<Pricelist> getPricelistSummaryQuery();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against simulation(s).
	 *
	 * @return The query definition object
	 */
	ItemSummaryQuery<Simulation> getSimulationSummaryQuery();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against price grid(s).
	 *
	 * @return The query definition object
	 */
	ItemSummaryQuery<PriceGrid> getPriceGridSummaryQuery();

	/**
	 * Creates a summary query object that can be parameterized and is then subsequently used in {@link #runSummaryQuery(ItemSummaryQuery)}.
	 * This query runs against rebate records.
	 *
	 * @return The query definition object
	 */
	ItemSummaryQuery<RebateRecordSet> getRebateRecordSummaryQuery();

	/**
	 * Runs (executes) a summary query defined by the parameterized query object which was created by one of the getXXXSummaryQuery calls.
	 *
	 * @param query The query to execute
	 * @return The result set of the query
	 */
	List<Map<String, Object>> runSummaryQuery(ItemSummaryQuery<?> query);

	/**
	 * Creates a configurator entry object and adds the specified input parameter as the first parameter to the configurator entry.
	 * If the parameter had a value in the previous round trip, that value is set automatically.
	 *
	 * @param type Input type of the parameter. Valid enum values are:
	 * 	PRODUCT,	PRODUCTGROUP,	CUSTOMER,	CUSTOMERGROUP,	USERENTRY,	STRINGUSERENTRY, OPTION,
	 * 	INTEGERUSERENTRY,	BOOLEANUSERENTRY,	DATEUSERENTRY,	TIMEUSERENTRY,	DATETIMEUSERENTRY,	MULTITIERENTRY,
	 * 	INPUTMATRIX, OPTIONS,	MATRIXLOOKUP,	LOOKUP,	FILTERBUILDER,	DMFILTERBUILDER,	DMFIELD,	DMDIMFILTER,
	 * 	CONFIGURATOR,	HIDDEN 
	 * @param name The name of the parameter
	 * @return The configurator entry with the input parameter added
	 */
	ConfiguratorEntry createConfiguratorEntry(InputType type, String name);

	/**
	 * Creates a configurator entry object and adds the specified input parameter as the first parameter to the configurator entry.
	 * If the parameter had a value in the previous round trip, that value is set automatically.
	 *
	 * @param type Input type of the parameter. Valid enum values are:
	 * 	PRODUCT,	PRODUCTGROUP,	CUSTOMER,	CUSTOMERGROUP,	USERENTRY,	STRINGUSERENTRY, OPTION,
	 * 	INTEGERUSERENTRY,	BOOLEANUSERENTRY,	DATEUSERENTRY,	TIMEUSERENTRY,	DATETIMEUSERENTRY,	MULTITIERENTRY,
	 * 	INPUTMATRIX, OPTIONS,	MATRIXLOOKUP,	LOOKUP,	FILTERBUILDER,	DMFILTERBUILDER,	DMFIELD,	DMDIMFILTER,
	 * 	CONFIGURATOR,	HIDDEN 
	 * @param name       The name of the parameter
	 * @param targetDate Target date for the input parameter (not applicable to all input types)
	 * @return The configurator entry with the input parameter added
	 */
	ConfiguratorEntry createConfiguratorEntry(InputType type, String name, Date targetDate);

	/**
	 * Creates an empty configurator entry object. Must be filled with input parameters by calling createParameter on the object.
	 *
	 * @return The configurator entry
	 */
	ConfiguratorEntry createConfiguratorEntry();

	/**
	 * Creates a configurator entry array object which is to be returned from
	 * a logic element. Allows dynamic setup of the configurator entries.
	 * 
	 * @param entries
	 *            0..N configurator entries. Supports also list nesting.
	 * @return The configurator entry array object with defined methods for
	 *         further anrichment of the array as
	 *         {@code ConfiguratorEntryArray#addEntry(ConfiguratorEntry)},
	 *         {@code ConfiguratorEntryArray#setEntries(List)} or
	 *         {@code ConfiguratorEntryArray#getEntries()}
	 */
	ConfiguratorEntryArray createConfiguratorEntryArray(Object ... entries);
	
	/**
	 *
	 *
	 * @param configuratorName The name of the configurator, similar to other options' names
	 * @param formulaName      The name of the configurator formula which is used to drive the input selections
	 * @return The result of the configurator
	 */
	Object configurator(String configuratorName, String formulaName);

	/**
	 *
	 *
	 * @param configuratorName The name of the configurator, similar to other options' names
	 * @param formulaName      The name of the configurator formula which is used to drive the input selections
	 * @param resultName	   The result that should be returned (instead of the full configurator results map)
	 * @return The result of the configurator
	 */
	Object configurator(String configuratorName, String formulaName, Object resultName);
	
	/**
	 *
	 *
	 * @param configuratorName The name of the configurator, similar to other options' names
	 * @param formulaName      The name of the configurator formula which is used to drive the input selections
	 * @param width			   The width of the configurator window. Can be a number (in pixels) or a percentage string.
	 * @param height		   The height of the configurator window. Can be a number (in pixels) or a percentage string.
	 * @return The result of the configurator
	 */
	Object configurator(String configuratorName, String formulaName,Object width, Object height);
	
	/**
	 *
	 *
	 * @param configuratorName The name of the configurator, similar to other options' names
	 * @param formulaName      The name of the configurator formula which is used to drive the input selections
	 * @return The result of the configurator
	 */
	Object inlineConfigurator(String configuratorName, String formulaName);

	/**
	 * Parses a date (entered as a String) according to the given pattern.
	 * <p>
	 * Hint: For parsing datetimes (i.e. dates also including information about the time of the day), use the {@link #parseDateTime(String, String)} method instead.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def date = api.parseDate("yyyy-MM-dd", "2019-01-23")
	 * </pre>
	 * </blockquote>
	 *
	 * @param pattern The pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html">DateTimeFormat</a> for pattern format specification)
	 * @param date    The date as a String in the format specified by the {@code pattern}
	 * @return The date
	 * @see #parseDateTime(String, String)
	 * @see #parseDateTime(String, String, int, int)
	 */
	Date parseDate(String pattern, String date);

	/**
	 * Parses a datetime (entered as a String) according to the given pattern.
	 * <p>
	 * Important note: This method will use the UTC time zone for parsing (unless the provided {@code pattern} uses a syntax explicitly
	 * specifying a time zone). If a different time zone is required, use the {@link #parseDateTime(String, String, int, int)} method which allows to
	 * set a time zone by providing an offset to UTC in hours and minutes.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def now = api.parseDateTime("yyyy-MM-dd'T'HH:mm:ss", "2019-01-23T13:44:52")
	 * now.getYear()            // 2019
	 * now.getMonthOfYear()     // 1
	 * now.getDayOfMonth()      // 23
	 * now.getHourOfDay()       // 13
	 * now.getMinuteOfHour()    // 44
	 * now.getSecondOfMinute()  // 52
	 * </pre>
	 * </blockquote>
	 *
	 * @param pattern  The pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html">DateTimeFormat</a> for pattern format specification)
	 * @param datetime The datetime as a String in the format specified by the {@code pattern}
	 * @return The datetime
	 * @see #parseDate(String, String)
	 * @see #parseDateTime(String, String, int, int)
	 */
	DateTime parseDateTime(String pattern, String datetime);

	/**
	 * Parses a datetime (entered as a String) according to a given pattern and time zone.
	 * <p>
	 * Important note: This method will always use the provided time zone, regardless of the possible time zone set explicitly in {@code datetime}.
	 *
	 * @param pattern         The pattern for parsing (see <a href="http://www.joda.org/joda-time/apidocs/index.html?org/joda/time/format/DateTimeFormat.html">DateTimeFormat</a> for pattern format specification)
	 * @param datetime        The datetime as a String in the format specified by the {@code pattern}
	 * @param tzHoursOffset   The offset in hours from UTC, from -23 to +23
	 * @param tzMinutesOffset The offset in minutes from UTC, from -59 to +59
	 * @return The datetime
	 * @see #parseDateTime(String, String)
	 * @see #parseDate(String, String)
	 */
	DateTime parseDateTime(String pattern, String datetime, int tzHoursOffset, int tzMinutesOffset);

	/**
	 * Creates a cross-list filter expression based on meta data.
	 * Example:
	 * You need to filter across multiple price lists on items that have the value "A" in the column with the elementName "XYZ".
	 * Now the elementName XYZ might be mapped to a different attributeXX column for every list.
	 * So a simple attributeXX=A filter does not work.
	 * This method does a metadata lookup and creates a proper filter clause based on it for every list.
	 * For every list this translates into something like:
	 * (attributeXX = A AND listId = &lt;id&gt;)
	 * These filters are ORed together.
	 *
	 * @param elementName  The element name (originates from the logic element name or MPL integration tag)
	 * @param fieldValue   The value (= operator applied) of that attribute
	 * @param listTypedIds A list of typedId of the "header" objects. E.g. 34.PL. Supported type codes are PL, MPL, PG and SIM
     * @return The filter object
	 */
	Filter createElementNameFilter(String elementName, Object fieldValue, String... listTypedIds);

	/**
	 * Gets the current complete calculation results of the given item (e.g., price list item).
	 * <p>
	 * Note: These are not returned when using other methods (e.g., {@link #pricelistItem(String, String)}).
	 *
	 * @param typedId The typedId of the item, e.g. 34.PLI. Supported type codes are PLI, XPLI, SIMI, XSIMI, PGI, XPGI and RR
	 * @return The map of individual calculation results with 'Element Name' as the key
	 */
	Map<String, Object> getItemCompleteCalculationResults(String typedId);

	/**
	 * Gets the current complete active (i.e. approved) calculation results of the given item.
	 * @param typedId The typedId of the item, e.g. 34.PLI. Supported type codes are PGI and XPGI
	 * @return The map of individual calculation results with 'Element Name' as the key
	 */
	Map<String, Object> getItemActiveCalculationResults(String typedId);

	/**
	 * Issues an HTTP call to an external entity. Defaulting to POST request and JSON content type.
     * If called from a logic element, it has the same timeout as defined for that element.
	 *
	 * @param url The full URL to the HTTPS or HTTP endpoint
	 * @param body The body of the request
     * @see #httpCall(String, String, String, String, Map[])
     * @return A map containing two elements 'responseBody' and 'errorCode'. If the content type is JSON and data is returned, responseBody will actually also be a parsed Map and not the raw data.
	 * @throws XExpression If the HTTP call fails
	 */
	Object httpCall(String url, String body) throws XExpression;

	/**
	 * Triggers a HTTPS request towards the Pricefx server, either local one or an external one.
     * If this method is called from a logic element, it has the same timeout as defined for that element.
	 * To be able to use this function, you need to set up a bound partition first in Configuration > All Modules > Bound Partitions.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def items = ...
	 * def requestBody = []
	 * items.each { item ->
	 *
	 *     def record = [
	 *         "name": "Increase",
	 *         "sku": item.sku,
	 *         "attribute1": item.newValue,
	 *     ]
	 *
	 *     requestBody.add([
	 *         data: record,
	 *         oldValues: [:]
	 *     ])
	 * }
	 * def resp = api.boundCall("localhost", "/integrate/PX/batch", api.jsonEncode(requestBody), true)
	 * }</pre>
	 * </blockquote>
	 *
	 * @param uniqueName            The unique name of the bound partition
	 * @param relativeUrl           The relative URL to the Pricefx server command. First slash excluded.
	 * @param body                  The body of the request
	 * @param responseBodyAsAString If the first param is true, the body is not converted to JSON but returned as a String
	 * @return A map containing two elements 'responseBody' and 'errorCode'
	 * @throws XExpression If the bound partition is not set or the call fails
	 * @see <a href="https://qa.pricefx.eu/pricefx-api/json/index.html">Pricefx backend API</a>
	 */
	Object boundCall(String uniqueName, String relativeUrl, String body, Boolean... responseBodyAsAString) throws XExpression;

	/**
	 * Returns a bound partition list.
	 *
	 * @return The bound partition unique names
	 */
	Object getBoundPartitionNames();

	/**
	 * Issues an HTTP call to an external entity.
     * If called from a logic element, it has the same timeout as defined for that element.
	 *
	 * @param url         The full URL to the HTTPS or HTTP endpoint
	 * @param body        The body of the request
	 * @param methodType  {@code "GET"} or {@code "POST"}
	 * @param contentType <ul>
	 *     				  <li>{@code "JSON"} - JSON response will be deserialized into map </li>
	 * 					  <li>{@code "APPLICATION_XML"} (or {@code "XML"} (deprecated)) for contentType {@code "application/xml"}</li>
	 * 					  <li>{@code "TEXT_XML"} for contentType {@code "text/xml"}</li>
	 * 					  <li>or any full content type, for instance {@code "text/plain"}</li>
	 * 					  </ul>
	 * 					  For XML content types, the response is
	 * 					  <a ="http://docs.groovy-lang.org/latest/html/api/groovy/util/slurpersupport/GPathResult.html" and
	 * 					  <a href="https://www.groovy-lang.org/processing-xml.html#_gpath">GPath</a> expressions can be used to work with it.
	 *
	 * @param properties  The map of properties. The following are allowed:
	 *                    <ul>
	 *                    <li>basicAuthenticationUserName {@code String} value - User name for basic authentication</li>
	 *                    <li>basicAuthenticationPassword {@code String} value - User password for basic authentication</li>
	 *                    <li>sslTrustAllCertificates {@code boolean} value - Ignores invalid certificates. Use for testing HTTPS connections only. Not for production use.</li>
	 *                    <li>proxyAuthenticationUserName {@code String} value - User name for proxy (if any)</li>
	 *                    <li>proxyAuthenticationPassword {@code String} value - User password for proxy (if any)</li>
	 *                    <li>proxyUrl {@code String} value - Full URL to proxy</li>
	 *                    <li>additionalHeaders {@code Map} value - Sub-map that will then add these headers verbatim to the request</li>
	 *                    </ul>
	 * @return A map containing two elements 'responseBody' and 'statusCode'
	 * @throws XExpression If the HTTP call fails
	 */
	Object httpCall(String url, String body, String methodType, String contentType, @SuppressWarnings("rawtypes") Map... properties) throws XExpression;

	/**
	 * Adds a new object to the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method cannot be used in distributed calculation.
	 * <ul>
	 *     <li>CFS</li>
	 *     <li>CalculationFlow</li>
	 *     <li>direct formula execution via JSON API</li>
	 * </ul>
	 * If any of the attributes is null, the record will not be updated.
	 * If you want a certain attribute to have a null value, do it explicitly, by putting e.g. the “NULL” marker value there or a similar one.
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def entry = [
	 *         lookupTableName : "ExchangeRates",
	 *         key1 : "USD",
	 *         key2 : "EUR",
	 *         attribute1 : 0.879
	 * ]
	 * api.add("MLTV2", entry)
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param values   The key-value pairs representing fields of the object you want to add. Note that all significant fields have to be present.
	 * @return The added object or {@code null} if error occurred
	 * @see #addOrUpdate(String, Map)
	 * @see #update(String, Map)
	 * @see #delete(String, Map)
	 */
	Object add(String typeCode, Map<String, Object> values);

	/**
	 * Adds or updates an object in the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method also works with the Post Step Logic Formula in workflows. This method cannot be used in distributed calculation.
	 *
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def entry = [
	 *         lookupTableName : "ExchangeRates",
	 *         key1 : "USD",
	 *         key2 : "EUR",
	 *         attribute1 : 0.879
	 * ]
	 * api.add("MLTV2", entry)
	 * </pre>
	 * </blockquote>
	 *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param values   The key-value pairs representing fields of the object you want to add or update. Note that all significant fields have to be present.
	 * @return The object added or updated
	 * @see #add(String, Map)
	 * @see #update(String, Map)
	 * @see #delete(String, Map)
	 */
	Object addOrUpdate(String typeCode, Map<String, Object> values);

	/**
	 * Updates an object in the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method cannot be used in distributed calculation.
	 * <p>
	 * Example 1:
     * <blockquote>
	 * <pre>
	 * api.addOrUpdate("PX", [
	 * 		"name": "Scoring_MM",
	 * 		"sku": sku
	 * ])
	 * </pre>
	 * </blockquote>
     *
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * api.addOrUpdate("PGI", [
	 * 		"priceGridName" : "LifeCycleMonitor",
	 * 		"sku"           : currentItem.sku,
	 * 		"label"         : currentItem.label
	 * ])
	 * </pre>
	 * </blockquote>
	 *
	 * Example 3:
	 * <blockquote>
	 * <pre>
	 * def large = [
	 * 		"lookupTableId"   : pp.id,
	 * 		"lookupTableName" : pp.uniqueName,
	 * 		"name"            : "L",
	 * 		"attribute1"      : percentileHigh,
	 * 		"attribute2"      : "999999999"
	 * ]
     * api.addOrUpdate("MLTV", large)
	 * </pre>
	 * </blockquote>
     *
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param values   The key-value pairs representing fields of the object you want to update. Note that all significant fields have to be present.
	 * @return The updated object or {@code null} if error occured
	 * @see #addOrUpdate(String, Map)
	 * @see #add(String, Map)
	 * @see #delete(String, Map)
	 */
	Object update(String typeCode, Map<String, Object> values);

	/**
	 * Deletes an object in the collection of objects of the given type.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
     * This method cannot be used in distributed calculation.
	 * <p>
     * Example:
     * <blockquote>
	 * <pre>
	 * api.delete("PX", ["id": existingRecord, "name": "ActivePrice"])
	 * </pre>
	 * </blockquote>
	 * @param typeCode A type code string of the type of the object to search for. All available type codes can be retrieved by calling {@code <hostname>/pricefx/<partition name>/fetch}
	 * @param values   The key-value pairs representing significant fields of the object you want to delete. Note that all significant fields have to be present, if not the ID has to be present
	 * @return The deleted object or {@code null} if error occured
	 * @see #addOrUpdate(String, Map)
	 * @see #add(String, Map)
	 * @see #update(String, Map)
	 */
	Object delete(String typeCode, Map<String, Object> values);

	/**
	 * Sends an email. Only one recipient can be defined.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def message = "The cost is missing"
     * def filter = Filter.equals("name", "Costs")
	 *
	 * api.findLookupTableValues("EmailNotifications", filter).each {
	 *     api.sendEmail(it.value, "Missing costs", message)
	 * }
	 * </pre>
	 * </blockquote>
	 *
	 * @param to      The email address the email is sent to
	 * @param subject The email subject
	 * @param message The email body
	 */
	void sendEmail(String to, String subject, String message);

	/**
	 * Sends an email. Only one recipient can be defined.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * <p>
 	 *
	 * @param toEmail		The email address the email is sent to
	 * @param toName		The name of the recipient. If not specified then toEmail is used.
	 * @param subject		The email subject
	 * @param message		The email body
	 * @param fromEmail 	The sender email address
	 * @param fromName		The name of the sender. If not specified then fromEmail is used.
	 * @param replyToEmail	The reply to email address
	 * @param replyToName	The reply to user name. If not specified then replyToEmail is used.
	 * @param useCustomSMTP	True if should be send through custom SMTP server.
	 */
	void sendEmail(String to, String toName, String subject, String message, String fromEmail, String fromName, String replyToEmail, String replyToName, Boolean useCustomSMTP);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CF and direct formula execution via JSON API).
	 * To enable sending of this type of events, this needs to be enabled on the Partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object The object to pass along with the event
	 */
	void customEvent(Object object);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * To enable sending of this type of events, this needs to be enabled on the Partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object          The object to pass along with the event
	 * @param customEventType A custom string to classify the custom event type (as the general event type is always CUSTOM)
	 */
	void customEvent(Object object, String customEventType);

	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * To enable sending of this type of events, this needs to be enabled on the Partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object          The object to pass along with the event
	 * @param customEventType A custom string to classify the custom event type (as the general event type is always CUSTOM)
	 * @param operation       One of these strings: UPDATE, DELETE, ADD, ITEMCHANGE
	 */
	void customEvent(Object object, String customEventType, String operation);
	
	/**
	 * Sends an event of the type "CUSTOM". Other event types are reserved internally.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 * To enable sending of this type of events, this needs to be enabled on the Partition by the system admin (event type "CUSTOM" in the Event Bitmask).
	 *
	 * @param object          The object to pass along with the event
	 * @param customEventType A custom string to classify the custom event type (as the general event type is always CUSTOM). Can be null.
	 * @param operation       One of these strings: UPDATE, DELETE, ADD, ITEMCHANGE. Can be null.
	 * @param omitNullValues  Also includes keys that have a null value (false) or omits them (true). True is the default for other method signatures that do not explicitly specify that value.
	 */	
	void customEvent(Object object, String customEventType, String operation, boolean omitNullValues);

	/**
	 * Updates the header information of a CalculableObject (e.g., a price grid, a price list, etc.).
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct formula execution via JSON API).
	 *
	 * @param calcObjTypeCode The type code of the object to update
	 * @param calcObjName     The (numeric) ID of the object to update
	 * @return The calculable Object Header
	 */
	Object updateCalculableObjectHeader(String calcObjTypeCode, String calcObjName);

	/**
	 * Creates an {@link AttributedResult} object to be returned as a result of an output element.
	 * <p>
	 * Since {@code AttributedResult} is configurable, it is used when it is not sufficient to only return a
	 * plain value, but it is also required to adjust formatting, e.g. color, underline, etc.
	 * <p>
	 * The usage is simple, instead of returning the output value, create a new {@code AttributedResult} instance like this:
	 * {@code api.attributedResult(plain value)}, customize it by calling its various methods, and return it finally.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def cost = api.getElement("Margin_pct")
	 * return api.attributedResult(cost)
	 * .withBackgroundColor(cost < 0.30 ? "red" : "#0101DF")
	 * .withSuffix(cost < 0.30 ? "!!!" : null)
	 * .withTextColor(cost < 0.30 ? "white" : null)
	 * .withTextDecoration(cost < 0.30 ? "underline" : null)
	 * }</pre>
	 * </blockquote>
	 *
	 * @param result The calculation result
	 * @return The AttributedResult object
	 * @see AttributedResult AttributedResult - for details on formatting
	 */
	AttributedResult attributedResult(Object result);

	/**
	 * Wraps a result to make sure the original result object appears in the JSON response.
	 * This way also more complex results like lists and maps can be exported as a logic element result.
	 *
	 * @param result The result
	 * @return A wrapped result object. Should only be used as a return statement of a logic element.
	 */
	WrappedResult wrap(Object result);

	/**
	 * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
	 * <pre>
	 * The individual cells can be styled via the use of:
	 * -  linkCell(Object value, String targetPage, String targetPageState)
	 * -  imageCell(Object imageUrl)
	 * -  styledCell(Object value, String textColor, String bgColor, String weight, String alignment)
	 * -  styledCell(Object value, String textColor, String bgColor, String weight)
	 * -  styledCell(Object value, String textColor, String bgColor)
	 * -  setColumnFormat(String column, FieldFormatType fft)
	 *
	 * Valid FieldFormatTypes are:
	 * NUMERIC, NUMERIC_LONG, MONEY, PERCENT, TEXT, MONEY_EUR, MONEY_USD, MONEY_GBP, MONEY_JPY, MONEY_CHF, MONEY_PLN, DATETIME, DATE, INTEGER
	 * </pre>
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def getTrafficColor(value) {
	 *     if (value <= 0) {
	 *         return "red"
	 *     } else if (value > 0 && value < 0.1) {
	 *         return "yellow"
	 *     } else if (value >= 0.1) {
	 *         return "green"
	 *     }
	 * }
	 *
	 * // initialize new matrix with three columns
	 * def matrix = api.newMatrix("Item", "Quantity", "Price", "Pricegrid", "Pricelist", "Margin Status", "Image", "CustomerId")
	 *
	 * // add more columns dynamically
	 * matrix.addColumn("Margin %")
	 * matrix.addColumn("Url")
	 *
	 * // set column formats
	 * matrix.setColumnFormat("Item", FieldFormatType.TEXT)
	 * matrix.setColumnFormat("Quantity", FieldFormatType.NUMERIC_LONG)
	 * matrix.setColumnFormat("Price", FieldFormatType.MONEY_EUR)
	 * matrix.setColumnFormat("Margin %", FieldFormatType.PERCENT)
	 * matrix.setColumnFormat("Url", FieldFormatType.LINK)
	 *
	 * // allow users to filter values
	 * matrix.setEnableClientFilter(true)
	 *
	 * // add data row
	 * matrix.addRow([
	 *         "Item": matrix.styledCell("Red bold text", "#ff0000", "transparent", "bold"),
	 *         "Quantity": 1.23456,
	 *         "Price": 78.9012,
	 *         "Margin %": 0.3456,
	 *
	 *         "Url": "<a href=\"https://www.google.com/search?q=price+f(x)\">link</a>",
	 *
	 *         // optionally you can add link to a price grid of a given ID
	 *         "Pricegrid" : matrix.linkToPriceGrid("Open Pricegrid", 123, null),
	 *
	 *         // or to a price list of a given ID
	 *         "Pricelist" : matrix.linkToPriceList("Open Pricelist", 123, null),
	 *
	 *         // or a to custom page (result identical to the previous)
	 *         "CustomerId" : matrix.linkCell("Open Customer", "customersPage", "123456"),
	 *
	 *         // or add a library images, such as Traffic, BlackTraffic or Arrow
	 *         "Margin Status" : matrix.libraryImage("BlackTraffic", getTrafficColor(-0.2f)),
	 *
	 *         // or add a general image
	 *         "Image" : matrix.imageCell("images/grid/approve.png"),
	 * ])
	 *
	 *
	 * return matrix
	 * }</pre>
	 * </blockquote>
     *
	 * @return ResultMatrix object
	 * @see #newMatrix(String...)
	 */
	ResultMatrix newMatrix();

	/**
	 * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
	 * See the documentation in {@link #newMatrix()}.
	 *
	 * @param columns The initial set of columns for the result matrix
	 * @return The ResultMatrix object
	 * @see #newMatrix()
	 */
	ResultMatrix newMatrix(String... columns);

	/**
	 * Creates a new result matrix DTO object that can be further customized and populated with calculated data.
	 * See the documentation in {@link #newMatrix()}.
	 *
	 * @see #newMatrix(String...)
	 * @param columns The initial set of columns for the result matrix
	 * 
	 * @return The ResultMatrix object
	 * @see #newMatrix()
	 */
	ResultMatrix newMatrix(Collection<String> columns);

	/**
	 * Creates a new gauge object that can be further customized.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * def gauge = api.newGauge()
	 * gauge.addSector(5, "#FF0000")
	 * gauge.addSector(25, "#FFFF00")
	 * gauge.addSector(null, "#00FF00")
	 * gauge.setValue(10)
	 * </pre>
	 * </blockquote>
	 * @return the ResultGauge object
	 */
	ResultGauge newGauge();

	/**
	 * Creates a new dashboard controller object that can be further customized.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>{@code
	 * def lookupTable = api.findLookupTable("SalesOrg")
	 * def controller = api.newController()
	 * controller.addButton("1. STEP: Change Price Strategy", "pricingParametersPage", lookupTable.typedId)
	 * controller.addHTML("<h2>Meatball</h2>")
	 * }</pre>
	 * </blockquote>
	 *
	 * @return the DashboardController object
	 */
	DashboardController newController();

	/**
	 * Walks a filter and all its sub filters, visiting each filter in the tree.
	 * A pair of closures is used to visit each filter. Each closure can
	 * replace the Filter that is visiting. If it does, a new tree will
	 * be created for every part of the tree that is affected, thus preserving
	 * the original tree. {@code removeNulls} is consulted only in case of
	 * (sub)filters which take a list of sub-filters.
	 * <p>
	 * Hence, this method may be used in many cases. For instance, you
	 * can use it for mapping the current property set to another one, e.g. to convert
	 * a PB filter to a PA one. Similarly to this example:
	 * <blockquote>
	 * <pre>{@code
	 * def propertyMappings = ["sku": "ProductID"]
	 *
	 * api.walkFilter(pbFilter, null, { filter ->
	 *     if (filter != null && propertyMappings[filter.property] != null) {
	 *         filter.property = propertyMappings[filter.property]
	 *     }
	 *     filter
	 * }, false);
	 * }</pre>
	 * </blockquote>
	 *
	 * @param filter The filter object
	 * @param visitBefore The before closure
	 * @param visitAfter The after closure
	 * @param removeNulls Removes nulls
	 * @return If any changes have been made, the new Filter; if not, the
	 * original Filter.
	 */
	Filter walkFilter(Filter filter, Closure<?> visitBefore, Closure<?> visitAfter, boolean removeNulls);

	/**
	 * @param dataLoadLabel The label of DMDataLoad as seen in the Data Load UI
	 * @param dataLoadType  The load type of DMDataLoad as seen in the Data Load UI. Namely:
	 *                      <ul>
	 *                      <li>Customers</li>
	 *                      <li>Products</li>
	 *                      <li>Flush</li>
	 *                      <li>Calendar</li>
	 *                      <li>Refresh</li>
	 *                      <li>Calculation</li>
	 *                      <li>Truncate</li>
	 *                      <li>Simulation</li>
	 *                      <li>ModelCalculation</li>
	 *                      <li>InternalCopy</li>
	 *                      <li>IndexMaintenance</li>
	 *                      </ul>
	 * @param target        The target of DMDataLoad as seen in the Data Load UI
	 * @return The Data Load if and only if there is exactly one Data Load matching (fail early strategy)
	 */
	Object findDataLoad(String dataLoadLabel, String dataLoadType, String target);
	
	/**
	 * @param rebateRecordLoadLabel The label of the rebate record load as seen in the Rebate Record UI
	 * @return The rebateRecordLoad if and only if there is exactly one load matching (fail early strategy)
	 */
	Object findRebateRecordLoad(String rebateRecordLoadLabel);
	
	/**
	 * Returns the current partition name.
	 * @return The partition name
	 */
	String currentPartitionName();
	
	/**
	 * Provides a wrapper for org.joda.time.DateTimeZone.forId(String tzString) method.
	 * For details see their documentation.
	 * <p>
	 * Example 1:
	 * <blockquote>
	 * <pre>
	 * def timezone1 = api.getTimeZone("Europe/Berlin")
	 * def timezone2 = api.getTimeZone("Europe/Prague")
	 * def timezone3 = api.getTimeZone("America/New_York")
	 * def timezone4 = api.getTimeZone("Australia/Brisbane")
	 * </pre>
	 * </blockquote>
	 *
	 * <p>
	 * Example 2:
	 * <blockquote>
	 * <pre>
	 * def startHour = 20
	 * def offset = api.getTimeZone("Europe/Prague").getOffset(new Date().getTime())
	 * def startHourUTC = startHour - offset.intValue()
	 * </pre>
	 * </blockquote>
	 * @param tzString The timezone String
	 * @return The DateTimeZone object
	 * @see <a href="https://en.wikipedia.org/wiki/List_of_tz_database_time_zones">https://en.wikipedia.org/wiki/List_of_tz_database_time_zones</a>
	 */
	DateTimeZone getTimeZone(String tzString);
	
	/**
	 * Generates a random uuid of the specified length.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * return uuid(15)   // returns e.g. "VcydxgltxrVZSTV"
	 * </pre>
	 * </blockquote>
	 *
	 * @param len The desired number of characters
     * @return The uuid string
     */
	String uuid(int len);
	/**
	 * Generates a random uuid of the specified length and radix.
	 * <p>
	 * Examples:
	 * <blockquote>
	 * <pre>
	 * return uuid(8, 2) // returns e.g. "01001010" (8 character ID, base=2)
	 *
	 * return uuid(8, 10)       // returns e.g. "47473046" (8 character ID, base=10)
	 *
	 * return uuid(8, 16)       // returns e.g. "098F4D35" (8 character ID, base=16)
	 * </pre>
	 * </blockquote>
	 *
	 * @param len The desired number of characters
	 * @param radix The number of allowable values for each character (must be &lt;=62)
     * @return The uuid string
	 */
	String uuid(int len, int radix);
	
	/**
	 * Generates a RFC4122, version 4 ID.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * return api.uuid()   // returns e.g. "92329D39-6F5C-4520-ABFC-AAB64544E172"
	 * </pre>
	 * </blockquote>
	 *
	 * @return The uuid string
	 */
	String uuid();

	/**
	 * Constructs a new composite key object which can be used as a key in a map.
	 * Equality of two keys is defined by the pair-wise equality of the corresponding key values.
	 *   
	 * @param keys Key values
	 * @return MultiKey composed of the passed ordered key values
	 */
	Object multiKey(Object... keys);

	/**
	 * Lists all roles available in the application.
	 * @return Map of [roleUniqueName : roleLabel] pairs
     */
	Map<String, String> roles();

	/**
	 * Creates an entityRef value (usually db ID) for the given entity. The value can be stored in attributeXX fields
	 * and you can set its meta type to ENTITY REFERENCE if you want the UI to show the data as links.
	 *
     * @param po The object of a typedCode P, C, DCR, Q, CT, RBA
	 * @return The string that represents the linkable state of the entity
     * @throws IllegalAccessException If {@code po} is of a different typeCode
	 */
	String entityRef(Object po) throws IllegalAccessException;

	/**
	 * Evaluates an arithmetic expression that is passed as a String.
	 *
	 * Supported operations are:
	 * <ul>
	 *     <li>addition {@code +}</li>
	 *     <li>subtraction {@code -}</li>
	 *     <li>multiplication {@code *}</li>
	 *     <li>division {@code /}</li>
	 *     <li>exponentiation {@code ^}</li>
	 *     <li>square root {@code sqrt}</li>
	 * </ul>
	 *
	 * Operator precedence and associativity rules are respected.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * evalExpression("((10 - 3^2 + 1) * -sqrt(1 * 2 + 3 * 4)) / 4")
	 * </pre>
	 * </blockquote>
	 *
	 * @param mathExpression The mathematical expression
	 * @return The result of the expression as double
	 */
	double evalExpression(String mathExpression);

	/**
	 * Stores a string value in the shared cache with the given key. Note the "global" nature of the key.
	 * 
	 * @param key The cache key
	 * @param value The value to store
	 * @see #getSharedCache(String)
	 * @see #removeSharedCache(String)
	 */
	void setSharedCache(String key, String value);
	
	/**
	 * Retrieves a value from the shared cache (and resets the TTL of the key).
	 * 
	 * @param key The cache key
	 * 
	 * @return The value or {@code null} found in the shared cache
	 * @see #getSharedCache(String)
	 */
	String getSharedCache(String key);
	
	/**
	 * Deletes a key from the shared cache.
	 * 
	 * @param key The cache key
	 * @see #setSharedCache(String, String)
	 */
	void removeSharedCache(String key);

	/**
	 * Provides a standard Base64 encoding of a string.
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.base64Encode("Price f(x)") // "UHJpY2UgZih4KQ=="
	 * </pre>
	 * </blockquote>
	 *
	 * @param string The plain text to be encoded
	 * @return The base64 encoded string
	 * @see #base64Decode(String)
	 * @see <a href="https://en.wikipedia.org/wiki/Base64">https://en.wikipedia.org/wiki/Base64</a>
	 */
	String base64Encode(String string);

	/**
	 * Decodes a base64 encoded string.
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.base64Decode("UHJpY2UgZih4KQ==") // "Price f(x)"
	 * </pre>
	 * </blockquote>
	 * @param base64String The base64 encoded text
	 * @return The plain text string
	 * @see #base64Encode(String)
	 * @see <a href="https://en.wikipedia.org/wiki/Base64">https://en.wikipedia.org/wiki/Base64</a>
	 */
	String base64Decode(String base64String);

	/**
	 * Returns a list of product reference records that tell which other objects, of the given type, reference these SKUs.
	 *
	 * @param typeCode Currently, only "Q" or "RBA" are supported
	 * @param skus A set of SKU strings that should be searched for
	 * @return A list of reference records
	 */
	List<?> getProductReferences(String typeCode, Set<String> skus);

	/**
	 * Returns a list of SKUs based on the definition of the product group
	 * @param pg The product group object
	 * @return List of SKU
	 */
	List<String> getSkusFromProductGroup(ProductGroup pg);

	/**
	 * Returns a list of Customer IDs based on the definition of the customer group
	 * @param cg The customer group object
	 * @return List of customer IDs
	 */
	List<String> getCustomerIdsFromCustomerGroup(CustomerGroup cg);

	/**
	 * Returns the language part of the current request's locale (e.g. "en").
	 *
	 * @return The language part string
	 */
	String getLocale();
	
	/**
	 * Returns the base URL component of the current instance. Example: https://www.pricefx.eu
	 *
	 * <p>
	 * Example:
	 * <blockquote>
	 * <pre>
	 * api.getBaseURL() // "https://www.pricefx.eu"
	 * </pre>
	 * </blockquote>
	 *
	 * @return The base URL
	 */
	String getBaseURL();

	CreationWorkflowDefinition newCreationWorkflow();

	CreationWorkflowStepDefinition newCreationWorkflowStep();

	/**
	 * Finds the next revision uniqueNames for the given Quotes/Contracts/RebateAgreements.
	 * @param typeCode Can be one of: 'Q', 'CT', 'RBA'
	 * @param uniqueNames List of uniqueNames
	 * @return Map of [uniqueName : List of nextRevUNs]
	 */
	Map<String, List<String>> findNextRevUNs(String typeCode, String... uniqueNames);

	/**
	 * Gives current item context in claim validation
	 *
	 * @return Claim Context instance
	 */
	Object getClaimContext();

	/**
	 * Indicates whether the current calculation is recalculating the entire list or only parts of it.
	 * <p>
	 * Note: Currently only applicable for price grid calculations.
	 * 
	 * @return True if the entire grid is calculated, false otherwise. {@code null} is returned if method is not applicable
	 */
	Boolean isFullListCalculation();

	/**
	 * Creates a new background calculation job for the specified price grid.
	 * Either a full calculation (parameters == null) or a partial calculation (parameter entry "skusToRecalc" as a list of SKU strings).
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct logic execution via JSON API).
	 * @param priceGridId The ID of the price grid to calculate
	 * @param parameters Optional job parameters
	 * @return The ID of the created job tracker
	 */
	Long triggerPriceGridCalculation(Long priceGridId, Map<String, Object> parameters);

	/**
	 * Creates a new background calculation job for the specified CFS.
	 * <p>
	 * Note: This operation will only work in contexts that allow object modification (CFS, CalculationFlow and direct logic execution via JSON API).
	 * @param cfsID The ID of CalculatedFieldSet
	 * @return The ID of the created job tracker 
	 */
	Long triggerCFSCalculation(Long cfsID);

	/**
	 * Provides a Map of current execution context info variables.
	 * <p>
	 * Currently supported variables are:
	 * <ul>
	 *     <li>
	 *         <b>commandName</b> Command name (if any) that was used to invoke current logic execution, e.g. "addproducts", "save", "price", "submit", "calculate"
	 *     </li>
	 * </ul>
	 *
	 *
	 * @return the context map
	 */
	Map currentContext();

	/**
	 * Resets specified fields on currentItem object to null. Applies only to price grid items. Fields that can not be found on the item are skipped.
	 *
	 * @param fields field names that should be reset
	 */
	void resetCurrentItemFields(String... fields);

	/**
	 * Evaluates product(group) filter logic
	 *
	 */
	List<Filter> evalProductFilterLogic(String logicName, String logicParam);

	/**
	 * Evaluates customer(group) filter logic
	 *
	 */
	List<Filter> evalCustomerFilterLogic(String logicName, String logicParam);


	/**
	 * Creates or updates user input attribute on LPG.
	 * <p>
	 * <b>Note:</b> Input attributes are saved to extension columns (like attribute12 for example) - same as output elements.
	 * Please consider this so that you do not accidentally overwrite any existing output element.
	 *
	 * @param fieldName Name of field to be used for storing values, e.g. attribute12
	 * @param elementName Name of the column, e.g. colour
	 * @param label Label of the column, e.g. Colour
	 * @param priceGridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
	 * @return Created or updated attribute
	 */
	Object createOrUpdatePricegridInputAttribute(String fieldName, String elementName, String label, Long priceGridId);

	/**
	 * Deletes a pricegrid input attribute.
	 *
	 * @param id Id of the input attribute to delete
	 * @return The deleted attribute object or null if no attrbiute was found
	 */
	Object removePricegridInputAttribute(Long id);

	/**
	 * Creates or updates user input attribute on PL.
	 * <p>
	 * <b>Note:</b> Input attributes are saved to extension columns (like attribute12 for example) - same as output elements.
	 * Please consider this so that you do not accidentally overwrite any existing output element.
	 *
	 * @param fieldName Name of field to be used for storing values, e.g. attribute12
	 * @param elementName Name of the column, e.g. colour
	 * @param label Label of the column, e.g. Colour
	 * @param priceGridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
	 * @return Created or updated attribute
	 */
	Object createOrUpdatePricelistInputAttribute(String fieldName, String elementName, String label, Long pricelistId);

	/**
	 * Deletes a pricelist input attribute.
	 *
	 * @param id Id of the input attribute to delete
	 * @return The deleted attribute object or null if no attrbiute was found
	 */
	Object removePricelistInputAttribute(Long id);

	/**
	 * Sets PG calculation output that will be shown in the header section after PG calculation.
	 *
	 * @param pricegridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
	 * @param name name of the calculation output, e.g. sales
	 * @param label	label of the calculation output, e.g. Global Sales
	 * @param value calculated value, e.g. €2.3M
	 * @param message message containing additional information
	 * @return saved calculation output
	 */
	Object setPricegridCalculationOutput(Long pricegridId, String name, String label, String value, String message);

	/**
	 * Removes PG calculation output.
	 *
	 * @param pricegridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
	 * @param name name of the calculation output that will be removed
	 */
	void removePricegridCalculationOutput(Long pricegridId, String name);

	/**
	 * Chart definition that will be used to construct and display a Highchart in the header section.
	 *
	 * @param definition highchart definition as a map
	 * @param pricegridId pricegrid id which can usually be obtained by using {@code api.currentItem("id")}
	 * @return saved chart definition
	 */
	Object setPricegridCalculationChart(final Map<String, ?> definition, Long pricegridId);

	/**
	 * Sets PL calculation output that will be shown in the header section after PL calculation.
	 *
	 * @param pricelistId pricelist id which can usually be obtained by using {@code api.currentItem("id")}
	 * @param name name of the calculation output, e.g. sales
	 * @param label	label of the calculation output, e.g. Global Sales
	 * @param value calculated value, e.g. €2.3M
	 * @param message message containing additional information
	 * @return saved calculation output
	 */
	Object setPricelistCalculationOutput(Long pricelistId, String name, String label, String value, String message);

	/**
	 * Removes PL calculation output.
	 *
	 * @param pricelistId pricelist id which can usually be obtained by using {@code api.currentItem("id")}
	 * @param name name of the calculation output that will be removed
	 */
	void removePricelistCalculationOutput(Long pricelistId, String name);

	/**
	 * Chart definition that will be used to construct and display a Highchart in the header section.
	 *
	 * @param definition highchart definition as a map
	 * @param pricelistId pricelist id which can usually be obtained by using {@code api.currentItem("id")}
	 * @return saved chart definition
	 */
	Object setPricelistCalculationChart(final Map<String, ?> definition, Long pricegridId);

	InputBuilderFactory inputBuilderFactory();
}
