package net.pricefx.lmformulaengine;

import com.googlecode.genericdao.search.Filter;
import groovy.lang.Binding;
import groovy.lang.GroovyShell;
import groovy.lang.Script;
import net.pricefx.common.api.FieldFormatType;
import net.pricefx.common.api.InputButtonAction;
import net.pricefx.common.api.InputType;
import net.pricefx.common.api.pa.DataType;
import net.pricefx.common.apibuilder.BuilderExtensions;
import net.pricefx.common.apibuilder.BuilderExtensions.GroovyDefaultImport;
import net.pricefx.common.apibuilder.quote.QuoteStructure;
import net.pricefx.common.util.CalendarUtil.TimePeriod;
import net.pricefx.common.util.CalendarUtil.TimeUnit;
import net.pricefx.domain.CustomerGroup;
import net.pricefx.domain.ProductGroup;
import net.pricefx.domain.TimedObjectWithStatus;
import net.pricefx.formulaengine.FormulaEngineResult;
import net.pricefx.formulaengine.PriceFxException;
import net.pricefx.formulaengine.StackTraceUtil;
import net.pricefx.formulaengine.XExpression;
import net.pricefx.formulaengine.scripting.LibraryMethod;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.lmformulaengine.model.LibraryBindingMap;
import net.pricefx.lmformulaengine.scripting.LmSandboxAPI;
import net.pricefx.logic.core.AbstractLogicSerializer;
import net.pricefx.logic.core.LogicException;
import net.pricefx.logic.core.LogicSerializer;
import net.pricefx.logic.core.PfxClientBuilder;
import net.pricefx.logic.core.config.Partition;
import net.pricefx.logic.core.metadata.MetadataCache;
import net.pricefx.logic.core.metadata.MetadataProvider;
import net.pricefx.logic.core.model.AbstractFormula;
import net.pricefx.logic.core.model.AbstractFormulaElement;
import net.pricefx.logic.core.model.ContextParameter;
import net.pricefx.logic.core.model.ElementCombinationType;
import net.pricefx.logic.core.model.FormulaElementResult;
import net.pricefx.logic.core.model.PricingFormula;
import net.pricefx.logic.core.model.PricingFormulaElement;
import net.pricefx.logic.core.model.WorkflowFormula;
import net.pricefx.pckg.processing.ProcessingContextImpl;
import net.pricefx.pckg.processing.ProcessingException;
import net.pricefx.pckg.transform.TransformCalculationLogic;
import net.pricefx.pckg.transform.TransformWorkflowFormula;
import net.pricefx.pckg.transform.descriptor.TypeDescriptor;
import net.pricefx.server.dto.TieredValue;
import net.pricefx.server.dto.calculation.AttributedResult;
import net.pricefx.server.dto.calculation.ConfiguratorEntry;
import net.pricefx.server.dto.calculation.ResultMatrix;
import org.apache.commons.lang3.StringUtils;
import org.codehaus.groovy.control.CompilerConfiguration;
import org.codehaus.groovy.control.customizers.ImportCustomizer;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.io.FileReader;
import java.io.Reader;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.TimeoutException;

import static net.pricefx.common.api.PriceFxInterface.CONTEXTPARAM_NAME_PRODUCTSKU;
import static net.pricefx.common.api.PriceFxInterface.FIELD_FE_CRITICALALERT;
import static net.pricefx.common.api.PriceFxInterface.FIELD_FE_REDALERT;
import static net.pricefx.common.api.PriceFxInterface.FIELD_FE_YELLOWALERT;
import static net.pricefx.common.api.PriceFxInterface.GROOVY_LIBRARY_UNIQUENAME;
import static net.pricefx.common.api.PriceFxInterface.PARAM_TARGETDATE;


public class LmDefaultFormulaEngine {

    private static final Logger log = LoggerFactory.getLogger(LmDefaultFormulaEngine.class);
    private static final Logger config_log = LoggerFactory.getLogger("configuration.Formula");
    private static final Marker LOG_MARKER = MarkerFactory.getMarker("FORMULAENGINE");
    private static final Long defaultTimeout = 5L;
    private static final Long maxTimeout = 900L;
    private static final Long maxLibTimeout = 120L;

    // groovy binding's variable names
    public static final String VAR_lib = "lib";
    public static final String VAR_libs = "libs";
    public static final String VAR_api = "api";
    private static final String VAR_in = "in";
    private static final String VAR_out = "out";

    //libs
    protected LibraryBindingMap groovyLibraries;

    private Partition partition;
    private AbstractFormula logic;
    private boolean syntaxCheckMode;
    private Path moduleDir;
    Map<String, Object> parameters;
    LmFormulaEngineContext ctx;
    GroovyShell groovyShell;
    LogicSerializer logicSerializer;

    public LmDefaultFormulaEngine(Partition partition, AbstractFormula logic, Path moduleDir, Map<String, Object> parameters, boolean syntaxCheckMode, AbstractLogicSerializer.CachedEntry cache) throws ProcessingException, LogicException {

        this.partition = partition;
        this.syntaxCheckMode = syntaxCheckMode;
        this.logic = logic;
        this.moduleDir = moduleDir;
        this.parameters = parameters;
        logicSerializer = new LogicSerializer(logic);
        groovyShell = new GroovyShell(this.getClass().getClassLoader(), getGroovyCompilerConfiguration());

        Object skuObject = parameters.get(CONTEXTPARAM_NAME_PRODUCTSKU);
        String sku = null;
        if (skuObject instanceof List) {
            sku = ((List) skuObject).get(0).toString();
        } else if (skuObject != null) {
            sku = skuObject.toString();
        }

        MetadataCache metadataCache = MetadataProvider.getInstance().getMetadataInstance(partition);
        AbstractLogicSerializer.convertParameters(cache, partition.getClient(), parameters, logic, metadataCache);
        ctx = new LmFormulaEngineContext(getTargetDate(), sku, logic.getFormulaNature());
        ctx.addParameters(parameters);
    }

    public LmDefaultFormulaEngine(Partition partition, Path logicDir, Path moduleDir, String paramsPresetName, boolean syntaxCheckMode) throws LogicException {
        this(
                partition,
                LogicSerializer.readOnDiskLogicWithElements(moduleDir, new ProcessingContextImpl(), moduleDir.relativize(logicDir), true),
                moduleDir,
                new HashMap<>(),
                syntaxCheckMode,
                new AbstractLogicSerializer.CachedEntry());

        parameters = loadParametersPreset(paramsPresetName);
        String sku = (String) parameters.get(CONTEXTPARAM_NAME_PRODUCTSKU);
        MetadataCache metadataCache = MetadataProvider.getInstance().getMetadataInstance(partition);
        AbstractLogicSerializer.convertParameters(new AbstractLogicSerializer.CachedEntry(), partition.getClient(), parameters, logic, metadataCache);

        ctx = new LmFormulaEngineContext(getTargetDate(), sku, logic.getFormulaNature());
        ctx.addParameters(parameters);
    }

    LmDefaultFormulaEngine(String env, String partitionName, String url, String username, String encryptedPassword, Path logicDir, String paramsPresetName) throws LogicException {
        this(
                Partition.createPartition(PfxClientBuilder.getInstance(), env, partitionName, url, username, encryptedPassword),
                logicDir,
                logicDir.getParent().getParent(),
                paramsPresetName,
                false);
    }

    LmDefaultFormulaEngine(String[] args) throws LogicException {
        this(args[0], args[1], args[2], args[3], args[4], Paths.get(args[5]), args.length >= 7 ? args[6] : null);
    }

    private Date getTargetDate() throws LogicException {
        try {
            final SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd");
            String targetDateString = (String) parameters.get(PARAM_TARGETDATE);
            if (StringUtils.isEmpty(targetDateString)) {
                Date today = new Date();
                parameters.put(PARAM_TARGETDATE, sdf.format(today));
                return today;
            } else {
                return sdf.parse(targetDateString);
            }
        } catch (ParseException e) {
            String msg = "Parameter targetDate is not in required format YYYY-MM-DD";
            log.error(msg);
            throw new LogicException(msg, e);
        }
    }

    public Map<String, Object> loadParametersPreset(String paramsPresetName) throws LogicException {
        Map<String, Object> result = new TreeMap();
        try {
            if (paramsPresetName != null) {
                List<ContextParameter> contextParameters = logicSerializer.readParamsFile(paramsPresetName);
                for (ContextParameter contextParameter : contextParameters) {
                    result.put(contextParameter.getName(), contextParameter.getValue());
                }
            }
        } catch (Exception e) {
            log.error(e.getMessage());
            throw new LogicException(e);
        }
        return result;
    }


    public FormulaEngineResult evaluateFormula() throws LogicException, XExpression {

        if (logic == null)
            return FormulaEngineResult.DATA_ERROR;

        ctx.setFormulaNature(logic.getFormulaNature());

        //we need to init the api also in case the pricing formula has no script elements itself. The lib could be called via regular syntax wrapper
        TypeDescriptor objectType;
        if (logic instanceof PricingFormula) {
            objectType = TransformCalculationLogic.DESCRIPTOR;
        } else if (logic instanceof WorkflowFormula) {
            objectType = TransformWorkflowFormula.DESCRIPTOR;
        } else {
            throw new IllegalArgumentException("Cannot decode objectType for a logic");
        }
        LmSandboxAPI api = new LmSandboxAPI(partition.getClient(), objectType, ctx);
        if (ctx.getGroovyGlobals() != null) {
            api.global = ctx.getGroovyGlobals();
        }

        Map<String, Script> previousScripts = new HashMap();
        initializeLibrary(api, ctx);

        Iterator<AbstractFormulaElement> it = logic.getElementsIterator();
        while (it.hasNext()) {
            AbstractFormulaElement element = it.next();
            evaluateElement(element, ctx, api, previousScripts);
        }


        //retain global variables over formula executions
        if (api != null) {
            if (api.retainGlobal)
                ctx.setGroovyGlobals(api.global);
            else
                ctx.setGroovyGlobals(null);

        }

        ctx.setFinalResult(ctx.getIntermediateResult());
        ctx.setSandboxAPI(null);

        return FormulaEngineResult.OK;
    }

    private FormulaElementResult evaluateElement(AbstractFormulaElement e, LmFormulaEngineContext ctx,
                                                 LmSandboxAPI api, Map<String,Script> previousScripts) throws XExpression {
        log.info(LOG_MARKER,"Element: {} ", e.getElementName());

        // element group needs to match the context group & the conditional element needs to evaluate to non-null/non-false/non-zero
        if (!isFormulaElementOperative(e, ctx)) {
            return null;
        }

        String elementName = null;

            try {
                elementName = e.getElementName();
                FormulaElementResult r = null;

                //reset ctx temporary sku assignments
                ctx.resetSKUContextSwitch();

                //refresh library classes to reset timeout timers
                refreshGroovyLibrary(api);

                // calculate sub result
                Object result = null;
                Object manualResult = null;

                if (e instanceof PricingFormulaElement)
                    manualResult = getManualOverrideResult(ctx, (PricingFormulaElement) e);

                if (ElementCombinationType.RESULT.name().equals(e.getCombinationType())) {

                    if(ctx.isAborted())
                        return r;

                    // if the element is a result itself
                    // ==> store the running intermediate result in there
                    if (manualResult != null) {
                        //we have a manual override; skip calculation for that
                        r = formulaElementResult(e, manualResult,true);
                        ctx.setIntermediateResult(manualResult);
                    } else {
                        result = ctx.getIntermediateResult();
                        // in syntax check mode we replace 0 results by a number to avoid subsequent DivBy0 exceptions
                        if(syntaxCheckMode) {
                            result = syntaxCheckModeResultFix(result);
                            log.debug(LOG_MARKER,"Intermediate result in syntaxCheckMode: {} ",result);
                        }
                        r = formulaElementResult(e, result);
                    }
                } else {

                }
                // if the element is a contributor or INFO
                // ==> calculate sub result

                if (ctx.isAborted())
                    return null;

                if (manualResult != null) {
                    //we have a manual override; skip calculation for that
                    //we have a manual override; skip calculation for that
                    if(manualResult instanceof BigDecimal || manualResult instanceof Integer || manualResult == null) {
                        result = manualResult;
                    } else if(manualResult instanceof Number) {
                        result = new BigDecimal(manualResult.toString());
                    } else if(manualResult instanceof String) {
                        result = manualResult;
                    } else {
                        result = manualResult;
                    }

                    r = formulaElementResult(e, result, true);
                } else {
                    if (ElementCombinationType.FUNCTION.name().equals(e.getCombinationType())) {
                        try {
                            if (ctx.isAborted())
                                return null;

                            // setting formula element timeout on every element execution

                            //1. the element class name must be different than variable name,
                            //   since class name has a priority for some reason
                            //2. hack: setting element classname with ".groovy" suffix makes this
                            //   clickable file link in the exception stacktrace
                            Script groovyObj;
                            groovyShell.getClassLoader().clearCache();
                            if (e.getFormulaExpression() == null) {
                                Reader in = new FileReader(logicSerializer.getElementPath(e).toFile());
                                groovyObj = groovyShell.parse(in, e.getElementName() + ".groovy").getClass().newInstance();
                            } else {
                                groovyObj = groovyShell.parse(e.getFormulaExpression())
                                        .getClass()
                                        .newInstance();
                            }

                            Binding binding = new Binding();

                            if (groovyLibraries != null) {
                                binding.setVariable(VAR_libs, groovyLibraries);
                                Map<String, Script> groovyLibraryBindings = groovyLibraries.get(GROOVY_LIBRARY_UNIQUENAME);
                                if (groovyLibraryBindings != null) binding.setVariable(VAR_lib, groovyLibraryBindings);
                            }

                            binding.setVariable(VAR_api, api);
                            binding.setVariable(VAR_in, ctx.getParametersClone());
                            binding.setVariable(VAR_out, ctx.getResultElementsResultMap());

                            if (ctx.getGroovyBindings() != null) {
                                for (Map.Entry<String, Object> entry : ctx.getGroovyBindings().entrySet())
                                    binding.setVariable(entry.getKey(), entry.getValue());
                            }

                            if (previousScripts != null) {
                                for(Map.Entry<String, Script> entry : previousScripts.entrySet())
                                    binding.setVariable(entry.getKey(), entry.getValue());
                            }
                            previousScripts.put(elementName, groovyObj);

                            groovyObj.setBinding(binding);

                            ctx.setElementTimeout(getEffectiveTimeout(e, GROOVY_LIBRARY_UNIQUENAME.equals(logic.getUniqueName())));
                            result = groovyObj.run();

                            ctx.setElementTimeout(null);

                        } catch (Exception ex) {
                            //catch all exceptions here from scripts
                            if(ex instanceof XExpression)
                                throw (XExpression)ex; //pass through XExpression exceptions
                            else if(ex instanceof TimeoutException) {
                                config_log.warn(LOG_MARKER,"Caught groovy script TIMEOUT: [{}]", elementName);
                                throw new XExpression(ex);
                            } else if(ex instanceof SecurityException) {
                                config_log.warn(LOG_MARKER,"Caught groovy SECURITY exception [{}]: {}", elementName, ex.getMessage());
                                throw new XExpression(ex);
                            } else {
                                throw new XExpression(StackTraceUtil.getScriptLineNumber(ex), ex);
                            }
                        }
                    }
                }


                if (result instanceof AttributedResult) {
                    AttributedResult attributedResult = (AttributedResult) result;
                    r = formulaElementResult(e, attributedResult.getResult());
                    applyAttributes(r, attributedResult);
                } else {
                    r = formulaElementResult(e, result);
                }


                //did we accumulate new warnings?
                // if so => copy them to the element result
                if(ctx.getWarnings() != null) {
                    for(int i = 0;i<ctx.getWarnings().size();i++) {
                        r.getWarnings().add(ctx.getWarnings().get(i));
                    }

                    ctx.clearWarnings();
                }

                //did we accumulate trace messages?
                // if so => copy them to the element result
                if(ctx.getTraceMessages() != null) {
                    r.setTraceMessages(new ArrayList(ctx.getTraceMessages()));
                    ctx.clearTraceMessages();
                }

                if(e instanceof PricingFormulaElement) {
                    PricingFormulaElement fe = (PricingFormulaElement)e;
                    processAlerts(ctx, r, fe);
                    r.setOverridable(fe.getAllowOverride());

                    // determine suffix; the elementsuffix field contains a name of a previous element or a text
                    // if it is a valid name we pick the referenced result, or just the text otherwise
                    if(fe.getElementSuffix() != null && !fe.getElementSuffix().isEmpty() && r.getSuffix() == null) {
                        if(ctx.getResultValue(fe.getElementSuffix()) != null) {
                            r.setSuffix(ctx.getResultValue(fe.getElementSuffix()).toString());
                        } else {
                            r.setSuffix(fe.getElementSuffix());
                        }
                    }
                }

                ctx.addResult(r);

                if(ctx.isAborted() && !syntaxCheckMode) {
                    //end calculation execution
                    return r;
                }

                return r;

            } catch (XExpression ex) {
                final String error = String.format("Caught error in [%s] : %s", elementName, ex.getMessage());
                config_log.warn(LOG_MARKER, error);
                ctx.setErrorElementName(elementName);
                ctx.setErrorMessage(elementName + " : " + ex.getMessage());
                ctx.setErrorStartPosition(ex.getStartPosition());
                ctx.setErrorEndPosition(ex.getEndPosition());
                throw ex;
            }

        } //end for


    private Long getEffectiveTimeout(AbstractFormulaElement e, boolean isLibrary) {

        Long elementTimeout = defaultTimeout;
        if (e.getElementTimeout() != null) {
            elementTimeout = e.getElementTimeout();
            if (isLibrary) {
                if(elementTimeout > maxLibTimeout) elementTimeout = maxLibTimeout;
            } else {
                if(elementTimeout > maxTimeout) elementTimeout = maxTimeout;
            }
        }
        return elementTimeout;
    }

    @SuppressWarnings({ "unchecked", "rawtypes" })
    protected void applyAttributes(FormulaElementResult r, AttributedResult attributedResult) {
        Object suffix = attributedResult.getAttribute("suffix");
        if (suffix instanceof String) {
            r.setSuffix((String) suffix);
        }

        Object overrideValueOptions = attributedResult.getAttribute("overrideValueOptions");
        if(overrideValueOptions instanceof List) {
            r.setOverrideValueOptions((List)overrideValueOptions);
        }
        Object overrideAllowEmpty = attributedResult.getAttribute("overrideAllowEmpty");
        if(overrideAllowEmpty instanceof Boolean) {
            r.setOverrideAllowEmpty((boolean) overrideAllowEmpty);
        }

        applyCellStyles(r, attributedResult);
    }

    protected void applyCellStyles(FormulaElementResult r, AttributedResult attributedResult) {
        StringBuilder cellStyles = new StringBuilder();

        String existingCellStyles = r.getCellStyles();
        if (existingCellStyles != null) {
            cellStyles.append(existingCellStyles);
            if (!existingCellStyles.endsWith(";")) {
                cellStyles.append(';');
            }
        }

        Object property = attributedResult.getAttribute("color");
        if (property instanceof String) {
            cellStyles.append("color:").append(property).append(';');
        }

        property = attributedResult.getAttribute("text-decoration");
        if (property instanceof String) {
            cellStyles.append("text-decoration:").append(property).append(';');
        }

        property = attributedResult.getAttribute("background-color");
        if (property instanceof String) {
            cellStyles.append("background-color:").append(property).append(';');
        }

        property = attributedResult.getAttribute("RAWCSS");
        if (property instanceof String) {
            cellStyles.append(property);
        }

        if (cellStyles.length() != 0) {
            r.setCellStyles(cellStyles.toString());
        }
    }

    /*
     * The interrupt timers are defined at class instantiation
     * If we keep on reusing the classes of the library, they'll expire (e.g. in a list scenario)
     */
    private void refreshGroovyLibrary(LmSandboxAPI api) {
        Binding binding = new Binding();
        binding.setVariable(VAR_api, api);
        binding.setVariable(VAR_libs, groovyLibraries);

        Map<String, Script> groovyLibraryBindings = groovyLibraries.get(GROOVY_LIBRARY_UNIQUENAME);
        if (groovyLibraryBindings != null) binding.setVariable(VAR_lib, groovyLibraryBindings);

        //refresh all groovy libraries
        for (String libraryKey : groovyLibraries.keySet()){
            if (groovyLibraries.get(libraryKey) != null){
                groovyLibraries.get(libraryKey).forEach((k,v) -> {if (v != null) v.setBinding(binding);});
            }
        }
    }

    @SuppressWarnings("unchecked")
    private void initializeLibrary(LmSandboxAPI api, LmFormulaEngineContext ctx)  {

        //refresh all groovy libraries
        if (groovyLibraries != null){
            refreshGroovyLibrary(api);
        } else {
            //initialize all groovy libraries
            groovyLibraries = new LibraryBindingMap(new HashMap<>(), moduleDir, groovyShell, api, ctx);
        }
    }

    private boolean isFormulaElementOperative(AbstractFormulaElement e, LmFormulaEngineContext ctx){
        // do not execute if:
        // - a formula element group's been set in the ctx, and...
        // - the current element has group restrictions, and...
        // - none of these groups match the ctx's group setting:
        List<String> elementGroups = e.getElementGroups();
        if (ctx.getFormulaElementGroup() != null && elementGroups != null && !e.getElementGroups().isEmpty()) {
            boolean found = false;
            for (String elementGroup : elementGroups) {
                if (ctx.getFormulaElementGroup().equals(elementGroup)) {
                    found = true;
                    break;
                }
            }
            if (!found) {
                return false;
            }
        }
        // do not execute if there is a reference to another formula element, serving as a gate to the current one, and
        // this condition element evaluates as
        // a) Number 0
        // b) Boolean false
        // c) Object null
        if (e.getConditionElementName() != null){
            Object conditionElementResult = ctx.getResultValue(e.getConditionElementName());
            if (conditionElementResult == null
                    || Boolean.FALSE.equals(conditionElementResult)
                    || (conditionElementResult instanceof Number && ((Number)conditionElementResult).intValue() == 0)){
                return false;
            }
        }
        return true;
    }


    // no need to optimize anything, we use compiler just for first load on jvm.
    private CompilerConfiguration getGroovyCompilerConfiguration() {

        CompilerConfiguration cc = new CompilerConfiguration();

        ImportCustomizer importCustomizer = new ImportCustomizer();

        //add timeout & sandbox
        addGroovyImports(importCustomizer,
                LibraryMethod.class,
                FieldFormatType.class,
                DataType.class,
                InputType.class,
                InputButtonAction.class,
                ConfiguratorEntry.class,
                ResultMatrix.class,
                Matrix2D.class,
                Filter.class,
                QuoteStructure.class,
                TieredValue.Tier.class,
                TieredValue.class,
                CustomerGroup.class,
                ProductGroup.class,
                TimePeriod.class,
                TimeUnit.class,
//                DMDataSlice.class,
//                DMCalendar.class,
                TimedObjectWithStatus.TimedObjectStatus.class,
                DateTime.class,
                PriceFxException.class,
//                LookupTableType.class,            //can't use since there are 2 LookupTable classes
//                LookupTableValueType.class,       //can't use since there are 2 LookupTable classes
                groovy.transform.Field.class,
                RoundingMode.class);

        for (GroovyDefaultImport i : BuilderExtensions.getGroovyDefaultImports()) {
            importCustomizer.addImport(i.alias(), i.className());
        }

        cc.addCompilationCustomizers(importCustomizer);
        return cc;
    }

    private static void addGroovyImports(ImportCustomizer importCustomizer, Class<?>...classes){
        for (Class<?> clazz : classes){
            importCustomizer.addImport(clazz.getSimpleName(), clazz.getName());
        }
    }

    private Object syntaxCheckModeResultFix(Object result) {
//        AttributedResult attributedResult = null;
//        if (result instanceof AttributedResult) {
//            attributedResult = (AttributedResult) result;
//            result = attributedResult.getResult();
//        }

        if (result == null) {
            result = new BigDecimal(LmFormulaEngineContext.CHECK_RESULT_STRING);
        } else if (result instanceof BigDecimal && ((BigDecimal) result).compareTo(BigDecimal.ZERO) == 0) {
            result = new BigDecimal(LmFormulaEngineContext.CHECK_RESULT_STRING);
        } else if (result instanceof Integer && ((Integer) result).intValue() == 0) {
            result = new BigDecimal(LmFormulaEngineContext.CHECK_RESULT_STRING);
        }

        return result;//attributedResult == null ? result : attributedResult;
    }

    private Object getManualOverrideResult(LmFormulaEngineContext ctx, PricingFormulaElement e) {
        Object val = ctx.getManualResultOverrideValue(e.getElementName());
        if (val == null)
            return null;

        if (e.getAllowOverride()) {
            return val;
        } else {
            if(ctx.isForcedManualResultOverride(e.getElementName()))
                return val;
            else
                return null; //ignoring override here
        }
    }

    /**
     * Checks if there are alerts raised and adds them into the result if necessary
     *
     * @param elementResult
     * @param element
     */
    private void processAlerts(LmFormulaEngineContext ctx, FormulaElementResult elementResult, PricingFormulaElement element) {

        if(element == null || elementResult == null)
            return;

        // groovy alerts first, but do not return; thus UI configured alerts prevail
        if(ctx.hasAlert()) {
            elementResult.setAlertType(ctx.getAlertType());
            elementResult.setAlertMessage(ctx.getAlertMessage());
            ctx.clearAlert();
        }

        // criticalAlert first
        PricingFormulaElement.AlertConfiguration criticalAlert = element.getAlertConfiguration(FIELD_FE_CRITICALALERT);
        if(criticalAlert != null) {
            if(alertTriggered(ctx, elementResult.getElementResult(), criticalAlert)){
                elementResult.setAlertType("CRITICAL");
                elementResult.setAlertMessage(criticalAlert.getMessage());
                //don't process red if we have critical already
                return;
            }
        }

        // redAlert second
        PricingFormulaElement.AlertConfiguration redAlert = element.getAlertConfiguration(FIELD_FE_REDALERT);
        if(redAlert != null) {
            if(alertTriggered(ctx, elementResult.getElementResult(), redAlert)){
                elementResult.setAlertType("RED");
                elementResult.setAlertMessage(redAlert.getMessage());
                //don't process yellow if we have red already
                return;
            }
        }


        PricingFormulaElement.AlertConfiguration yellowAlert = element.getAlertConfiguration(FIELD_FE_YELLOWALERT);
        if(yellowAlert != null) {
            if(alertTriggered(ctx, elementResult.getElementResult(), yellowAlert)){
                elementResult.setAlertType("YELLOW");
                elementResult.setAlertMessage(yellowAlert.getMessage());
            }
        }
    }

    private boolean alertTriggered(LmFormulaEngineContext ctx, Object elementResult, PricingFormulaElement.AlertConfiguration alert) {

        if(alert == null)
            return false;

        BigDecimal compareVal = null;
        if(elementResult instanceof BigDecimal)
            compareVal = (BigDecimal)elementResult;

        if(elementResult instanceof Integer)
            compareVal = new BigDecimal((Integer)elementResult);


        String operator = alert.getOperator();
        BigDecimal threshold = alert.getThreshold();

        if(operator == null || operator.isEmpty()) {
            return false;
        }

        //check null/non-null ops
        if("isnull".equals(operator) && elementResult == null)
            return true;

        if("isnotnull".equals(operator) && elementResult != null)
            return true;

        if(elementResult == null)
            return false;	//cannot trigger anymore; null check handled above

        //check on threshold
        if(threshold != null) {
            if(compareVal == null)
                return false;	//either elementResult == null or of non-numeric type

            if(processAlertOperator(operator,compareVal, threshold))
                return true;
        }

        //check on comparison element
        if(alert.getComparisonElement() != null) {
            if(ctx.getResultValue(alert.getComparisonElement()) instanceof BigDecimal) {
                return processAlertOperator(operator, compareVal,(BigDecimal) ctx.getResultValue(alert.getComparisonElement()));
            }
            if(ctx.getResultValue(alert.getComparisonElement()) instanceof Integer) {
                return processAlertOperator(operator, compareVal,new BigDecimal((Integer) ctx.getResultValue(alert.getComparisonElement())));
            }
            if(ctx.getResultValue(alert.getComparisonElement()) != null &&
                    "=".equals(operator)) {
                return ctx.getResultValue(alert.getComparisonElement()).equals(elementResult);
            }
            if(ctx.getResultValue(alert.getComparisonElement()) != null &&
                    "!=".equals(operator)) {
                return !ctx.getResultValue(alert.getComparisonElement()).equals(elementResult);
            }
        }
        return false;
    }

    private boolean processAlertOperator(String operator, BigDecimal compareVal, BigDecimal compareTo) {

        if(compareVal == null || compareTo == null)
            return false;

        if("<".equals(operator)) {
            return compareVal.compareTo(compareTo) < 0;
        } else if ("<=".equals(operator)) {
            return compareVal.compareTo(compareTo) <= 0;
        } else if ("|<|".equals(operator)) {
            return compareVal.abs().compareTo(compareTo) <= 0;
        } else if (">=".equals(operator)) {
            return compareVal.compareTo(compareTo) >= 0;
        } else if (">".equals(operator)) {
            return compareVal.compareTo(compareTo) > 0;
        } else if ("|>|".equals(operator)) {
            return compareVal.abs().compareTo(compareTo) > 0;
        } else if ("!=".equals(operator)) {
            return compareVal.compareTo(compareTo) != 0;
        } else if ("=".equals(operator)) {
            return compareVal.compareTo(compareTo) == 0;
        } else
            return false;
    }

    public FormulaElementResult formulaElementResult(AbstractFormulaElement element) {
        FormulaElementResult result = new FormulaElementResult();
        result.setElementName(element.getElementName());
        return result;
    }

    public FormulaElementResult formulaElementResult(AbstractFormulaElement element, Object result, Long displayOptions) {
        FormulaElementResult res = formulaElementResult(element);
//        if (res instanceof DecoratedResult) {
//            DecoratedResult dr = (DecoratedResult) res;
//            this.elementResult = dr.result;
//            this.elementResultDescription = dr.description;
//        }
//        else{
            res.setElementResult(result);
//        }
        res.setDisplayOptions(displayOptions);
        return res;
    }

    private FormulaElementResult formulaElementResult(AbstractFormulaElement e, Object result) {
        FormulaElementResult res = formulaElementResult(e);
/*        if (result instanceof DecoratedResult) {
            DecoratedResult dr = (DecoratedResult) result;
            res.setElementResult(dr.result);
            res.setElementResultDescription(dr.description);
        }
        else{*/
        res.setElementResult(result);
//        }
        res.setElementLabel(e.getElementLabel());
        res.setElementResultDescription(e.getElementDescription());
        if (e instanceof PricingFormulaElement) {
            PricingFormulaElement fe = (PricingFormulaElement) e;
            res.setDisplayOptions(fe.getDisplayOptions());
            res.setFormatType(fe.getFormatType());
            res.setCssProperties(fe.getCssProperties());
            res.setSummarize(fe.getSummarize() != null && fe.getSummarize().booleanValue());
            res.setHideOnNull(fe.getHideOnNull() != null && fe.getHideOnNull().booleanValue());
            res.setUserGroup(fe.getUserGroup());
//            res.setResultGroup(fe.getResultGroup());
        }
        return res;
    }


    public FormulaElementResult formulaElementResult(AbstractFormulaElement element, Object result, boolean manual) {
        FormulaElementResult res = formulaElementResult(element, result);
        res.setOverridden(manual);
        return res;
    }

    public LmFormulaEngineContext getContext() {
        return ctx;
    }




}
