package net.pricefx.lmformulaengine.scripting;

import com.googlecode.genericdao.search.Filter;
import net.pricefx.logic.core.model.AbstractFormula;
import net.pricefx.logic.core.model.AbstractFormulaElement;
import net.pricefx.logic.core.model.ElementCombinationType;
import net.pricefx.logic.core.model.FormulaErrorElementResult;
import net.pricefx.logic.core.LogicException;
import net.pricefx.logic.core.PfxCommonService;
import net.pricefx.pckg.client.okhttp.PfxClient;
import net.pricefx.pckg.transform.descriptor.TypeDescriptor;

import java.util.*;

import static net.pricefx.common.api.PriceFxInterface.*;

public class LmSandboxApiAgent {
    public static final String LOGIC_NAME = "StudioEvaluator";
    public static final String ELEMENT_DEBUG = "result";

    private PfxClient client;
    private AbstractFormula formula;
    private Map<String, Object> parameters = null;

    public LmSandboxApiAgent(PfxClient client, TypeDescriptor objectType, Map<String, Object> parameters) {
        this.client = client;
        this.parameters = parameters;
        formula = createVirtualFormula(objectType);
    }

    public static AbstractFormula createVirtualFormula(TypeDescriptor objectType) {
        AbstractFormula virtFormula = AbstractFormula.createLogic(objectType);
        virtFormula.setUniqueName(LOGIC_NAME);

        AbstractFormulaElement element = virtFormula.createElement();
        element.setElementName(ELEMENT_DEBUG);
        element.setCombinationType(ElementCombinationType.FUNCTION.name());
        virtFormula.addElement(element);

        return virtFormula;
    }


    public Object getParameter(String parameterName) {
        return parameters.get(parameterName);
    }

    public Object getParameter(String[] parameterNames) {
        Object result = null;
        for (String parameterName : parameterNames) {
            result = parameters.get(parameterName);
        }
        return result;
    }

    private static String toSourceCodeFilterOperator(int operator) throws LogicException {
        switch (operator) {
            case Filter.OP_EQUAL:
                return "equal";
            case Filter.OP_LESS_THAN:
                return "lessThan";
            case Filter.OP_GREATER_THAN:
                return "greaterThan";
            case Filter.OP_LESS_OR_EQUAL:
                return "lessOrEqual";
            case Filter.OP_GREATER_OR_EQUAL:
                return "greaterOrEqual";
            case Filter.OP_LIKE:
                return "like";
            case Filter.OP_ILIKE:
                return "ilike";
            case Filter.OP_IN:
                return "in";
            case Filter.OP_NOT_IN:
                return "notIn";
            case Filter.OP_NULL:
                return "isNull";
            case Filter.OP_NOT_NULL:
                return "isNotNull";
            case Filter.OP_EMPTY:
                return "equal";
            case Filter.OP_NOT_EMPTY:
                return "isNotEmpty";
            case Filter.OP_AND:
                return "and";
            case Filter.OP_OR:
                return "or";
            case Filter.OP_NOT:
                return "not";
            case Filter.OP_SOME:
                return "some";
            case Filter.OP_ALL:
                return "all";
            case Filter.OP_NONE:
                return "none";
            case Filter.OP_CUSTOM:
                return "custom";
            default:
                throw new LogicException("Unsupported Filter operator " + operator);
        }
    }

    private static String toSourceCode(Filter filter) throws LogicException {
        String start = "Filter." + toSourceCodeFilterOperator(filter.getOperator());
        if (filter.isTakesSingleValue()) {
            return start + "(" + toSourceCode(filter.getProperty()) + ", " + toSourceCode(filter.getValue()) + ")";
        }
        return "";
    }

    private static String toSourceCode(Object[] value) throws LogicException {
        if (value == null) {
            return null;
        }
        String expr = "";
        String sep = "";
        for (Object arrayValue : value) {
            expr += sep + toSourceCode(arrayValue);
            sep = ", ";
        }
        return expr;
    }

    private static String toSourceCode(Object value) throws LogicException {
        if (value == null) {
            return "null";
        } else if (value instanceof Number || value instanceof Boolean) {
            return value.toString();
        } else if (value instanceof String) {
            return "\"" + value.toString() + "\"";
        } else if (value instanceof Filter) {
            return toSourceCode((Filter) value);
        } else if (value instanceof Object[]) {
            return toSourceCode((Object[]) value);
        }
        return toSourceCode(value.toString());
//        try {
//            Utils.getObjectMapper().writeValueAsString(value);
//        } catch (JsonProcessingException e) {
//            throw new LogicException(e.getMessage(), e);
//        }
    }

    private static String toSourceCode(List arrayValues) throws LogicException {
        String expr = "";
        String sep = "";
        for (Object arrayValue : arrayValues) {
            expr += sep + toSourceCode(arrayValue);
            sep = ", ";
        }
        return expr;
    }

    private static String buildCommand(String methodName, Object... params) throws LogicException {
        String expr = "api." + methodName + "(" + toSourceCode(params) + ")";
        return expr;
    }


    public FormulaErrorElementResult callApiExpression(String expression, Object... params) throws LogicException {
        Iterator<AbstractFormulaElement> it = formula.getElementsIterator();
        AbstractFormulaElement element = it.next();
        element.setFormulaExpression(expression);

        List<FormulaErrorElementResult> results = PfxCommonService.testExecFormula(client, formula, parameters);

        if (results == null || results.size() == 0) {
            throw new LogicException("Couldn't execute " + element.getFormulaExpression());
        }

        return results.get(0);
    }

    public Object callFormulaExpression(String expression, Object... params) throws LogicException {
        return callApiExpression(expression, params).getElementResult();
    }

    //do not change this call structure of the function unless or make sure getStackTrace will still work
    public Object callApiUsingFormula(Object...  params) throws LogicException {
        String methodName = Thread.currentThread().getStackTrace()[2].getMethodName();
        String expr = buildCommand(methodName, params);
        return callFormulaExpression(expr, params);
    }

    //do not change this call structure of the function unless or make sure getStackTrace will still workpublic Map<String, Object> callApi(String methodName, TypeReference typeReference, Object...  params) throws LogicException {
    public <T> T callApi(Object...  params) throws LogicException {
        String methodName = Thread.currentThread().getStackTrace()[2].getMethodName();
        return (T) ((List) PfxCommonService.debugApiFunction(client, methodName, Arrays.asList(params),
                (String) parameters.get(PARAM_FM_ITEM), (String) parameters.get(PARAM_TARGETDATE))).get(0);
    }
}
