package net.pricefx.server.dto.calculation;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.googlecode.genericdao.search.Filter;

import net.pricefx.common.api.CalculationResultType;
import net.pricefx.common.api.FieldFormatType;
import net.pricefx.common.api.PriceFxInterface;
import net.pricefx.domain.PriceGrid;
import net.pricefx.domain.Pricelist;
import net.pricefx.formulaengine.scripting.Matrix2D;
import net.pricefx.logic.core.TypeCode;


/**
 * This class is used to transport a calculated matrix-style result
 * It is basically a list of maps with the addition that the column names are clearly named and can have a format
 * This means:
 * - the UI knows exactly which columns to display (even if some maps do not have those as keys)
 * - the entry maps can have more "columns" (or keys) than those that are displayed; maybe useful for calculation
 * 
 * @author cat
 *
 */
@JsonIgnoreProperties(ignoreUnknown = true)
public class ResultMatrix implements ComplexCalculationResult, Serializable {

	private static final long serialVersionUID = 2266960694562305277L;
	
	private List<Map<String,Object>> entries;
	private List<String> columns;
	private FieldFormatType defaultFormat;
	private Map<String,FieldFormatType> columnFormats;
	private Map<String,String> columnTooltips;
	private boolean enableClientFilter;
	private String title;
	private String preferenceName;
	private boolean disableSorting;
	private boolean escapeHTML;
	private Map<String,String> columnTranslations;	
	
	private Map<String, Map<String, String>> onRowSelectEvents = new HashMap<>();
	
	public ResultMatrix() {
		this((Map<?, ?>) null);
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public ResultMatrix(Map jsonMap) {
		if(jsonMap == null) {
			entries = new ArrayList<Map<String,Object>>();
			columns = new ArrayList<String>();
		} else {
			entries = (List<Map<String, Object>>) jsonMap.get("entries");
			columns = (List<String>) jsonMap.get("columns");
		}
		
	}	
	
	/**
	 * Gets the current title
	 * @return The current title
	 */
	public String getTitle() {
		return title;
	}

	/**
	 * Sets the title that will be shown as window title of the result matrix in the client
	 * @param title The title
	 */
	public void setTitle(String title) {
		this.title = title;
	}

	/**
	 * Gets the preference name
	 * @return The preference name
	 */
	public String getPreferenceName() {
		return preferenceName;
	}

	/**
	 * UI sorting of matrix grid
	 * @return true if sorting is disabled on all columns. Default is false
	 */
	public boolean isSortingDisabled() {
		return disableSorting;
	}

	/**
	 * Set sorting capability in UI grid
	 * @param disableSorting true: sorting is disabled in UI on all columns
	 */
	public void setDisableSorting(boolean disableSorting) {
		this.disableSorting = disableSorting;
	}

	/**
	 * Is HTML escaping enabled?
	 * @return true if the matrix should display raw text as is (i.e. not try to interpret it as HTML)
	 */
	public boolean isEscapeHTML() {
		return escapeHTML;
	}

	/**
	 * Set HTML escaping flag. Default: false (i.e. values will be interpreted as HTML)
	 * @param escapeHTML
	 */
	public void setEscapeHTML(boolean escapeHTML) {
		this.escapeHTML = escapeHTML;
	}

	/**
	 * Sets the so-called preference name. I.e. all result matrixes with the same preference name will share the same preference setting.
	 * The preference setting can be managed as usual in the client and can be set as global to predefine a good setting
	 * 
	 * @param preferenceName The preference name
	 */
	public void setPreferenceName(String preferenceName) {
		if(preferenceName != null)
			this.preferenceName = preferenceName.replaceAll("[^a-zA-Z0-9]+","");
		else
			this.preferenceName = null;
	}

	/**
	 * Creates a new matrix object with an initial set of columns
	 * @param columns The initial set of column names
	 */
	public ResultMatrix(String ... columns) {
		this();
		for (int i = 0; i < columns.length; i++)
			addColumn(columns[i]);
	}
	
	/**
	 * Adds translated labels for the given column. 
	 * @param column The column name (as defined e.g. in the constructor)  
	 * @param translations A map with the translations. The key of the map is the locale, like "de".
	 */	
	public void addColumnTranslation(String column, Map<String,String> translations) {
		if(columnTranslations == null) {
			columnTranslations = new HashMap<String,String>();			
		}
		
//		columnTranslations.put(column, DomainObject.jsonSerialize(translations));
	}
		
	/**
	 * Gets current column translations in raw format
	 * @return
	 */
    public Map<String, String> getColumnTranslations() {
		return columnTranslations;
	}


	/**
     * Creates a new matrix object with an initial set of columns
     * @param columns Collection of column names
     */
	public ResultMatrix(Collection<String> columns) {
	    this(columns.toArray( new String[columns.size()]) );
	}

	/**
	 * Adds a column to the existing set of columns
	 * @param column The columns name
	 */
	public void addColumn(String column) {
		if(!columns.contains(column))
			columns.add(column);
	}
	
	/**
	 * Adds a column to the existing set of columns including a tooltip
	 * @param column The columns name
	 */
	public void addColumn(String column, String tooltip) {
		if(!columns.contains(column))
			columns.add(column);
		
		setColumnTooltip(column,tooltip);
	}	

    /**
     * Adds columns to the existing set of columns
     * @param columns Collection of columns names
     */
	public void addColumns(Collection<String> columns) {
	    for (String col : columns){
	        addColumn(col);
	    }
	}

	/**
	 * Adds a data row. The keys of the row-map should match the column names defined earlier
	 * @param row
	 */
	public void addRow(Map<String,Object> row) {
		entries.add(row);
	}

    /**
     * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
     * @param row Row as List
     */
   public void addRow(List<Object> row) {
       if (row != null){
           Map<String,Object> m = new HashMap<>();
           for (int c=0; c < Math.max(row.size(), columns.size()); c++){
               m.put(columns.get(c), row.get(c));
           }
           entries.add(m);
       }
    }
   
   /**
    * Adds a data row. The order of the values is assumed to be the same as the matrix column order.
    * @param row Row as Array
    */
   @SuppressWarnings("unchecked")
   public void addRow(Object... values) {
       if (values != null){
           if (values.length == 1 && values[0] instanceof Map) {
               addRow((Map<String,Object>) values[0]);
           }
           else{
               addRow(Arrays.asList(values));
           }
       }
   }

	/**
	 * Gets a list of defined columns
	 * @return The columns
	 */
	public List<String> getColumns() {
		return columns;
	}
	
	/**
	 * Returns the current list entries as a list (=the rows) of maps (= the row content)
	 * @return List of maps
	 */
	public List<Map<String,Object>> getEntries() {
		return entries;
	}
	
	/**
	 * Retrieves a particular row based on a value check
	 * @param key The column name to search on
	 * @param value The value to match the row's column value to
	 * @return The row
	 */
	public Map<String,Object> findRow(String key, Object value) {
		for(Map<String,Object> entry : entries) {
			Object val = entry.get(key);
			if(val == null && value == null) return entry;
			if(val != null && val.equals(value)) return entry;
		}
		return null;
	}
	
	/**
	 * Transpose the matrix.
	 * 
	 * <br/><br/>
	 * 
	 * Example:
	 * <PRE>
	 * ProductGroup | AvgPrice | AvgMargin <- header row
	 * -----------------------------------
	 *     Beef     |   4.1€   |   3.6%
	 *     Pork     |   2.2€   |   1.1%
	 *     Lamb     |   4.6€   |   5.2%
	 * </PRE>
	 * 
	 * Transposed matrix:
	 * <PRE>
	 * transpose("ProductGroup")
	 * 
	 * ProductGroup | Beef | Pork | Lamb <- header row
	 * ---------------------------------
	 *   AvgPrice   | 4.1€ | 2.2€ | 4.6€
	 *   AvgMargin  | 3.6% | 1.1% | 5.2%
	 * </PRE>
	 * 
	 * 
	 * @param newHeaderColumnName Name of the column that will become the new header row
	 * @return The transposed ResultMatrix (new instance)
	 */
	public ResultMatrix transpose(String newHeaderColumnName) throws Exception {
		throw new Exception("Method ResultMatrix.transpose() is not supported by debugger");
	}
	
	/**
	 * Transpose the matrix.
	 * 
	 * <br/><br/>
	 * 
	 * Example:
	 * <PRE>
	 * ProductGroup | AvgPrice | AvgMargin  <- header row
	 * ------------------------------------
	 *     Beef     |   4.1€   |   3.6%
	 *     Pork     |   2.2€   |   1.1%
	 *     Lamb     |   4.6€   |   5.2%
	 * </PRE>
	 * 
	 * Transposed matrix:
	 * <PRE>
	 * transpose("ProductGroup", "PG Label")
	 * 
	 *  Beef | Pork | Lamb |   PG Label    <- header row
	 * -----------------------------------
	 *  Beef | Pork | Lamb | ProductGroup
	 *  4.1€ | 2.2€ | 4.6€ |   AvgPrice
	 *  3.6% | 1.1% | 5.2% |   AvgMargin
	 * </PRE>
	 * 
	 * 
	 * @param newHeaderColumnName Name of the column that will become the new header row
	 * @param oldColLabelsInNewColumnName Name of the new column that will receive the old column headers as values (optional)
	 * @return The transposed ResultMatrix (new instance)
	 */
	public ResultMatrix transpose(String newHeaderColumnName, String oldColLabelsInNewColumnName) throws Exception {
		throw new Exception("Method ResultMatrix.transpose() is not supported by debugger");
	}
	
	/**
	 * Sets a default column format
	 * @param fft
	 */
	public void setDefaultFormat(FieldFormatType fft) {
		defaultFormat = fft;
	}
	
	/**
	 * Gets the default format
	 * @return Default format
	 */
	public FieldFormatType getDefaultFormat() {
		return defaultFormat;
	}

	/**
	 * Gets individual column formats (if specified)
	 * @return A Map with column name and format
	 */
	public Map<String, FieldFormatType> getColumnFormats() {
		return columnFormats;
	}

	public Map<String, Map<String, String>> getOnRowSelectEvents() {
		return onRowSelectEvents;
	}

	/**
	 * Sets a column format for a specified column
	 * @param column The column name
	 * @param fft The format
	 */
	public void setColumnFormat(String column, FieldFormatType fft) {
		if(columnFormats == null)
			columnFormats = new HashMap<String,FieldFormatType> ();
		columnFormats.put(column, fft);
	}

	/**
	 * Gets the currently assigned column tooltips
	 * @return
	 */
	public Map<String, String> getColumnTooltips() {
		return columnTooltips;
	}

	/**
	 * Adds a tooltip to a column
	 * @param column The column name
	 * @param tooltip The tooltip text
	 */
	public void setColumnTooltip(String column, String tooltip) {
		
		if(columnTooltips == null)
			columnTooltips = new HashMap<String,String>();
		
		if(tooltip == null)
			columnTooltips.remove(column);
		else
			columnTooltips.put(column, tooltip);
	}

	public boolean isEnableClientFilter() {
		return enableClientFilter;
	}

	public void setEnableClientFilter(boolean enableClientFilter) {
		this.enableClientFilter = enableClientFilter;
	}

	/**
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @return The styled cell object that can then be further styled with e.g. the flow pattern .withXXX() methods
	 */
	public ResultMatrixStyledCell styledCell(Object value) {
		ResultMatrixStyledCell rmcc = new ResultMatrixStyledCell();
		rmcc.value = value;
		return rmcc;
	}
	
	/**
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @param textColor The text color of the cell
	 * @param bgColor The background color of the cell (HTML hex value)
	 * @return The styled cell object
	 */
	public ResultMatrixStyledCell styledCell(Object value, String textColor,String bgColor) {
		ResultMatrixStyledCell rmcc = new ResultMatrixStyledCell();
		rmcc.value = value;
		rmcc.color = textColor;
		rmcc.background = bgColor;
		return rmcc;
	}
	
	/**
	 * 
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @param textColor The text color of the cell
	 * @param bgColor The background color of the cell (HTML hex value)
	 * @param weight The font weight
	 * @return The styled cell object
	 */
	public ResultMatrixStyledCell styledCell(Object value, String textColor,String bgColor, String weight) {
		ResultMatrixStyledCell rmsc = new ResultMatrixStyledCell();
		rmsc.value = value;
		rmsc.color = textColor;
		rmsc.background = bgColor;
		rmsc.fontweight = weight;
		return rmsc;
	}
	
	/**
	 * Creates a styled cell that can then be added as row value
	 * @param value The value of the cell
	 * @param textColor The text color of the cell
	 * @param bgColor The background color of the cell (HTML hex value)
	 * @param weight The font weight
	 * @param alignment Cell alignement value (valid CSS)
	 * @return The styled cell object
	 */
	public ResultMatrixStyledCell styledCell(Object value, String textColor,String bgColor, String weight, String alignment) {
		ResultMatrixStyledCell rmsc = new ResultMatrixStyledCell();
		rmsc.value = value;
		rmsc.color = textColor;
		rmsc.background = bgColor;
		rmsc.fontweight = weight;
		rmsc.alignment = alignment;
		return rmsc;
	}	
	
	/**
	 * Creates an image cell that can then be added as row value. Value should be a valid image URL
	 * @param value The image URL
	 * @return The image cell object
	 */
	public ResultMatrixImageCell imageCell(Object value) {
		ResultMatrixImageCell rmic = new ResultMatrixImageCell();
		rmic.value = value;
		return rmic;
	}
	
	public ResultMatrixImageCell libraryImage(String mainName, String statusName) {
		ResultMatrixImageCell rmic = new ResultMatrixImageCell();
		rmic.value = "images/library/" + mainName + "_" + statusName + ".png";
		return rmic;
	}
	
	public ResultMatrixLinkCell linkToPriceGrid(Object value, long priceGridID, Filter filter) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();

		rmlc.value = value;
		rmlc.targetPage = "priceGridPage";

        HashMap<String, Object> state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
                priceGridID + "." + TypeCode.PRICEGRID.getShortName());
        state.put(PriceFxInterface.PARAM_FILTER_CRITERIA, filter);

		rmlc.targetPageState = state;

		return rmlc;
	}

	public ResultMatrixLinkCell linkToPriceList(Object value, long priceListID, Filter filter) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();

        rmlc.value = value;
        rmlc.targetPage = "priceListPage";

        HashMap<String, Object> state = new HashMap<String, Object>(2);
        state.put(PriceFxInterface.FIELD_ID,
                priceListID + "." + TypeCode.PRICELIST.getShortName());
        state.put(PriceFxInterface.PARAM_FILTER_CRITERIA, filter);

		rmlc.targetPageState = state;

		return rmlc;
	}
	
	public ResultMatrixLinkCell linkCell(Object value, String targetPage, String targetPageState) {
		ResultMatrixLinkCell rmlc = new ResultMatrixLinkCell();
		rmlc.value = value;
		rmlc.targetPage = targetPage;
		rmlc.targetPageState = targetPageState;
		return rmlc;
	}
	
	/**
	 * Configures a client-side UI event. Use case is usually a refresh of some sort when a result matrix row is selected
	 * Example code:
	 * <pre>
	 * matrix.onRowSelection().triggerEvent(api.dashboardWideEvent("CustomerIdChanged-ED")).withColValueAsEventDataAttr("CustomerID", "customerId")
	 * </pre>
	 * The embedded dashboard that receives the event could then look like:
	 * <pre>
	 *  return api.dashboard("EmbeddedDashboard")
	 *
	 *    // Here the user can assign values to the embedded
	 *    // dashboard inputs common to all three modes
	 *    .setParam("Customer Id", api.input("Customer Id"))
	 *    .setParam("Year", api.input("Year"))
	 *
	 *    // Show the dashboard embedded ...
	 *    .showEmbedded()
	 *    // ... and reevaluate it on "CustomerIdChanged-ED"
	 *    // Note: "dashboardWideEvent()" fce makes the event local
	 *    // to the containing dashboard instance
	 *    .andRecalculateOn(api.dashboardWideEvent("CustomerIdChanged-ED"))
	 *    // Pull the "customerId" attribute out of the event payload and expose it
	 *    // as the "Customer Id" input to the embedded dashboard
	 *    .withEventDataAttr("customerId").asParam("Customer Id")
	 * </pre>   
	 * @return The selection event to further configure it in a flow pattern
	 */
	public ResultMatrixOnRowSelection onRowSelection() {
		return new ResultMatrixOnRowSelection();
	}
	
	@Override
	public String toString() {
		return String.format("(%s),(%s)", columns,entries);
	}

	@Override
	public String getResultType() {
		return CalculationResultType.MATRIX.name();
	}
	
	public abstract class AbstractResultMatrixCell {
		public Object value;
		public String type;
	}
	
	public class ResultMatrixStyledCell extends AbstractResultMatrixCell {
		public String color;
		public String background;
		public String fontweight;
		public String alignment;
		public String extraCSS;
		public String tooltip;
		
		public ResultMatrixStyledCell() {
			type = "StyledCell";
		}
		
		/**
		 * Set the foreground color of the value font
		 * @param color
		 * @return
		 */
		public ResultMatrixStyledCell withColor(String color) {
			this.color = color;
			return this;
		}
		
		/**
		 * Set the background color of the cell
		 * @param bgColor
		 * @return
		 */
		public ResultMatrixStyledCell withBackground(String bgColor) {
			this.background = bgColor;
			return this;
		}
		
		/**
		 * Set the font weight (in standard CSS terms) of the value text
		 * @param fontWeight
		 * @return
		 */
		public ResultMatrixStyledCell withFontWeight(String fontWeight) {
			this.fontweight = fontWeight;
			return this;
		}
		
		/** 
		 * Specify the alignment of the cell
		 * @param alignment
		 * @return
		 */
		public ResultMatrixStyledCell withAlignment(String alignment) {
			this.alignment = alignment;
			return this;
		}	
		
		/**
		 * Show a tooltip text in the cell
		 * @param tooltip
		 * @return
		 */
		public ResultMatrixStyledCell withToolTip(String tooltip) {
			this.tooltip = tooltip;
			return this;
		}	
		
		/**
		 * Apply any other valid CSS instructions to the cell
		 * @param extraCSS
		 * @return
		 */
		public ResultMatrixStyledCell withCSS(String extraCSS) {
			this.extraCSS = extraCSS;
			return this;
		}			
	}
	
	public class ResultMatrixLinkCell extends AbstractResultMatrixCell {
		public String targetPage;
		public Object targetPageState;
		
		public ResultMatrixLinkCell() {
			type = "LinkCell";
		}
	}	
	
	public class ResultMatrixImageCell extends AbstractResultMatrixCell {
				
		public ResultMatrixImageCell() {
			type = "ImageCell";
		}
	}	
	
	public class ResultMatrixOnRowSelection {
		
		public ResultMatrixOnRowSelection() {
		}
		
		public TriggerEvent triggerEvent(String eventName) {
			if (eventName == null || eventName.isEmpty()) {
				throw new IllegalArgumentException("Illegal argument in withColValueAsEventDataAttr(...)");
			}
			HashMap<String, String> mappings = new HashMap<String, String>();
			onRowSelectEvents.put(eventName, mappings);
			return new TriggerEvent(mappings);
		}
		
		public class TriggerEvent {
			
			private final Map<String, String> mappings;
			
			public TriggerEvent(HashMap<String, String> mappings) {
				this.mappings = mappings;
			}

            public TriggerEvent withColValueAsEventDataAttr(String colName, String attrName) {
                if (colName == null || colName.isEmpty()) {
                    throw new IllegalArgumentException("Illegal argument in withColValueAsEventDataAttr(...)");
                }
                if (attrName == null || attrName.isEmpty()) {
                    throw new IllegalArgumentException("Illegal argument in withColValueAsEventDataAttr(...)");
                }
                mappings.put(colName, attrName);
                return TriggerEvent.this;
            }
        }
    }
}
