package net.pricefx.testing;

import net.pricefx.logic.core.PfxClientBuilder;
import net.pricefx.logic.core.metadata.MetadataCache;
import net.pricefx.logic.core.metadata.MetadataProvider;
import net.pricefx.logic.core.model.AbstractFormula;
import net.pricefx.logic.core.model.FormulaErrorElementResult;
import net.pricefx.logic.core.AbstractLogicSerializer;
import net.pricefx.logic.core.LogicException;
import net.pricefx.logic.core.LogicSerializer;
import net.pricefx.logic.core.PfxCommonService;
import net.pricefx.logic.core.config.Config;
import net.pricefx.logic.core.config.Partition;
import net.pricefx.pckg.processing.ProcessingContextImpl;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class LogicTester {

    Partition partition;
    AbstractFormula logic;

    Map<String, FormulaErrorElementResult> savedResults;
    Map<String, FormulaErrorElementResult> currentResults;

    Config config;
    Path projectDir;

    /**
     *
     * @param env environment where to run the test
     * @param partitionName partition where to run the test
     * @param logicDir name of the logic directory (incl. date if there is any)
     * @throws Exception
     */
    LogicTester(String env, String partitionName, String logicDir) throws LogicException {
        projectDir = Paths.get(System.getProperty("user.dir"));
        config = new Config().load(projectDir, PfxClientBuilder.getInstance());
        setPartition(env, partitionName);
        loadLogic(logicDir);
    }

    /**
     *
     * @param logicDir
     * @throws LogicException
     */
    private void loadLogic(String logicDir) throws LogicException {
        logic = LogicSerializer.readOnDiskLogicWithElements(projectDir, new ProcessingContextImpl(), Paths.get(logicDir), false);
    }

    /**
     * Finds a partition object for a given environment and partition name
     * @param env environment where to run the test
     * @param partitionName partition where to run the test
     * @return the found partition object or null
     * @throws LogicException if it is not possible to find given env/partition in the config.groovy
     */
    private Partition getPartition(String env, String partitionName) throws LogicException {
        return config.findPartition(env, partitionName);
    }

    /**
     * Creates the Pfx client to the specified Pfx instance
     * @param env environment where to run the test
     * @param partitionName partition where to run the test
     * @throws LogicException if it is not possible to find given env/partition in the config.groovy
     */
    public void setPartition(String env, String partitionName) throws LogicException {
        partition = getPartition(env, partitionName);
    }

    /**
     * Reads the saved test results of a given logic for a given preset from the disk
     * @param presetName preset name (filename without extensions)
     * @return the list of the formula element results
     * @throws LogicException if the results/*.json file couldn't be read or parsed
     */
    public List<FormulaErrorElementResult> getSavedResults(String presetName) throws IOException {
        return LogicSerializer.readResultJson(logic.getDirPath(), presetName);
    }

    /**
     * Calls the test formula of a given logic. Parameter
     * @param presetName preset name (filename without extensions)
     * @return
     * @throws Exception
     */
    public List<FormulaErrorElementResult> getActualResults(String presetName) throws Exception {
        LogicSerializer logicSerializer = new LogicSerializer(logic);
        Map<String, Object> paramsMap = LogicSerializer.convertParamListToMap(logicSerializer.readParamsFile(presetName));
        MetadataCache metadataCache = MetadataProvider.getInstance().getMetadataInstance(partition);
        AbstractLogicSerializer.convertParameters(new AbstractLogicSerializer.CachedEntry(), partition.getClient(), paramsMap, logic, metadataCache);
        return PfxCommonService.testExecFormula(partition.getClient(), logic, paramsMap);
    }

    /**
     * Converts List of FormulaErrorElementResult to a HashMap with ElementName as the key
     * @param results from test formula
     * @return Map of results
     */
    public Map<String, FormulaErrorElementResult> convertFormulaResultsToMap(List<FormulaErrorElementResult> results) {
        Map<String, FormulaErrorElementResult> res = new HashMap();
        for (FormulaErrorElementResult result : results) {
            res.put(result.getElementName(), result);
        }
        return res;
    }

    /**
     * Loads the saved results and runs the test formula for the given preset
     * @param presetName preset name (filename without extensions)
     * @throws Exception
     */
    public void loadResults(String presetName) throws Exception {
        setSavedResults(presetName);
        setActualResults(presetName);
    }

    /**
     * Reads the saved test results of a given logic for a given preset from the disk
     * @param presetName preset name (filename without extensions)
     * @throws LogicException
     */
    public void setSavedResults(String presetName) throws IOException {
        savedResults = convertFormulaResultsToMap(getSavedResults(presetName));
    }

    /**
     * Runs the test formula for the given logic and parameter preset
     * @param presetName preset name (filename without extensions)
     * @throws Exception
     */
    public void setActualResults(String presetName) throws Exception {
        currentResults = convertFormulaResultsToMap(getActualResults(presetName));
    }


    public Map<String, FormulaErrorElementResult> getSavedResults() {
        return savedResults;
    }

    public Map<String, FormulaErrorElementResult> getCurrentResults() {
        return currentResults;
    }
}
